(*
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: September 2013
  (c) INRIA
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import Recdef.
Require Import SiaamTable.
Require Import SiaamStructures.
Require Import SiaamUtils.
Require Import SiaamReach.
Require Import SiaamRules.
Require Import SiaamProofWF.





Definition obj_mark (o : object) : Prop :=
  exists (f : fid), set_In (f,vmark) o.

Definition obj_clean (o : object) : Prop :=
  forall (f : fid)(v: value), set_In (f,v) o -> v <> vmark.

Lemma obj_mark_or_clean : forall (o : object), { obj_mark o } + { obj_clean o }.
Proof.
  intros o.
  induction o as [ | (f,v) os].
  Case "o is nil".
    right. unfold obj_clean.
    intros g w H. inversion H.
  Case "o is (f,v)::os".
    inversion IHos.
    SCase "mark".
      left. unfold obj_mark.
      unfold obj_mark in H.
      inversion H as [g G].
      exists g.
      simpl. tauto.
    SCase "clean".
      value_cases (destruct v) SSCase; 
      try (
         right; unfold obj_clean; intros g w;
         simpl; intros G; inversion G as [E | S];
           [inversion E; intros C; inversion C |
           unfold obj_clean in H; apply H with (f := g); trivial]; fail).
     SSCase "vmark".
       left.
       unfold obj_mark.
       exists f.
       simpl. 
       tauto.
Qed.

Lemma obj_clean_not_mark : forall (o : object), obj_clean o -> ~ obj_mark o.
Proof.
  intros o C.
  intros M.
  unfold obj_clean in C.
  unfold obj_mark in M.
  inversion M as [f S].
  apply C in S. apply S. trivial.
Qed.

Lemma obj_mark_not_clean : forall (o : object), obj_mark o -> ~ obj_clean o.
Proof.
  intros o M C.
  unfold obj_mark in M.
  inversion M as [f S].
  unfold obj_clean in C.
  apply C in S.
  apply S. trivial.
Qed.


Definition addr_mark (h: heap) (a : addr): Prop :=
  exists (o : object), set_In (a,o) h /\ obj_mark o.

Definition addr_clean (h : heap) (a : addr) : Prop :=
  forall (o : object), set_In (a,o) h -> obj_clean o.





Lemma addr_clean_larger_heap : forall (h1 h2 : heap) (a :addr),
  wf_heap h2 ->
    set_In a (addrs_from_heap h1) ->
      addr_clean h1 a ->
        subset h1 h2 ->
          addr_clean h2 a.
Proof.
  intros h1 h2 a W Sa A S.
  unfold addr_clean.
  intros o H.
  unfold addr_clean in A.
  apply A.
  apply addrs_from_heap_correct1 in Sa.
  inversion Sa as [o' S1].
  assert (C2 : o' = o).
    unfold wf_heap in W. apply W with (a := a).
    apply S. trivial. trivial.
  subst.
  trivial.
Qed.


Lemma addr_mark_not_clean : forall (h : heap)(a : addr),
  addr_mark h a -> ~ addr_clean h a.
Proof.
  intros h a M.
  intros C.
  unfold addr_mark in M.
  unfold addr_clean in C.
  inversion M as [o H].
  inversion H as [S Mo].
  apply C in S.
  apply obj_mark_not_clean in Mo.
  apply Mo.
  trivial.
Qed.

Lemma addr_clean_not_mark : forall (h : heap)(a : addr),
  addr_clean h a -> ~ addr_mark h a.
Proof.
  intros h a C.
  intros M.
  unfold addr_mark in M.
  unfold addr_clean in C.
  inversion M as [o H].
  inversion H as [S Mo].
  apply C in S.
  apply obj_mark_not_clean in Mo.
  apply Mo.
  trivial.
Qed.


Inductive oreachable : heap -> addr -> addr -> aid -> owners -> Prop :=
  | or_refl : forall (h : heap)(a : addr)(o : aid)(owns : owners),
                set_In (a,Some o) owns ->
                    oreachable h a a o owns
  | or_trans : forall (h : heap)(a b c: addr)(o : aid)(owns : owners),
                 set_In (a,Some o) owns -> 
                   ofield h a b -> 
                     set_In (b,Some o) owns ->
                       oreachable h b c o owns ->
                         oreachable h a c o owns.




Definition trans_mark (h: heap)(a:addr)(o:aid)(owns:owners): Prop :=
  exists (b: addr), oreachable h a b o owns /\ addr_mark h b.

Definition trans_clean (h : heap) (a: addr) (o : aid) (owns : owners) : Prop :=
  forall (b : addr), oreachable h a b o owns -> addr_clean h b.


Lemma trans_mark_not_clean : forall (h : heap)(a: addr) (o : aid) (owns : owners),
  trans_mark h a o owns -> ~ trans_clean h a o owns.
Proof.
  intros h a o owns M C.
  unfold trans_mark in M.
  inversion M as [b H].
  inversion H as [H1 H2].
  unfold trans_clean in C.
  apply C in H1.
  apply addr_clean_not_mark in H1.
  apply H1.
    trivial.
Qed.

Lemma trans_clean_not_mark : forall (h : heap)(a: addr) (o : aid) (owns : owners),
  trans_clean h a o owns -> ~ trans_mark h a o owns.
Proof.
  intros h a o owns C M.
  unfold trans_mark in M.
  inversion M as [b [R A]].
  unfold trans_clean in C.
  apply C in R.
  apply addr_clean_not_mark in R.
  apply R. trivial.
Qed.



Definition loc_mark (h : heap)(l : locals)(o : aid)(owns: owners): Prop :=
  (exists (i: vid), set_In (i, vmark) l) \/
  (exists (i : vid)(a : addr), set_In (i,vadd a) l /\ trans_mark h a o owns).

Definition loc_clean (h : heap)(l: locals)(o:aid)(owns: owners): Prop :=  
  forall (i: vid)(v:value), 
    set_In (i,v) l ->
      v <> vmark /\ (forall (a:addr), v = vadd a -> trans_clean h a o owns).


Lemma loc_mark_not_clean : forall (h : heap)(l : locals)(o : aid)(owns: owners),
  loc_mark h l o owns -> ~ loc_clean h l o owns.
Proof.
  intros h l o owns M.
  intros C.
  unfold loc_mark in M.
  unfold loc_clean in C.
  inversion M as [ [i Si] | [i [a [S T]]]].
  Case "i Si".
    apply C in Si. apply Si. trivial.
  Case "i a S T".
    apply C in S. inversion S as [S1 S2].
    apply trans_mark_not_clean in T.
    apply T. apply S2. trivial.
Qed.


Lemma loc_clean_not_mark : forall (h : heap)(l : locals)(o : aid)(owns: owners),
  loc_clean h l o owns -> ~ loc_mark h l o owns.
Proof.
  intros h l o owns C M.
  unfold loc_clean in C.
  unfold loc_mark in M.
  inversion M as [ [i Si] | [i [a [S T]]]].
  Case "i Si".
    apply C in Si. apply Si. trivial.
  Case "i a S T".
    apply C in S. inversion S as [S1 S2].
    apply trans_mark_not_clean in T.
    apply T. apply S2. trivial.
Qed.  



Definition ac_mark (h : heap) (a : actor) (owns : owners): Prop :=
  loc_mark h (snd a) (fst a) owns.

Definition ac_clean (h : heap)(a: actor)(owns:owners): Prop :=
  loc_clean h (snd a) (fst a) owns.


Lemma ac_mark_not_clean : forall (h : heap) (a : actor) (owns : owners),
  ac_mark h a owns -> ~ ac_clean h a owns.
Proof.
  unfold ac_mark. unfold ac_clean.
  intros h a owns M C.
  apply loc_mark_not_clean in M.
  apply M. trivial.
Qed.

Lemma ac_clean_not_mark : forall (h : heap) (a : actor) (owns : owners),
  ac_clean h a owns -> ~ ac_mark h a owns.
Proof.
  unfold ac_mark. unfold ac_clean.
  intros h a owns M C.
  apply loc_mark_not_clean in M.
  apply M. trivial.
Qed.

Definition conf_mark (c : conf) : Prop :=
  exists (a : actor), set_In a (acs c) /\ ac_mark (shp c) a (ows c).

Reserved Notation " t1 '=@' a '=>' t2 " (at level 40).

Inductive mred : aid -> conf -> conf -> Prop :=
  | mred_silent : forall (c1 c2 : conf) (e a: aid),
                    c1 =s e => c2 -> c1 =@ a => c2
  | mred_fread : forall (c1 c2 : conf)(e a: aid), c1 =fr e => c2  -> c1 =@ a => c2
  | mred_fwrite : forall (c1 c2 : conf)(e a: aid), c1 =fw e => c2  -> c1 =@ a => c2
  | mred_newact : forall (c1 c2 : conf)(e a: aid), c1 =nwa e => c2 -> c1 =@ a => c2
  | mred_newmbox : forall (c1 c2 : conf)(e a: aid), c1 =nwm e => c2 -> c1 =@ a => c2
  | mred_newobj : forall (c1 c2 : conf)(e a: aid), c1 =nwo e => c2 -> c1 =@ a => c2
  | mred_send : forall (c1 c2 : conf)(e a: aid), c1 =snd e => c2 -> c1 =@ a => c2
  | mred_rcv : forall (c1 c2 : conf)(e a: aid), 
                  e <> a -> c1 =rcv e => c2 -> c1 =@ a => c2

    where " t1 '=@' a '=>' t2 " := (mred a t1 t2).



Reserved Notation " t1 '=@' a '=>*' t2 " (at level 40).

Inductive mredt : forall (a:aid), conf -> conf -> Prop :=
  | mredt_refl : forall (c : conf)(a : aid), c =@ a =>* c 
  | mredt_trans : forall (c1 c2 c3 : conf)(a : aid),
                  c1 =@ a => c2 ->
                  c2 =@ a =>* c3 ->
                    c1 =@ a =>* c3

    where " t1 '=@' a '=>*' t2 " := (mredt a t1 t2).


Lemma reds_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =s a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros e S E.
  subst. simpl.
  destruct e as (u,lu).
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.
      

Lemma redfr_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =fr a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.
      

Lemma redfw_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =fr a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.

Lemma rednwa_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =nwa a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  apply set_add_intro. right.
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.

Lemma rednwm_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =nwm a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.

Lemma rednwo_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =nwo a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.

Lemma redsnd_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =snd a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  trivial.
Qed.

Lemma redrcv_frame : forall (c1 c2 : conf) (a : aid)(e : actor),
  c1 =rcv a => c2 ->
    set_In e (acs c1) ->
      (fst e) <> a ->
        set_In e (acs c2).
Proof.
  intros c1 c2 a e R.
  generalize dependent e.
  induction R; intros (u,lu) S E.
  subst. simpl.
  apply up_table_correct8.
    simpl in E. auto.
    trivial.
Qed.



Lemma action_keeps_clean : 
  forall (h : heap)(l1 l2 : locals)(o : aid)(owns: owners),
  action l1 l2 ->
    loc_clean h l1 o owns ->
      loc_clean h l2 o owns.
Proof.
  unfold loc_clean.
  intros h l1 l2 o owns A L1.
  generalize dependent h.
  generalize dependent o.
  generalize dependent owns.
  action_cases (induction A) Case ; intros owns o h L1 k v S.
  Case "a_nwbool".
    subst.
    apply set_add_elim in S.
    inversion S as [E | I].
    SCase "E".
      inversion E. subst. split.
        intros C. inversion C.
        intros x V. inversion V.
    SCase "I".
      apply L1 with (i := k). trivial.
  Case "a_nwnat".
    subst.
    apply set_add_elim in S.
    inversion S as [E | I].
    SCase "E".
      inversion E. subst. split.
        intros C. inversion C.
        intros x V. inversion V.
    SCase "I".
      apply L1 with (i := k). trivial.
  Case "a_nwnull".
    subst.
    apply set_add_elim in S.
    inversion S as [E | I].
    SCase "E".
      inversion E. subst. split.
        intros C. inversion C.
        intros x V. inversion V.
    SCase "I".
      apply L1 with (i := k). trivial.
  Case "a_compat".
    subst.
    unfold up_locals in S.
    apply up_table_elim in S.
    inversion S as [E | I].
      inversion E. subst. split.
        apply L1 in H0. tauto.
        apply L1 in H0. tauto.
      apply L1 in I. trivial.
  Case "a_forget".
    apply H in S. apply L1 with (i := k). trivial.
Qed.


Lemma reds_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =s (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
  intros c1 c2 a1 a2 a3 W S1 S2 C1 R L.
  unfold ac_clean.
  unfold ac_clean in C1.
  destruct a1 as (e1, l1).
  destruct a2 as (e2, l2).
  destruct a3 as (e3,l3).
  simpl in C1.
  simpl in L.
  simpl in R.
  simpl.
  destruct (Aeq_aid e2 e1).
  Case "e1 = e2".
    subst.
    assert (A1: wf_actors (acs c1)). unfold wf_conf in W. tauto.
    assert (A2: l2 = l1).
      unfold wf_actors in A1. unfold wf_table in A1. apply A1 with (n := e1).
      trivial. trivial.
    subst.
    induction R.
    subst. simpl. simpl in L.
    assert (A3: l = l1).
      unfold wf_actors in A1. unfold wf_table in A1. apply A1 with (n := a).
      trivial. trivial.
    subst.
    assert (A4 : Some (a,l') = lookup_actor (up_actors a l' (acs c1)) a).
      symmetry. apply lookup_actor_up.
      rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      rewrite <- A4 in L.
      inversion L.
      subst.
      apply action_keeps_clean with (l1 := l1).
        trivial.
        trivial.
  Case "e1 <> e2".
    assert (A: set_In (e1,l1) (acs c2)).
      apply reds_frame with (c1 := c1)(a := e2).
        trivial. trivial. simpl. auto.
    induction R.
    subst. simpl. simpl in L. simpl in A.
    assert (B : Some (e1,l1) = lookup_actor (up_actors a l' (acs c1)) e1).
      apply lookup_actor_up_not.
        unfold wf_conf in W. tauto.
        trivial.
        trivial.
      rewrite <- B in L. inversion L. subst.
      trivial.
Qed.
    

    
Lemma trans_clean_not_owned : 
  forall (h : heap) (a : addr) (o : aid) (owns : owners),
   ~ (set_In (a, Some o) owns) -> trans_clean h a o owns.
Proof.
  intros h a o owns H.
  unfold trans_clean.
  intros b Rb.
  induction Rb.
  Case "refl".
    apply ex_falso_quodlibet. tauto.
  Case "trans".
    apply ex_falso_quodlibet. tauto.
Qed.


Lemma up_local_var_keeps_clean : 
  forall (l : locals)(e : aid)(a: addr)(f: fid)(i j : vid)(v w : value)
         (o : object)(c: conf),
  wf_conf c ->
  set_In (e,l) (acs c) ->
  set_In (a, o) (shp c) ->
  set_In (f, v) o -> 
  set_In (a, Some e) (ows c) ->
  v_compat w v ->
  loc_clean (shp c) l e (ows c) ->
  set_In (i, w) l ->
  set_In (j, vadd a) l ->
    loc_clean (shp c) (up_locals i v l) e (ows c).
Proof.
  intros l e a f i j v w o c; 
  intros W S1 S2 S3 S4 C L S5 S6.
  unfold loc_clean.
  unfold loc_clean in L.
  intros k u H.
  unfold up_locals in H. 
  apply up_table_elim in H.
  inversion H as [E | I].
  Case "E".
    inversion E. subst.
    assert (A: trans_clean (shp c) a e (ows c)).
      apply L in S6. inversion S6 as [N F]. apply F. trivial.
    assert (B : oreachable (shp c) a a e (ows c)).
      apply or_refl.
        trivial.
    split.
      unfold trans_clean in A. apply A in B.
       unfold addr_clean in B.
       apply B in S2.
       unfold obj_clean in S2.
       apply S2 with (f := f). trivial.
      intros b V.
        destruct (set_In_dec Aeq_oentry (b, Some e) (ows c)).
        SCase "b owner = e".
          assert (Rb : oreachable (shp c) a b e (ows c)).
            apply or_trans with (b := b).
              trivial.
              unfold ofield. exists o.
              split.
                trivial.
                exists f. rewrite V in S3. trivial.
                trivial.
              apply or_refl.
                trivial.
          unfold trans_clean.
          unfold trans_clean in A.
          intros x Ox.
          apply A.
          apply or_trans with (b := b).
            trivial.
            unfold ofield. exists o.
              split. trivial.
              exists f. rewrite V in S3. trivial.
              trivial.
            trivial.
         apply trans_clean_not_owned. trivial.
  Case "I".
    apply L with (i := k). trivial.
Qed.
    
  


Lemma redfr_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =fr (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
  intros c1 c2 a1 a2 a3 W S1 S2 C1 R L.
  destruct a1 as (e1,l1).
  destruct a2 as (e2,l2).
  destruct a3 as (e3,l3).
  simpl in R.
  simpl in L.
  destruct (Aeq_aid e1 e2).
  Case "e1 =e2".
    subst.
    apply lookup_actor_elim in L. inversion L as [E S].
    subst.
    induction R.
    subst.
    simpl. simpl in L. simpl in S.
    unfold up_actors in S. apply up_table_correct6 in S. subst.
    assert (WA : wf_actors (acs c1)).
      unfold wf_conf in W. tauto.
    assert (A: l = l1).
      unfold wf_actors in WA. unfold wf_table in WA. apply WA with (n := e).
      trivial. trivial.
    subst.
    unfold ac_clean. simpl.
    unfold ac_clean in C1. simpl in C1.
    apply up_local_var_keeps_clean with 
    (a:= a) (f := f) (j := j) (w := w) (o := o).
      trivial. trivial. trivial. trivial. trivial. trivial. trivial. trivial.
      trivial.
  Case "e1 <> e2".
    assert (A: set_In (e1,l1) (acs c2)). 
      apply redfr_frame with (c1 := c1)(a := e2).
      trivial. trivial. simpl. trivial.
    assert (WA: wf_actors (acs c2)).
      apply redfr_preserves_wf with (c2 := c2)(a := e2) in W.
      unfold wf_conf in W. tauto. trivial.
    assert (B: l3 = l1).
      unfold wf_actors in WA. unfold wf_table in WA.
      apply WA with (n := e1).
        apply lookup_actor_elim in L. inversion L as [E  S]. subst.
          trivial.
        trivial.
      subst.
      apply lookup_actor_elim in L. inversion L as [E S]. subst.
      induction R.
      subst. simpl. simpl in A. simpl in WA. simpl in L. simpl in S.
      trivial.
Qed.

Definition value_clean (h: heap)(v : value) (o : aid) (owns: owners) : Prop :=
  v <> vmark /\ (forall (a:addr), v = vadd a -> trans_clean h a o owns).


Lemma clean_addr_upobject_clean :
  forall (h : heap)(a x: addr)(oa : object)(v : value) 
         (f : fid) (o : aid) (owns:owners),
  wf_heap h ->
    value_clean h v o owns ->
      set_In (a,oa) h -> 
        addr_clean h x -> 
          addr_clean (up_heap a (up_object f v oa) h) x.
Proof.
  intros h a x oa v f o owns W V S A.
  destruct (Aeq_addr x a).
  Case "x = a".
    subst. unfold addr_clean.
    intros ob Sb.
    unfold up_heap in Sb.
    apply up_table_elim in Sb.
    inversion Sb as [Eb | Ib].
    SCase "Eb".
      inversion Eb.
      unfold obj_clean.
      intros g vg Sg.
      unfold up_object in Sg.
      apply up_table_elim in Sg.
      inversion Sg as [Eg | Ig].
        inversion Eg. subst. unfold value_clean in V. tauto.
        unfold addr_clean in A. apply A in S. unfold obj_clean in S.
         apply S in Ig. trivial.
    SCase "Ib".
      unfold addr_clean in A. apply A. trivial.
  Case "x <> a".
    unfold addr_clean in A.
    destruct (set_In_dec Aeq_addr x (addrs_from_heap h)).
    SCase "x in h".
      apply addrs_from_heap_correct1 in s.
      inversion s as [o' Sx].
      assert (Ix : set_In (x,o') (up_heap a (up_object f v oa) h)).
        unfold up_heap. apply up_table_correct8. auto. trivial.
      unfold addr_clean.
      intros ox Hx.
      assert (WF : wf_heap (up_heap a (up_object f v oa) h)).
        rewrite wf_heap_table. unfold up_heap.
        apply wf_up_table. trivial.
      assert (Eo: ox = o').
        unfold wf_heap in WF. apply WF with (a0 := x).
        trivial. trivial.
      subst.
      apply A. trivial.
    SCase "x not in h".
      unfold addr_clean. intros o' Sx.
      apply ex_falso_quodlibet. apply n0.
      rewrite addrs_from_heap_indexes.
      apply indexes_in_up_table22 with (i := a) (v := (up_object f v oa)).
      unfold up_heap in Sx.
      apply indexes_in_table_correct1 with (Jeq_dec := Aeq_addr) in Sx.
      trivial.
Qed.


Lemma oreachable_up_object_not_addr : 
  forall (h1 h2 : heap)(a b x: addr)(oa : object)(v : value) 
         (f : fid) (o : aid) (owns:owners),
  set_In (a,oa) h1 ->
   h2 = up_heap a (up_object f v oa) h1 ->
    oreachable h2 b x o owns ->
      (forall (c : addr), v <> vadd c) ->
        oreachable h1 b x o owns.
Proof.
  intros h1 h2 a b x oa v f o owns S H R V.
  induction R.
  Case "refl".
    constructor. trivial.
  Case "trans".
    rename a0 into e.
    apply or_trans with (b := b).
      trivial.
      unfold ofield in H1. inversion H1 as [oe [Se Fe]].
      rewrite H in Se.
      unfold up_heap in Se.
      apply up_table_elim in Se.
      inversion Se as [Ee | Ie].
      SCase "Ee".
        inversion Ee. subst.
        inversion Fe as [g Ig].
        unfold up_object in Ig.
        apply up_table_elim in Ig.
        inversion Ig as [N | Y].
        SSCase "N".
          apply ex_falso_quodlibet. inversion N. 
          assert (NV : v <> vadd b). apply V.
          apply NV. auto.
        SSCase "Y".
          unfold ofield.
          exists oa.
          split.
            trivial.
            exists g. trivial.        
      SCase "Ie".
        inversion Fe as [g Ig].
        unfold ofield.
        exists oe.
        split.
          trivial.
          exists g. trivial.
     trivial.
     apply IHR.
     trivial.
Qed.


Lemma oreachable_trans :
  forall (h : heap)(a b c : addr) (o : aid) (owns : owners),
  oreachable h a b o owns ->
    oreachable h b c o owns ->
      oreachable h a c o owns.
Proof.
  intros h a b c o owns R1 R2.
  induction R1.
  Case "refl".
    trivial.
  Case "trans".
    apply or_trans with (b := b).
      trivial.
      trivial.
      trivial.
    apply IHR1.
    trivial.
Qed.


Lemma oreachable_up_object_addr : 
  forall (h1 h2 : heap)(a b c x: addr)(oa : object) 
         (f : fid) (o : aid) (owns:owners),
  wf_heap h1 ->
  complete_heap h1 ->
   set_In (a,oa) h1 ->
   h2 = up_heap a (up_object f (vadd c) oa) h1 ->
    oreachable h2 b x o owns ->
         (oreachable h1 b a o owns /\ oreachable h1 c x o owns)
         \/ (oreachable h1 b x o owns).
Proof.
  intros h1 h2 a b c x oa f o owns W CH Sa H R.
  induction R.
  Case "refl".
    right. constructor. trivial.
  Case "trans".
    rename a0 into s.
    rename c0 into x.
    assert (A : oreachable h1 b a o owns /\ oreachable h1 c x o owns \/
      oreachable h1 b x o owns).
      apply IHR. trivial.
    inversion A as [A1 | A2].
    SCase "A1".
      destruct (Aeq_addr s a).
      SSCase "s = a".
        subst. left. split.
          constructor. trivial.
          tauto.
      SSCase "s <> a".
        unfold ofield in H1. inversion H1 as [os [Is Fs]].
        assert (A2: set_In (s,os) h1).
          apply up_table_correct4 with (Jeq_dec := Aeq_addr)
                                       (n := a)
                                       (v := (up_object f (vadd c) oa)).
           auto.
           rewrite H in Is. trivial.
        left. split.
        apply or_trans with (b:= b).
          trivial.
          unfold ofield. exists os. tauto. trivial.
          tauto.
        tauto.
    SCase "A2".
      destruct (Aeq_addr s a).
      SSCase "s = a".
        subst.
        unfold ofield in H1.
        inversion H1 as [oa' [Soa Foa]].
        assert (B : oa' = up_object f (vadd c) oa).
          apply wf_up_table with 
          (Jeq_dec := Aeq_addr)(n:= a)(a:= (up_object f (vadd c) oa)) in W.
          unfold wf_table in W.
          apply W with (n:= a).
          trivial.
          apply up_table_correct2.
          apply indexes_in_table_correct1 with (v := oa).
          trivial.
        inversion Foa as [g Ig].
        rewrite B in Ig.
        unfold up_object in Ig.
        apply up_table_elim in Ig.
        inversion Ig as [Eg | Sg].
        SSSCase "g = f".
          inversion Eg. subst. left. split.
          constructor. trivial.
          trivial.
        SSSCase "g <> f".
          right. apply or_trans with (b := b).
            trivial.
            unfold ofield.
            exists oa.
            split.
            trivial.
            exists g.
            trivial.
            trivial.
            trivial.
      SSCase "s <> a".
        unfold ofield in H1.
        inversion H1 as [os [Is Fs]].
        rewrite H in Is.
        unfold up_heap in Is.
        apply up_table_elim in Is.
        inversion Is as [Cs | Ss].
        SSSCase "Cs".
          inversion Cs. apply ex_falso_quodlibet. apply n. trivial.
        SSSCase "Ss".
          right.
          apply or_trans with (b := b).
          trivial.
          unfold ofield. exists os. tauto.
          trivial. trivial.
Qed.





Lemma ac_clean_all_values_clean :
  forall (a: aid)(l : locals)(h : heap)(owns: owners),
   ac_clean h (a,l) owns <-> 
    (forall (v : value), 
     set_In v (values_in_table vid value Aeq_value l) -> 
       value_clean h v a owns).
Proof.
  intros a l h owns.
  split.
  Case "->".
   intros AC.
   intros v S.
   unfold ac_clean in AC.
   simpl in AC.
   unfold loc_clean in AC.
   unfold value_clean.
   apply values_in_table_correct2 in S.
   inversion S as [n Sn].
   apply (AC n). trivial.
  Case "<-".
    intros V.
    unfold ac_clean. simpl.
    unfold loc_clean.
    intros i v S.
    unfold value_clean in V.
    apply V.
    apply values_in_table_correct1 with (Aeq_dec := Aeq_value) in S.
    trivial.
Qed.

Lemma clean_values_upobject_clean : 
  forall (h1 h2 : heap)(a : addr)(oa : object)(v w: value) 
         (f : fid) (o : aid) (owns:owners),
  wfc_heap h1 ->
   trans_clean h1 a o owns ->
    value_clean h1 v o owns ->
     value_clean h1 w o owns -> 
      set_In (a,oa) h1 ->
       set_In (a, Some o) owns ->
        h2 = up_heap a (up_object f w oa) h1 ->
         value_clean h2 v o owns.
Proof.
  intros h1 h2 a oa v w f o owns W T Vv Vw SH SO H.
  unfold value_clean.
  split. unfold value_clean in Vv. tauto.
    intros b Vb.
    unfold trans_clean.
    unfold value_clean in Vv.
    inversion Vv as [Tr Fv].
    clear Tr.
    apply Fv in Vb.
    clear Fv.
    intros x Rx.
    value_cases (destruct w) Case.
    Case "vnull".
      apply oreachable_up_object_not_addr 
      with (h1 := h1)(a:=a)(oa := oa) (v:= vnull)(f:=f) in Rx.
        unfold addr_clean. intros ox Sox.
          rewrite H in Sox. unfold up_heap in Sox.
          apply up_table_elim in Sox.
          inversion Sox as [Eox | Iox].
          SCase "Eox".
            inversion Eox. unfold trans_clean in T.
            assert (Aa : addr_clean h1 a). apply T. constructor. trivial.
            unfold addr_clean in Aa.
            assert (Coa: obj_clean oa). apply Aa. trivial.
            unfold obj_clean.
            intros g vg Sg.
            unfold up_object in Sg.
            apply up_table_elim in Sg.
            inversion Sg as [Eg | Ig].
            SSCase "Eg".
              inversion Eg. intros C. inversion C.
            SSCase "Ig".
              unfold obj_clean in Coa. apply Coa with (f := g). trivial.
          SCase "Iox".
            unfold trans_clean in Vb.
            unfold addr_clean in Vb.
            apply Vb with (b0 := x). trivial. trivial. 
       trivial. trivial. intros c C. inversion C.
    Case "vmark".
      apply ex_falso_quodlibet. unfold value_clean in Vw. tauto.
    Case "vnat".
      apply oreachable_up_object_not_addr 
      with (h1 := h1)(a:=a)(oa := oa) (v:= vnat n)(f:=f) in Rx.
        unfold addr_clean. intros ox Sox.
          rewrite H in Sox. unfold up_heap in Sox.
          apply up_table_elim in Sox.
          inversion Sox as [Eox | Iox].
          SCase "Eox".
            inversion Eox. unfold trans_clean in T.
            assert (Aa : addr_clean h1 a). apply T. constructor. trivial.
            unfold addr_clean in Aa.
            assert (Coa: obj_clean oa). apply Aa. trivial.
            unfold obj_clean.
            intros g vg Sg.
            unfold up_object in Sg.
            apply up_table_elim in Sg.
            inversion Sg as [Eg | Ig].
            SSCase "Eg".
              inversion Eg. intros C. inversion C.
            SSCase "Ig".
              unfold obj_clean in Coa. apply Coa with (f := g). trivial.
          SCase "Iox".
            unfold trans_clean in Vb.
            unfold addr_clean in Vb.
            apply Vb with (b0 := x). trivial. trivial. 
       trivial. trivial. intros c C. inversion C.
    Case "vbool".
      apply oreachable_up_object_not_addr 
      with (h1 := h1)(a:=a)(oa := oa) (v:= vbool b0)(f:=f) in Rx.
        unfold addr_clean. intros ox Sox.
          rewrite H in Sox. unfold up_heap in Sox.
          apply up_table_elim in Sox.
          inversion Sox as [Eox | Iox].
          SCase "Eox".
            inversion Eox. unfold trans_clean in T.
            assert (Aa : addr_clean h1 a). apply T. constructor. trivial.
            unfold addr_clean in Aa.
            assert (Coa: obj_clean oa). apply Aa. trivial.
            unfold obj_clean.
            intros g vg Sg.
            unfold up_object in Sg.
            apply up_table_elim in Sg.
            inversion Sg as [Eg | Ig].
            SSCase "Eg".
              inversion Eg. intros C. inversion C.
            SSCase "Ig".
              unfold obj_clean in Coa. apply Coa with (f := g). trivial.
          SCase "Iox".
            unfold trans_clean in Vb.
            unfold addr_clean in Vb.
            apply Vb with (b0 := x). trivial. trivial. 
       trivial. trivial. intros c C. inversion C.
     Case "vadd".
       rename a0 into c.
       apply oreachable_up_object_addr 
       with (h1 := h1) (a:= a)(c := c)(oa := oa)(f := f) in Rx.
         inversion Rx as [R1 | R2].
           SCase "R1".
             rewrite H. apply clean_addr_upobject_clean with (o := o)(owns := owns).
               unfold wfc_heap in W. tauto.
               trivial.
               trivial.
               unfold value_clean in Vw.
               assert (Tc : trans_clean h1 c o owns). apply Vw. trivial.
               unfold trans_clean in Tc. apply Tc. tauto.
           SCase "R2".
             rewrite H. apply clean_addr_upobject_clean with (o := o)(owns := owns).
               unfold wfc_heap in W. tauto.
               trivial.
               trivial.
               unfold trans_clean in Vb.
               apply Vb. trivial.
               unfold wfc_heap in W. tauto.
               unfold wfc_heap in W. tauto.
               trivial. trivial.
    Case "vaid".
      apply oreachable_up_object_not_addr 
      with (h1 := h1)(a:=a)(oa := oa) (v:= vaid a0)(f:=f) in Rx.
        unfold addr_clean. intros ox Sox.
          rewrite H in Sox. unfold up_heap in Sox.
          apply up_table_elim in Sox.
          inversion Sox as [Eox | Iox].
          SCase "Eox".
            inversion Eox. unfold trans_clean in T.
            assert (Aa : addr_clean h1 a). apply T. constructor. trivial.
            unfold addr_clean in Aa.
            assert (Coa: obj_clean oa). apply Aa. trivial.
            unfold obj_clean.
            intros g vg Sg.
            unfold up_object in Sg.
            apply up_table_elim in Sg.
            inversion Sg as [Eg | Ig].
            SSCase "Eg".
              inversion Eg. intros C. inversion C.
            SSCase "Ig".
              unfold obj_clean in Coa. apply Coa with (f := g). trivial.
          SCase "Iox".
            unfold trans_clean in Vb.
            unfold addr_clean in Vb.
            apply Vb with (b0 := x). trivial. trivial. 
       trivial. trivial. intros c C. inversion C.
    Case "vmid".
      apply oreachable_up_object_not_addr 
      with (h1 := h1)(a:=a)(oa := oa) (v:= vmid m)(f:=f) in Rx.
        unfold addr_clean. intros ox Sox.
          rewrite H in Sox. unfold up_heap in Sox.
          apply up_table_elim in Sox.
          inversion Sox as [Eox | Iox].
          SCase "Eox".
            inversion Eox. unfold trans_clean in T.
            assert (Aa : addr_clean h1 a). apply T. constructor. trivial.
            unfold addr_clean in Aa.
            assert (Coa: obj_clean oa). apply Aa. trivial.
            unfold obj_clean.
            intros g vg Sg.
            unfold up_object in Sg.
            apply up_table_elim in Sg.
            inversion Sg as [Eg | Ig].
            SSCase "Eg".
              inversion Eg. intros C. inversion C.
            SSCase "Ig".
              unfold obj_clean in Coa. apply Coa with (f := g). trivial.
          SCase "Iox".
            unfold trans_clean in Vb.
            unfold addr_clean in Vb.
            apply Vb with (b0 := x). trivial. trivial. 
       trivial. trivial. intros c C. inversion C.
Qed.

Lemma oreachable_upheap_not_owner : 
  forall (h1 h2 : heap) (a b x: addr)(o2 : object) (o e : aid) (owns : owners),
    wf_owners owns -> 
    oreachable h2 b x e owns ->
     set_In (a, Some o) owns ->
      e <> o ->
       h2 = up_heap a o2 h1 ->
        oreachable h1 b x e owns.
Proof.
  intros h1' h2 a' b x o2' o' e owns W R.
  generalize dependent h1'.
  generalize dependent a'.
  generalize dependent o2'.
  generalize dependent o'.
  induction R as [h2 b e owns | h2 b x y e owns]; intros o o2 a h1 S N H2.
  Case "refl".
    constructor. trivial.
  Case "trans".
    apply or_trans with (b := x).
      trivial.
      unfold ofield. unfold ofield in H0. inversion H0 as [ob Hb].
        inversion Hb as [Ib Fb].
        rewrite H2 in Ib. 
        unfold up_heap in Ib.
        apply up_table_elim in Ib.
        inversion Ib as [Eb | Sb].
        SCase "Eb".
          inversion Eb. subst.
          apply ex_falso_quodlibet.
          apply N.
          assert (A: Some e = Some o).
            apply (W a). trivial. trivial.
          inversion A. trivial.
        SCase "Ib".
          exists ob. tauto.
          trivial.
          apply IHR with (o' := o) (o2' := o2)(a' := a).
            trivial. trivial. trivial. trivial.
Qed.
    
Lemma oreachable_owner : 
  forall (h : heap) (a b : addr) (o : aid) (owns : owners),
     oreachable h a b o owns ->
       set_In (b , Some o) owns.
Proof.
  intros h a b o owns R.
  induction R.
    trivial.
    apply IHR.
Qed.

Lemma clean_value_upheap_not_owner :
  forall (h1 h2 : heap) (a : addr) (o2 : object) (o e : aid) 
         (v : value) (owns : owners),
  wfc_heap h1 ->
   wf_owners owns ->
    set_In (a, Some o) owns ->
     e <> o ->
      value_clean h1 v e owns ->
       h2 = up_heap a o2 h1 ->
        value_clean h2 v e owns.
Proof.
  intros h1 h2 a o2 o e v owns WH WO S N V H.
  unfold value_clean.
  split.
    unfold value_clean in V. tauto.
    intros b Vb.
      unfold trans_clean.
      intros x Rx.
      unfold addr_clean.
      intros ox Sox.
      rewrite H in Sox.
      unfold up_heap in Sox.
      apply up_table_elim in Sox.
      inversion Sox as [Eox | Iox].
      Case "Eox".
        apply ex_falso_quodlibet. inversion Eox. subst.
        apply oreachable_owner in Rx.
        assert (A : Some e = Some o). apply (WO a). trivial. trivial.
        inversion A. auto.
      Case "Iox".
        unfold value_clean in V.
        apply oreachable_upheap_not_owner 
        with (h1 := h1)(a := a)(o2 := o2)(o:= o) in Rx.
          assert (A : trans_clean h1 b e owns).
            apply V. trivial.
           unfold trans_clean in A.
           apply A in Rx.
           unfold addr_clean in Rx.
           apply Rx. trivial.
        trivial. trivial. trivial. trivial.
Qed.
  
    

Lemma redfw_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =fw (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
  intros c1 c2 a1 a2 a3 W S1 S2 C R L.
  destruct a1 as (e1,l1).
  destruct a2 as (e2, l2).
  destruct a3 as (e3,l3).
  simpl in R.
  induction R.
  subst. simpl. simpl in L.
  destruct (Aeq_aid e1 e).
  Case "e1 = e".
    subst.
    assert (A1: l1 = l).
      SCase "Proof of A1".
        assert (WF : wf_actors (acs c1)). unfold wf_conf in W. tauto.
        unfold wf_actors in WF. unfold wf_table in WF.
        apply WF with (n := e).
        trivial.
        trivial.
    subst.
    assert (A2 : lookup_actor (acs c1) e = Some (e,l)).
      apply lookup_actor_intro. unfold wf_conf in W. tauto. trivial.
    rewrite A2 in L. inversion L. subst.
    apply ac_clean_all_values_clean.
    intros u Su.
    apply clean_values_upobject_clean with 
    (h1 := (shp c1))(a := a)(oa := o)(f := f)(w := w).
      unfold wf_conf in W. tauto.
      unfold ac_clean in C. simpl in C. unfold loc_clean in C.
      apply C in H1. apply H1. trivial.
      rewrite ac_clean_all_values_clean in C. apply (C u). trivial.
      rewrite ac_clean_all_values_clean in C. apply (C w). 
        apply values_in_table_correct1 with (Aeq_dec := Aeq_value) in H0. trivial.
      trivial.
      trivial.
      trivial.
  Case "e1 <> e".
    assert (A : lookup_actor (acs c1) e1 = Some (e1, l1)).
      apply lookup_actor_intro. unfold wf_conf in W. tauto. trivial.
    rewrite A in L. inversion L. subst.
    apply ac_clean_all_values_clean.
    intros u Su.
    apply clean_value_upheap_not_owner
    with (h1 := (shp c1))(a := a)(o2 := up_object f w o)(o := e).
      unfold wf_conf in W. tauto. 
      unfold wf_conf in W. tauto.
      trivial.
      trivial.
      rewrite ac_clean_all_values_clean in C.
        apply C. trivial.
      trivial.
Qed.


Lemma oreachable_new_object :
  forall (h1 h2 : heap)(e :aid)(a b x : addr)(o : object)(ows1 ows2 : owners),
  complete_heap h1 ->
   ~ set_In a (addrs_from_heap h1) ->
       h2 = set_add Aeq_sentry (a,o) h1 ->
         ows2 = set_add Aeq_oentry (a, Some e) ows1 ->
             oreachable h2 b x e ows2 ->
               b <> a ->
               oreachable h1 b x e ows1.
Proof.
  intros h1 h2 e a b x o ow1 ows2 CH N1 H A R N3.
  induction R.
  Case "refl".
    rename a0 into b. rename o0 into e. rename owns into ow2.
    rewrite A in H0. apply set_add_elim in H0.
    inversion H0 as [E | S].
      inversion E. apply ex_falso_quodlibet. apply N3. trivial.
      constructor. trivial.
  Case "trans".
    rename c into x. rename b into c. rename a0 into b.
    rename owns into ow2. rename o0 into e.
    rewrite A in H0.
    apply set_add_elim in H0.
    inversion H0 as [E | I].
      inversion E. apply ex_falso_quodlibet. apply N3. trivial.
      apply or_trans with (b := c).
        trivial.
        unfold ofield. rewrite H in H1.
          unfold ofield in H1. inversion H1 as [ob [Sb Fb]].
          apply set_add_elim in Sb.
          inversion Sb as [Eb | Ib].
            inversion Eb. apply ex_falso_quodlibet. apply N3. trivial.
            exists ob. tauto.
       rewrite A in H2. apply set_add_elim in H2.
       inversion H2 as [Ec | Ic].
         inversion Ec. apply ex_falso_quodlibet. unfold ofield in H1.
           inversion H1 as [ob [Sb Fb]]. inversion Fb as [f Sc].
           rewrite H in Sb. apply set_add_elim in Sb.
           inversion Sb as [Ebob | Ibob].
             inversion Ebob. apply N3. trivial.
             unfold complete_heap in CH. apply CH with (b := c) in Ibob.
             apply N1. subst. trivial.
             apply addrs_in_fields_correct2 in Sc. trivial.
             trivial.
       apply IHR.
        trivial. trivial. 
        unfold ofield in H1. inversion H1 as [ob [Eb Fb]]. inversion Fb as [f Fc].
        intros C. unfold complete_heap in CH.
        rewrite H in Eb. apply set_add_elim in Eb.
        inversion Eb as [Ebob | Ibob].
          inversion Ebob. apply N3. trivial.
          apply CH with (b := c) in Ibob.
            subst. apply N1. trivial.
            apply addrs_in_fields_correct2 in Fc. trivial.
Qed.
           
Lemma trans_clean_oreachable_trans_clean :
  forall (h : heap)(e : aid)(a b : addr)(owns : owners),
   trans_clean h a e owns ->
   oreachable h a b e owns ->
     trans_clean h b e owns.
Proof.
  intros h e a b owns T R.
  induction R.
  Case "refl".
    trivial.
  Case "trans".
    apply IHR.
    unfold trans_clean.
    intros x Rx.
    unfold trans_clean in T.
    apply T.
    apply or_trans with (b := b).
      trivial. trivial. trivial. trivial.
Qed.

Lemma trans_clean_new_object : 
 forall (h1 h2 : heap)(e :aid)
        (a : addr)(o : object)(ow1 ow2 : owners),
 complete_heap h1 ->
 not (set_In a (addrs_from_heap h1)) ->
 not (set_In a (addrs_from_owners ow1)) ->
 subset (addrs_in_fields o) (addrs_from_heap h1) ->
 h2 = set_add Aeq_sentry (a,o) h1 ->
 ow2 = set_add Aeq_oentry (a, Some e) ow1 ->
 (forall (v : value), set_In v (values_in_table fid value Aeq_value o) ->
                      value_clean h1 v e ow1) ->
   trans_clean h2 a e ow2.
Proof.
  intros h1 h2 e a o ow1 ow2 CH S1 S2 NS H OA C.
  unfold trans_clean.
  intros b Rb.
  induction Rb.
  Case "refl".
    unfold addr_clean.
    intros oa Soa.
    rewrite H in Soa.
    simpl in Soa.
    apply set_add_elim in Soa.
    inversion Soa as [E | I].
      SCase "E".
        inversion E. unfold obj_clean.
        intros f v Sf.
        unfold value_clean in C.
        apply C.
        apply values_in_table_correct1 with (n := f). trivial.
      SCase "I".
        apply ex_falso_quodlibet. apply S1.
        apply addrs_from_heap_correct2.
        exists oa. trivial.
  Case "trans".
    rename o0 into e.
    unfold addr_clean.
    apply oreachable_new_object 
    with (h1 := h1)(a := a)(o := o)(ows1 := ow1) in Rb.
      intros oc Soc.
      rewrite H in Soc.
      apply set_add_elim in Soc.
      inversion Soc as [Eoc | Ioc].
        inversion Eoc. apply ex_falso_quodlibet. 
          apply oreachable_owner in Rb.
          apply S2. subst.
          rewrite addrs_from_owners_indexes.
          apply indexes_in_table_correct1 with (v := Some e).
          trivial.
        unfold value_clean in C.
          unfold trans_clean in C.
          assert (A : set_In (vadd b) (values_in_table fid value Aeq_value o)).
            unfold ofield in H1.
            inversion H1 as [oa [Soa Foa]].
            rewrite H in Soa.
            apply set_add_elim in Soa.
            inversion Soa as [Eoa | Ioa].
              inversion Eoa. subst. inversion Foa as [f Vf].
                apply values_in_table_correct1 with (n := f).
                trivial.
              apply ex_falso_quodlibet. apply S1.
                rewrite addrs_from_heap_indexes. 
                apply indexes_in_table_correct1 with (v := oa).
                trivial.
          apply C in A. inversion A as [Vb ORb].
          apply ORb in Rb. unfold addr_clean in Rb.
          apply Rb. trivial. trivial.
        trivial.
        trivial.
        trivial.
        trivial.
        unfold ofield in H1.
          inversion H1 as [oa [Soa Foa]].
          inversion Foa as [f Vf].
          apply addrs_in_fields_correct2 in Vf.
          intros N. rewrite H in Soa.
          apply set_add_elim in Soa. 
          inversion Soa as [Eoa | Ioa].
            inversion Eoa. subst.
              apply NS in Vf. apply S1. trivial.
            apply S1. apply addrs_from_heap_correct2.
              exists oa. trivial.
Qed.
        
  
Lemma trans_clean_new_object_not_new :
  forall (h1 h2 : heap) (e : aid)
         (a b : addr)(o : object)(ow1 ow2 : owners),
  complete_heap h1 ->
  not (set_In a (addrs_from_heap h1)) ->
  not (set_In a (addrs_from_owners ow1)) ->
  subset (addrs_in_fields o)(addrs_from_heap h1) ->
  h2 = set_add Aeq_sentry (a,o) h1 ->
  ow2 = set_add Aeq_oentry (a, Some e) ow1 ->
 (forall (v : value), set_In v (values_in_table fid value Aeq_value o) ->
                      value_clean h1 v e ow1) ->
  trans_clean h1 b e  ow1 ->
  b <> a ->
    trans_clean h2 b e ow2.
Proof.
  intros h1 h2 e a b o ow1 ow2.
  intros CH S1 S2 S3 H SO C T N.
  unfold trans_clean.
  intros x RX.
  induction RX.
  Case "refl".
    rename a0 into b.
    rename o0 into e.
    rename owns into ow2.
    rename h into h2.
    unfold addr_clean.
    intros ob Sob.
    rewrite H in Sob.
    apply set_add_elim in Sob.
    inversion Sob as [Eob | Iob].
    SCase "b = a".
      inversion Eob.
      apply ex_falso_quodlibet. tauto.
    SCase "b ob in h1".
      unfold trans_clean in T.
      assert (A : addr_clean h1 b).
        apply T. constructor.
        rewrite SO in H0.
        apply set_add_elim in H0.
        inversion H0 as [E | I].
          inversion E. tauto.
          trivial.
      unfold addr_clean in A. apply A. trivial.
  Case "trans".
    rename owns into ow2.
    rename o0 into e.
    rename c into x.
    rename b into c.
    rename a0 into b.
    rename h into h2.
    apply IHRX.
      trivial. trivial. trivial.
      apply trans_clean_oreachable_trans_clean with (a := b).
        trivial.
        apply oreachable_new_object with (h2 := h2)(a := a)(o := o)(ows2 := ow2).
          trivial. trivial. trivial. trivial.
            apply or_trans with (b := c). 
              trivial. trivial. trivial.
              constructor. trivial. trivial.
              unfold ofield in H1.
                inversion H1 as [ob [Sob Fob]].
                inversion Fob as [f Ic].
                rewrite H in Sob.
                apply set_add_elim in Sob.
                inversion Sob as [E | I].
                  inversion E. apply ex_falso_quodlibet. tauto.
                  apply addrs_in_fields_correct2 in Ic.
                    unfold complete_heap in CH.
                    apply CH with (b := c) in I.
                    intros Ca. apply S1. subst. trivial.
                    trivial.
Qed.
            


Lemma oreachable_new_object_not_owner :
  forall (h1 h2 : heap)(e e':aid)(a b x : addr)(o : object)(ows1 ows2 : owners),
    ~ set_In a (addrs_from_owners ows1) ->
       h2 = set_add Aeq_sentry (a,o) h1 ->
         ows2 = set_add Aeq_oentry (a, Some e) ows1 ->
           e <> e' ->
            set_In (b, Some e') ows1 ->
             oreachable h2 b x e' ows2 ->
               oreachable h1 b x e' ows1.
Proof.
  intros h1 h2 e e' a b x o ows1 ows2.
  intros NS H OW NE SB RX.
  induction RX.
  Case "refl".
    rename a0 into b.
    rename o0 into ob.
    constructor. trivial.
  Case "trans".
    rename a0 into s.
    rename o0 into es.
    apply or_trans with (b := b).
      trivial.
      unfold ofield. unfold ofield in H1.
        inversion H1 as [os [Is Fs]]. rewrite H in Is.
        apply set_add_elim in Is. inversion Is as [Es | Ss].
          inversion Es. apply ex_falso_quodlibet.
            subst. apply NS. rewrite addrs_from_owners_indexes.
            apply indexes_in_table_correct1 with (Jeq_dec := Aeq_addr) in SB.
            trivial.
          exists os. tauto.
          rewrite OW in H2. apply set_add_elim in H2. 
            inversion H2 as [ES | SS].
              inversion ES. subst. tauto.
              trivial.
          apply IHRX. trivial. trivial. trivial.
            rewrite OW in H2. apply set_add_elim in H2.
            inversion H2 as [E | I].
              inversion E. subst. tauto.
              trivial.
Qed.


Lemma trans_clean_new_object_not_owner :
  forall (h1 h2 : heap) (e e': aid)
         (a b : addr)(o : object)(ow1 ow2 : owners),
  complete_heap h1 ->
   set_In (b, Some e') ow1 ->
    trans_clean h1 b e' ow1 ->
     not (set_In a (addrs_from_heap h1)) ->     
      not (set_In a (addrs_from_owners ow1)) ->
       e <> e'-> 
        h2 = set_add Aeq_sentry (a,o) h1 ->
         ow2 = set_add Aeq_oentry (a, Some e) ow1 ->
           trans_clean h2 b e' ow2.
Proof.
  intros h1 h2 e e' a b o ow1 ow2.
  intros CH S T NS1 NS2 NE H OW.
  unfold trans_clean.
  intros x Rx.
  apply oreachable_new_object_not_owner 
  with (h1 := h1)(e := e)(a := a)(o := o) (ows1 := ow1) in Rx.
    assert (TX: trans_clean h1 x e' ow1).
      apply trans_clean_oreachable_trans_clean with (a := b).
      trivial. trivial.
    unfold addr_clean. intros ox Sox.
    unfold trans_clean in TX.
    rewrite H in Sox.
    apply set_add_elim in Sox.
    inversion Sox as [Eox | Iox].
      inversion Eox. subst. apply oreachable_owner in Rx.
        apply ex_falso_quodlibet. apply NS2.
        rewrite addrs_from_owners_indexes.
        apply indexes_in_table_correct1 with (v := Some e').
        trivial.
      assert (AC: addr_clean h1 x).
        apply (TX x). constructor. 
        apply oreachable_owner with (h := h1)(a := b). trivial.
      apply AC. trivial.
      trivial. trivial. trivial. trivial. trivial.
Qed.


Lemma rednwo_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =nwo (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
 intros c1 c2 a1 a2 a3.
 destruct a1 as [e1 l1].
 destruct a2 as [e2 l2].
 destruct a3 as [e3 l3].
 simpl.
 intros W S1 S2 C R L.
 assert (CC : ac_clean (shp c1) (e1, l1) (ows c1)). trivial.
 induction R.
 subst.
 simpl.
 simpl in L.
 assert (A : l = l2).
   assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
   unfold wf_actors in WA. apply (WA e). trivial. trivial.
 subst.
 destruct (Aeq_aid e1 e).
 Case "e = e1".
   subst.
   assert (A: lookup_actor (up_actors e (up_locals i (vadd a) l2) (acs c1)) e =
                Some (e, up_locals i (vadd a) l2)).
     apply lookup_actor_intro.
     assert (WA : wf_actors (acs c1)). unfold wf_conf in W. tauto.
     unfold wf_actors. unfold up_actors. apply wf_up_table. trivial.
     unfold up_actors. apply up_table_correct2.
     apply indexes_in_table_correct1 with (v := l2). trivial.
   rewrite A in L.
   inversion L. subst. clear L. clear A.
   assert (A : l2 = l1).
     assert (WA : wf_actors (acs c1)). unfold wf_conf in W. tauto.
     unfold wf_actors in WA. unfold wf_table in WA.
     apply (WA e). trivial. trivial.
   subst. remember (up_locals i (vadd a) l1) as l2.
   unfold ac_clean. unfold loc_clean. simpl.
   intros j w Si.
   split.
   SCase "<> vmark".
     unfold ac_clean in C. unfold loc_clean in C. simpl in C.
     rewrite Heql2 in Si. unfold up_locals in Si.
     apply up_table_elim in Si.
     inversion Si as [Ei | Ij].
       inversion Ei. intros N. inversion N.
       apply C in Ij. tauto.
   SCase "trans_clean".
     intros b Vb.
     unfold ac_clean in C. unfold loc_clean in C. simpl in C.
     rewrite Heql2 in Si. unfold up_locals in Si.
     apply up_table_elim in Si. inversion Si as [Ei | Ij].
     SSCase "Ei".
       remember (set_add Aeq_sentry (a,o) (shp c1)) as h2.
       inversion Ei. subst. inversion H6. subst.
       apply trans_clean_new_object
       with (h1 := (shp c1))(o := o)(ow1 := ows c1)(a:= a).
         unfold wf_conf in W. unfold wfc_heap in W. tauto.
         intros K. apply H1.
           assert (A: subset (addrs_from_heap (shp c1))
                           (addrs_from_owners (ows c1))).
             unfold wf_conf in W. unfold consistent_heap_actors in W.
             tauto.
           apply A. trivial.
         trivial.
         apply subset_trans with (y := addrs_in_locals l1). 
           apply sub_values_sub_addrs_in_locals. trivial.
           apply subset_trans with (y := addrs_from_actors (acs c1)).
             unfold subset. intros x Sx.
             apply addrs_from_actors_correct2 with (e := e) (l := l1).
               trivial. trivial.
             unfold wf_conf in W. unfold consistent_heap_actors in W.
               tauto.
        trivial.
        trivial.
        rewrite ac_clean_all_values_clean in CC.
          intros u Su. apply H2 in Su. rewrite values_from_locals_values in Su.
          apply CC. trivial.
     SSCase "Ij".
       apply trans_clean_new_object_not_new 
         with (h1 := (shp c1))(a := a)(o := o)(ow1 := (ows c1)).
         unfold wf_conf in W. unfold wfc_heap in W. tauto.
         assert (A: subset (addrs_from_heap (shp c1))
                           (addrs_from_owners (ows c1))).
             unfold wf_conf in W. unfold consistent_heap_actors in W.
             tauto.
          intros Ca. apply A in Ca. tauto.
         trivial.
         apply subset_trans with (y := addrs_in_locals l1). 
           apply sub_values_sub_addrs_in_locals. trivial.
           apply subset_trans with (y := addrs_from_actors (acs c1)).
             unfold subset. intros x Sx.
             apply addrs_from_actors_correct2 with (e := e) (l := l1).
               trivial. trivial.
             unfold wf_conf in W. unfold consistent_heap_actors in W.
               tauto.
         trivial.
         trivial.
         rewrite ac_clean_all_values_clean in CC.
          intros u Su. apply H2 in Su. rewrite values_from_locals_values in Su.
          apply CC. trivial.
         apply C in Ij. inversion Ij as [NW TW]. apply TW. trivial.
         rewrite Vb in Ij.
           apply addrs_from_actors_correct2 with (a := b) in S1.
           assert (A : subset (addrs_from_actors (acs c1))
                              (addrs_from_heap (shp c1))).
              unfold wf_conf in W. unfold consistent_heap_actors in W. tauto.
           apply A in S1.
           assert (B: subset (addrs_from_heap (shp c1)) (addrs_from_owners (ows c1))).
             unfold wf_conf in W. unfold consistent_heap_owners in W. tauto.
           apply B in S1. 
           intros K. apply H1. subst. trivial.
           apply values_addrs_in_locals1.
           apply values_in_table_correct1 with (n := j). trivial.
 Case "e <> e1".
   unfold ac_clean. unfold loc_clean. simpl.
   intros j w Siw.
   apply lookup_actor_elim in L. inversion L. subst.
   rewrite ac_clean_all_values_clean in C.
   unfold value_clean in C.
   unfold up_actors in H5.
   apply up_table_elim in H5.
   inversion H5 as [NE | SE].
     SCase "NE".
       inversion NE. tauto.
     SCase "SE".
       assert (A: l3 = l1).
         assert (B : wf_actors (acs c1)). unfold wf_conf in W. tauto.
         unfold wf_actors in B. unfold wf_table in B. apply B with (n := e1).
         trivial. trivial.
       subst.
       apply values_in_table_correct1 with(Aeq_dec := Aeq_value) in Siw.
       apply C in Siw.
       split.
         tauto.
         intros x VW.
         inversion Siw as [Vw Tw].
         apply Tw in VW.
         destruct (in_table_dec addr (option aid) Aeq_addr (ows c1) x).
         SSCase "x has owner e1".
           inversion e0 as [oex Sex].
             destruct (oex). 
             SSSCase "oex = Some ex".
               rename a0 into ex.
               destruct (Aeq_aid ex e1).
               SSSSCase "ex = e1".
                 subst.
                 apply trans_clean_new_object_not_owner
                 with (h1 := shp c1)(e := e)(o := o)(ow1 := ows c1)(a := a).
                   unfold wf_conf in W. unfold wfc_heap in W. tauto.
                   trivial. trivial.
                   intros K. apply H1. 
                     assert (SU : subset (addrs_from_heap (shp c1)) 
                            (addrs_from_owners (ows c1))).
                      unfold wf_conf in W. unfold consistent_heap_owners.
                      tauto.
                     apply SU. trivial.
                   trivial.
                   auto.
                   trivial.
                   trivial.
               SSSSCase "ex <> e1".
                 destruct (Aeq_aid ex e).
                 SSSSSCase "ex = e".
                   subst. apply trans_clean_not_owned.
                   intros K. apply set_add_elim in K.
                   inversion K as [K1 | K2].
                     inversion K1. tauto.
                     assert (KW: wf_owners (ows c1)). unfold wf_conf in W. tauto.
                   unfold wf_owners in KW. 
                   unfold wf_table in KW.
                   assert (KE: Some e1 = Some e).
                     apply KW with (n := x). trivial. trivial.
                   inversion KE. tauto.
                 SSSSSCase "ex <> e".
                   apply trans_clean_not_owned.
                   intros K. apply set_add_elim in K.
                   inversion K as [KE | KI].
                     inversion KE. tauto.
                     assert (KW : wf_owners (ows c1)). unfold wf_conf in W. tauto.
                       assert (KS: Some ex = Some e1).
                         unfold wf_owners in KW. unfold wf_table in KW.
                         apply KW with (n := x). trivial. trivial.
                       inversion KS. tauto.
             SSSCase "oex = None".
               apply trans_clean_not_owned.
                 intros K. apply set_add_elim in K.
                 inversion K as [KE | KS].
                   inversion KE. tauto.
                   assert (KW : wf_owners (ows c1)). unfold wf_conf in W. tauto.
                   assert (KN : Some e1 = None).
                     unfold wf_owners in KW. unfold wf_table in KW.
                     apply KW with (n := x). trivial. trivial.
                   inversion KN.
         SSCase "x has no owner".
           apply trans_clean_not_owned.
           intros K. apply n0.
           apply set_add_elim in K.
           inversion K as [KE | KS].
             inversion KE. apply ex_falso_quodlibet. tauto.
             exists (Some e1). trivial.
Qed.
   
Lemma up_local_not_mark_keeps_clean :
  forall (h : heap)(e : aid) (l : locals)(i : vid)(v : value) (owns : owners),
  ac_clean h (e,l) owns ->
    value_clean h v e owns ->
      ac_clean h (e, up_locals i v l) owns.
Proof.
  intros h e l i v owns.
  intros AC VC.
  unfold ac_clean. simpl.
  unfold loc_clean.
  intros j w S.
  unfold ac_clean in AC.
  unfold loc_clean in AC.
  simpl in AC.
  unfold up_locals in S.
  apply up_table_elim in S.
  inversion S as [E | I].
  Case "E".
    inversion E. subst. unfold value_clean in VC. trivial.
  Case "I".
    apply AC with (i := j). trivial.
Qed.


Lemma rednwm_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =nwm (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
 intros c1 c2 a1 a2 a3.
 destruct a1 as [e1 l1].
 destruct a2 as [e2 l2].
 destruct a3 as [e3 l3].
 simpl.
 intros W S1 S2 C R L.
 assert (CC : ac_clean (shp c1) (e1, l1) (ows c1)). trivial.
 induction R.
 subst. simpl. simpl in L. 
 apply lookup_actor_elim in L.
 assert (A : l = l2).
   assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
   unfold wf_actors in WA. apply (WA e). trivial. trivial.
 subst.
 clear H.
 inversion L as [LE LS].
 subst.
 destruct (Aeq_aid e1 e).
 Case "e1 = e".
   subst.
   assert (A : l2 = l1).
     assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
     unfold wf_actors in WA. apply (WA e). trivial. trivial.
   subst.
   unfold up_actors in LS.
   apply up_table_elim in LS.
   inversion LS as [LSE | LSS].
   SCase "LSE".
     inversion LSE. apply up_local_not_mark_keeps_clean.
     trivial.
     unfold value_clean. split.
       intros K. inversion K.
       intros a V. inversion V.
   SCase "LSS".
     assert (A: l3 = l1).
       assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
       unfold wf_actors in WA. apply (WA e). trivial. trivial.
     subst. trivial.
 Case "e1 <> e".
   unfold up_actors in LS. apply up_table_elim in LS.
   inversion LS as [E | S].
     inversion E. apply ex_falso_quodlibet. tauto.
     assert (A : l3 = l1).
       assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
       unfold wf_actors in WA. apply (WA e1). trivial. trivial.
     subst.
     trivial.
Qed.

Lemma rednwa_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =nwa (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
 intros c1 c2 a1 a2 a3.
 destruct a1 as [e1 l1].
 destruct a2 as [e2 l2].
 destruct a3 as [e3 l3].
 simpl.
 intros W S1 S2 C R L.
 assert (CC : ac_clean (shp c1) (e1, l1) (ows c1)). trivial.
 induction R.
 subst. simpl. simpl in L. 
 apply lookup_actor_elim in L.
 assert (A : le = l2).
   assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
   unfold wf_actors in WA. apply (WA e). trivial. trivial.
 subst.
 inversion L as [LE LS].
 subst.
 clear L.
 assert (N1 : ne <> e).
   intros K. apply H1. subst. rewrite aids_from_actors_indexes.
   apply indexes_in_table_correct1 with (v := l2).
   trivial.
 clear H.
 assert (N2 : ne <> e1).
   intros K. apply H1. subst. rewrite aids_from_actors_indexes.
   apply indexes_in_table_correct1 with (v := l1).
   trivial.
 destruct (Aeq_aid e1 e).
 Case "e1 = e".
   subst.
   assert (A : l2 = l1).
     assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
     unfold wf_actors in WA. apply (WA e). trivial. trivial.
   subst.
   apply set_add_elim in LS.
   inversion LS as [LSE | LSS].
     inversion LSE. apply ex_falso_quodlibet. auto.
     unfold up_actors in LSS. apply up_table_elim in LSS.
       inversion LSS as [LSSE | LSSS].
     SCase "LSSE".
       inversion LSSE. apply up_local_not_mark_keeps_clean.
         trivial.
         unfold value_clean. split.
           intros K. inversion K.
           intros a V. inversion V.
     SCase "LSSS".
       assert (A : l3 = l1).
         assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
         unfold wf_actors in WA. apply (WA e). trivial. trivial.
       subst. trivial.
 Case "e1 <> e".
   apply set_add_elim in LS.
   inversion LS as [LSE | LSS].
     inversion LSE. apply ex_falso_quodlibet. auto.
     unfold up_actors in LSS. apply up_table_elim in LSS.
       inversion LSS as [LSSE | LSSS].
         inversion LSSE. apply ex_falso_quodlibet. auto.
         assert (A: l3 = l1).
           assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
           unfold wf_actors in WA. apply (WA e1). trivial. trivial.
         subst. trivial.
Qed.

Lemma set_owner_update_in :
  forall (adds : set addr) (o1 o2: option aid) (ow : owners) (x : addr),
  set_In (x, o2) (set_owner_update adds o1 ow) ->
    not (set_In x adds) -> 
      set_In (x, o2) ow.
Proof.
  intros adds.
  induction adds as [ | z addss]; intros o1 o2 ow x S N.
  Case "adds = nil".
    simpl in S. trivial.
  Case "adds = z::addss".
    simpl in S.
    apply IHaddss in S.
    unfold up_owners in S.
    apply up_table_elim in S.
    inversion S as [K | SS].
      inversion K. apply ex_falso_quodlibet. subst. apply N. simpl. tauto.
      trivial.
    intros K. apply N. simpl. tauto.
Qed.

Lemma set_owner_update_in_not :
  forall (adds : set addr) (o1 o2: option aid) (ow : owners) (x : addr),
  set_In (x, o2) ow ->
    not (set_In x adds) -> 
      set_In (x, o2) (set_owner_update adds o1 ow).
Proof.
  intros adds.
  induction adds as [ | z addss]; intros o1 o2 ow x S NS.
  Case "adds = nil".
    simpl. trivial.
  Case "adds = z::addss".
    simpl.
    apply IHaddss.
      unfold up_owners.
      apply up_table_correct8.
        intros K. subst. apply NS. simpl. tauto.
        trivial.
        intros K. apply NS. simpl. tauto.
Qed.

Lemma set_owner_update_correct1 : 
  forall (adds : set addr) (oe : option aid) (ow1 ow2 : owners)(x : addr),
    ow2 = set_owner_update adds oe ow1 ->
      set_In x adds ->
        set_In x (addrs_from_owners ow1) ->
          set_In (x, oe) ow2.
Proof.
  intros adds.
  induction adds as [ | z addss]; intros oe ow1 ow2 x SO S1 S2.
  Case "adds = nil".
    inversion S1.
  Case "adds = z::addss".
    simpl in S1. inversion S1 as [E | I].
    SCase "E".
      subst. clear S1. simpl.
      destruct (set_In_dec Aeq_addr x addss).
        apply IHaddss with (ow1 := up_owners x oe ow1).
          trivial. trivial.
          rewrite addrs_from_owners_indexes. unfold up_owners.
          apply indexes_in_up_table1. rewrite addrs_from_owners_indexes in S2.
          trivial.
        apply set_owner_update_in_not.
          unfold up_owners. apply up_table_correct2.
          rewrite addrs_from_owners_indexes in S2. trivial.
        trivial.
    SCase "I".
      simpl in SO.
      apply IHaddss with (ow1 := up_owners z oe ow1).
      trivial.
      trivial.
      rewrite addrs_from_owners_indexes. unfold up_owners.
        apply indexes_in_up_table11.
        rewrite addrs_from_owners_indexes in S2.
        trivial.
Qed.


Lemma trans_owner_update_correct1 :
  forall (adds : set addr)(h : heap)(ow1 ow2 : owners)(oe : option aid) (a x: addr),
   Some ow2 = trans_owner_update h ow1 oe a ->
    Some adds = taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) ->
     set_In x adds -> 
      set_In x (addrs_from_owners ow1) ->
       set_In (x, oe) ow2.
Proof.
  intros adds h ow1 ow2 oe a x TR TA S SA.
  unfold trans_owner_update in TR.
  rewrite <- TA in TR.
  inversion TR.
    apply set_owner_update_correct1 with (adds := adds)(ow1 := ow1).
      trivial. trivial. trivial.
Qed.


Lemma trans_owner_update_correct2 :
  forall (adds : set addr)(h : heap)(ow1 ow2 : owners)(oe : option aid) (a x: addr),
   Some ow2 = trans_owner_update h ow1 oe a ->
    Some adds = taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) ->
     not (set_In x adds) -> 
      set_In (x, oe) ow1 ->
        set_In (x, oe) ow2.
Proof.
  intros adds h ow1 ow2 oe a x S T NS I.
  unfold trans_owner_update in S.
    rewrite <- T in S.
    inversion S.
    apply set_owner_update_in_not.
      trivial. trivial.
Qed.

Lemma check_owner_correct1 :
  forall (adds : list addr)(owns : owners) (oe : option aid),
  check_owner adds owns oe  = true ->
    forall (a : addr), In a adds -> set_In (a,oe) owns.
Proof.
  intros adds.
  induction adds as [ | z addss]; intros owns oe C.
  Case "adds = nil".
    intros a I. inversion I.
  Case "adds = z::addss".
    intros a I. simpl in I.
    inversion I as [E | Ia].
    SCase "E".
      subst. clear I. simpl in C.
      remember (lookup_owner owns a) as oa.
      destruct oa.
      SSCase "Some o".
        destruct o.
        SSSCase "Some e".
          rename a0 into e.
          destruct oe.
          SSSSCase "Some a0".
            destruct (Aeq_aid a0 e). 
              subst. rewrite lookup_owner_tlookup in Heqoa.
                apply tlookup_correct1 with (Jeq_dec := Aeq_addr). auto.
              inversion C.
          SSSSCase "None".
            inversion C.
        SSSCase "None".
          destruct oe.
            inversion C.
            rewrite lookup_owner_tlookup in Heqoa.
              apply tlookup_correct1 with (Jeq_dec := Aeq_addr). auto.
      SSCase "None".
        inversion C.
    SCase "Ia".
      simpl in C.
      destruct (lookup_owner owns z).
        destruct o.
          destruct oe.
            destruct (Aeq_aid a1 a0).
              apply IHaddss. trivial. trivial.
              inversion C.
            inversion C.
            destruct oe.
              inversion C.
              apply IHaddss. trivial. trivial.
         inversion C.
Qed.
            

Lemma check_owner_correct2:
  forall (s : list addr) (owns : owners) (oe : option aid),
    wf_owners owns ->
     (forall (a : addr), In a s -> set_In (a, oe) owns) ->
      check_owner s owns oe = true.
Proof.
  intros s.
  induction s as [ | u ss]; intros owns oe W H.
  Case "s = nil".
    simpl. trivial.
  Case "s = u::ss".
    simpl.
    assert (A1: set_In (u, oe) owns). apply H. simpl. tauto.
    assert (A2: lookup_owner owns u = Some oe).
      rewrite lookup_owner_tlookup.
      apply tlookup_correct2. unfold wf_owners in W. trivial. trivial.
    rewrite A2.
    destruct oe.
    SCase "Some e".
      rename a into e.
      destruct (Aeq_aid e e). 
        apply IHss.
          trivial.
          intros x Ix. apply H. simpl. tauto.
        apply ex_falso_quodlibet. auto.
    SCase "None".
      apply IHss.
        trivial.
        intros x Ix. apply H. simpl. tauto.
Qed.


Lemma oreachable_owner_update_not_owner :
  forall (h : heap)(a b e: addr)(ac: aid)(oe : option aid) (ow1 ow2 : owners),
  wfc_heap h ->
    set_In a (addrs_from_heap h) ->
      oreachable h a b ac ow2 ->
        ow2 = up_owners e oe ow1 ->
          oe <> Some ac ->
          oreachable h a b ac ow1.
Proof.
  intros h a b e ac oe ow1 ow2.
  intros W S R.
  generalize dependent e.
  generalize dependent oe.
  generalize dependent ow1.
  induction R ; intros ow1 oe e U N.
  Case "refl".
    constructor. rewrite U in H.
    unfold up_owners in H.
    apply up_table_elim in H.
    inversion H as [E | I].
      inversion E. apply ex_falso_quodlibet. auto.
      trivial.
  Case "trans".
    apply or_trans with (b := b).
    rewrite U in H. unfold up_owners in H.
      apply up_table_elim in H.
      inversion H as [E | I].
        inversion E. apply ex_falso_quodlibet. auto.
        trivial.
    trivial.
    rewrite U in H1. unfold up_owners in H1.
      apply up_table_elim in H1.
      inversion H1 as [E | I].
        inversion E. apply ex_falso_quodlibet. auto.
        trivial.
    apply IHR with (oe := oe) (e := e).
      trivial.
      assert (CH : complete_heap h). unfold wfc_heap in W. tauto.
        unfold ofield in H0. inversion H0.
        inversion H0 as [oa [Soa [f If]]].
        unfold complete_heap in CH.
        apply CH with (a:= a) (o := oa). trivial.
          apply addrs_in_fields_correct2 with (f := f). trivial.
        trivial.
        trivial.
Qed.


Lemma oreachable_set_owner_update_not_owner :
  forall (h : heap) (s : set addr)(a b : addr)(ac : aid)(os : option aid)
         (ow1 ow2 : owners),
  wfc_heap h ->
    set_In a (addrs_from_heap h) ->
      oreachable h a b ac ow2 ->
        ow2 = set_owner_update s os ow1 ->
          os <> Some ac ->
            oreachable h a b ac ow1.
Proof.
  intros h s.
  induction s as [ | x ss]; intros a b ac os ow1 ow2 W S R U N.
  Case "s = nil".
    simpl in U. subst. trivial.
  Case "s = x::ss".
    simpl in U.
    apply IHss with (a := a)(b:= b)(ac := ac) in U .
      apply oreachable_owner_update_not_owner 
      with (e := x)(oe := os)(ow2 := up_owners x os ow1).
        trivial. trivial. trivial. trivial. trivial.
      trivial. trivial. trivial. trivial.
Qed.

Lemma trans_owner_check_true_taddr_some :
  forall (h : heap)(ow : owners)(a : addr)(oa : option aid),
   trans_owner_check h ow oa a = true ->
     exists (s : set addr), taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) = Some s.
Proof.
  intros h ow a oa T.
  unfold trans_owner_check in T.
  remember (taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a])) as ta.
  destruct ta.
  Case "Some s".
    exists s. trivial.
  Case "None".
    inversion T.
Qed.


Lemma ofield_in_address:
  forall (h : heap)(a b : addr),
  complete_heap h ->
    ofield h a b ->
      set_In a (addrs_from_heap h) ->
        set_In b (addrs_from_heap h).
Proof.
  intros h a b CH F S.
  unfold ofield in F.
  inversion F as [o [So [f If]]].
  unfold complete_heap in CH.
  apply CH with (a := a)(o := o).
    trivial.
    apply addrs_in_fields_correct2 with (f := f). trivial.
Qed.


Lemma oreachable_trans_owner_update_not_owner :
  forall (h : heap) (a b e: addr)(ac : aid)(oe : option aid)
         (ow1 ow2 : owners),
  wfc_heap h ->
    set_In a (addrs_from_heap h) ->
      oreachable h a b ac ow2 ->
        Some ow2 = trans_owner_update h ow1 oe e ->
         oe <> Some ac ->
           oreachable h a b ac ow1.
Proof.
  intros h a b e ac oe ow1 ow2.
  intros W S R T N.
  unfold trans_owner_update in T.
  destruct (taddr h [e] (set_diff Aeq_addr (addrs_from_heap h) [e])).
   apply oreachable_set_owner_update_not_owner 
    with (s := s)(os := oe)(ow2 := ow2).
        trivial. trivial. trivial. inversion T. trivial. trivial.
   inversion T. 
Qed.
  

Lemma redsnd_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =snd (fst a2) => c2 ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
 intros c1 c2 a1 a2 a3.
 destruct a1 as [e1 l1].
 destruct a2 as [e2 l2].
 destruct a3 as [e3 l3].
 simpl.
 intros W S1 S2 C R L.
 assert (CC : ac_clean (shp c1) (e1, l1) (ows c1)). trivial.
 induction R.
 subst. simpl. simpl in L. 
 apply lookup_actor_elim in L.
 inversion L as [LE LS].
 subst.
 assert (A: l3 = l1).
   assert (WA: wf_actors (acs c1)). unfold wf_conf in W. tauto.
   unfold wf_actors in WA. apply (WA e1). trivial. trivial.
 subst.
 clear LS. clear L.
 unfold ac_clean.
 unfold loc_clean.
 simpl.
 intros i v Si.
 unfold ac_clean in C. unfold loc_clean in C. simpl in C.
 split.
   apply C in Si. tauto.
   intros x Vx. subst.
   unfold trans_clean.
   intros z Rz.
   apply oreachable_trans_owner_update_not_owner 
   with (e := a)(oe := None) (ow1 := ows c1) in Rz.
     unfold trans_clean in C. apply C in Si.
       inversion Si as [V AC]. apply AC with (a := x). trivial.
       trivial.
     unfold wf_conf in W. tauto.
     unfold wf_conf in W.
       assert (A : consistent_heap_actors (shp c1)(acs c1)). tauto.
       unfold consistent_heap_actors in A.
         apply addrs_from_actors_correct2 with (a := x) in S1 .
         apply A. trivial. apply values_addrs_in_locals1.
         apply values_in_table_correct1 with( n := i).
         trivial.
       auto.
       intros K. inversion K.
Qed.

Lemma redrcv_keeps_clean : forall (c1 c2 : conf) (a1 a2 a3: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
    set_In a2 (acs c1) -> 
      ac_clean (shp c1) a1 (ows c1) ->
        c1 =rcv (fst a2) => c2 ->
         (fst a2) <> (fst a1) ->
          Some a3 = lookup_actor (acs c2) (fst a1) ->
            ac_clean (shp c2) a3 (ows c2).
Proof.
 intros c1 c2 a1 a2 a3.
 destruct a1 as [e1 l1].
 destruct a2 as [e2 l2].
 destruct a3 as [e3 l3].
 simpl.
 intros W S1 S2 C R N L.
 assert (CC : ac_clean (shp c1) (e1, l1) (ows c1)). trivial.
 induction R.
 rewrite H2 in H4.
 rewrite H8 in L.
 rewrite H8.
 simpl. simpl in L. clear H8. 
 apply lookup_actor_elim in L.
 inversion L as [LE LS].
 unfold up_actors in LS.
 apply up_table_elim in LS.
 rewrite LE in LS. rewrite LE. clear L. clear LE.
 inversion LS as [LSE | LSS].
 Case "LSE".
   inversion LSE. apply ex_falso_quodlibet. auto.
 Case "LSS".
   assert (A : l3 = l1).
     assert (B: wf_actors (acs c1)). unfold wf_conf in W. tauto.
     unfold wf_actors in B. unfold wf_table in B.
     apply B with (n:= e1). trivial. trivial.
   rewrite A. clear A. clear LSS. clear LS.
   unfold ac_clean. unfold loc_clean. simpl.
   intros j w Sj.
   unfold ac_clean in CC. unfold loc_clean in CC.
   assert (SSj : set_In (j,w) l1). trivial.
   apply CC in SSj.
   split. tauto.
     intros x Vx. rewrite Vx in Sj.
       unfold trans_clean.
       intros z Rz.
       apply oreachable_trans_owner_update_not_owner 
       with (e := snd m)(oe := Some e)(ow1 := ows c1)in Rz.
         inversion SSj as [Vw Fw]. apply Fw in Vx.
         unfold trans_clean in Vx.
         apply Vx. simpl. trivial.
       unfold wf_conf in W. tauto.
       assert (A : consistent_heap_actors (shp c1)(acs c1)). unfold wf_conf in W. tauto.
         unfold consistent_heap_actors in A.
         apply addrs_from_actors_correct2 with (a := x) in S1 .
         apply A. trivial. apply values_addrs_in_locals1.
         apply values_in_table_correct1 with (n := j).
         trivial.
       trivial.
       intros K. inversion K. auto.
Qed.


Lemma mred_preserves_actors :
  forall (c1 c2 : conf)(e : aid),
  c1 =@ e => c2 ->
    subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2)).
Proof.
  intros c1 c2 e R.
  induction R.
  Case "silent".
    inversion H. subst. simpl.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes.
    apply indexes_in_up_table1.
  Case "fread".
    inversion H. subst. simpl.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes.
    apply indexes_in_up_table1.
  Case "fwrite".
    inversion H. subst. simpl.
    rewrite aids_from_actors_indexes.
    apply subset_refl.
  Case "newact".
    inversion H. subst. simpl.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes.
    apply subset_trans with
    (y := indexes_in_table aid locals Aeq_aid 
          (up_actors e (up_locals i (vaid ne) le) (acs c1))).
    apply indexes_in_up_table1.
    apply indexes_in_add_table1.
  Case "newmbox".
    inversion H. subst. simpl.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes.
    apply indexes_in_up_table1.   
  Case "newobj".
    inversion H. subst. simpl.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes.
    apply indexes_in_up_table1.   
  Case "send".
    inversion H. subst. simpl.
    apply subset_refl.   
  Case "rcv".
    inversion H0. subst. simpl.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes.
    apply indexes_in_up_table1.
Qed.   

Lemma reds_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =s e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.

Lemma redfr_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =fr e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.

Lemma redfw_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =fw e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.


Lemma rednwa_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =nwa e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists le. trivial.
Qed.

Lemma rednwm_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =nwm e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.

Lemma rednwo_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =nwo e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.

Lemma redsnd_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =snd e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.

Lemma redrcv_has_label_actor :
  forall (c1 c2 : conf) (e : aid),
    c1 =rcv e => c2 ->
      exists (le: locals), set_In (e,le) (acs c1).
Proof.
  intros c1 c2 e R.
  induction R. exists l. trivial.
Qed.


Theorem ac_isolation : forall (c1 c2 : conf) (a1 a2: actor),
  wf_conf c1 ->
   set_In a1 (acs c1) ->
     ac_clean (shp c1) a1 (ows c1) -> 
      c1 =@ (fst a1) =>* c2 ->
       Some a2 = lookup_actor (acs c2) (fst a1) -> 
        ac_clean (shp c2) a2 (ows c2).
Proof.
  intros c1 c2 a1 a2.
  intros W S1 C R L.
  destruct a1 as [e1 l1].
  destruct a2 as [e2 l2].
  simpl in R.
  simpl in L.
  generalize dependent l1.
  generalize dependent e2.
  generalize dependent l2.
  induction R; intros l2 e2 L l1 S1 C.
  Case "refl".
    apply lookup_actor_elim in L. inversion L as [E S2].
    assert (A: wf_actors (acs c)). unfold wf_conf in W. tauto.
    unfold wf_actors in A. unfold wf_table in A.
    subst. 
    assert (B: l2 = l1). apply A with ( n := a). trivial. trivial.
    subst. trivial.
  Case "trans".
    induction H.
    SCase "silent".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_silent with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply reds_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply reds_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply reds_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
    SCase "fread".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_fread with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply redfr_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply redfr_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply redfr_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
    SCase "fwrite".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_fwrite with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply redfw_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply redfw_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply redfw_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
    SCase "newact".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_newact with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply rednwa_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply rednwa_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply rednwa_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
    SCase "newmbox".
       rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_newmbox with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply rednwm_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply rednwm_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply rednwm_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
   SCase "newobj".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_newobj with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply rednwo_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply rednwo_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply rednwo_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
    SCase "send".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := e).
        apply mred_send with (e := e). trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply redsnd_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply redsnd_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply redsnd_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
    SCase "rcv".
      rename e2 into e3.
      rename l2 into l3.
      assert (LL : Some (e3, l3) = lookup_actor (acs c3) a). trivial.
      apply lookup_actor_elim in LL. inversion LL as [E S3].
      subst. clear LL.
      assert (A1 : subset (aids_from_actors (acs c1)) (aids_from_actors (acs c2))).
        apply mred_preserves_actors with (e := a).
        apply mred_rcv with (e := e). trivial. trivial.
      assert (A2 : set_In a (aids_from_actors (acs c2))).
        apply A1. rewrite aids_from_actors_indexes.
        apply indexes_in_table_correct1 with (v := l1).
        trivial.
      assert (B : exists (l2 : locals), set_In (a, l2) (acs c2)).
        apply indexes_in_table_correct2 with (Jeq_dec := Aeq_aid).
        rewrite aids_from_actors_indexes in A2. trivial.
      inversion B as [l2 S2].
      assert (E : exists (le : locals), set_In (e,le) (acs c1)).
        apply redrcv_has_label_actor with (c2 := c2).
        trivial.
      inversion E as [le Se].
      assert (W2 : wf_conf c2).
        apply redrcv_preserves_wf with (c1 := c1)(a := e). trivial. trivial.
      apply IHR with (l1 := l2).
        trivial.
        trivial.
        trivial.
        apply redrcv_keeps_clean with (c1 := c1)(a1:= (a,l1)) (a2 := (e,le)).
          trivial.
          trivial.
          trivial.
          trivial.
          simpl. trivial.
          simpl. trivial.
          simpl. symmetry. apply lookup_actor_intro. unfold wf_conf in W2. tauto.
          trivial.
Qed.

Section Liveness.

Definition a: addr := mkaddr 1.
Definition f : fid := mkfid 1.
Definition oa : object := [(f, vmark)].
Definition h: heap := [(a,oa)].
Definition mi : msgid := mkmsgid 1.
Definition m : message := (mi, a).
Definition e1 : aid := mkaid 1.
Definition e2 : aid := mkaid 2.
Definition mb1 : mbox := mkmb e2 [].
Definition mb2 : mbox := mkmb e2 [m].
Definition i : vid := mkvid 1.
Definition j : vid := mkvid 2.
Definition l1 : locals := [(i, vadd a)].
Definition l2 : locals := [(j,vnull)].
Definition l3 : locals := [(j,vadd a)].
Definition ac1 : actor := (e1, l1).
Definition ac2 : actor := (e2,l2).
Definition ac3 : actor := (e2, l3).
Definition ow1 : owners := [(a, Some e1)].
Definition ow2 : owners := [(a, None)].
Definition ow3 : owners := [(a, Some e2)].
Definition acts1 : actors := [ac1, ac2].
Definition acts2 : actors := [ac1, ac3].
Definition mbi : mid := mkmid 1.
Definition mbs1 : mailboxes := [(mbi, mb1)].
Definition mbs2 : mailboxes := [(mbi, mb2)].
Definition cf1 : conf := mkcf acts1 ow1 mbs1 h.
Definition cf2 : conf := mkcf acts1 ow2 mbs2 h.
Definition cf3 : conf := mkcf acts2 ow3 mbs1 h.

Lemma marked_oa : obj_mark oa.
Proof.
  unfold obj_mark. unfold oa. exists f. simpl. tauto.
Qed.

Lemma clean_ac2 : ac_clean h ac2 ow1.
Proof.
  unfold ac_clean. simpl. unfold loc_clean.
  intros k u H.
  unfold l2 in H. simpl in H.
  inversion H as [E | F].
    inversion E. split. intros K. inversion K. intros x V. inversion V.
    tauto.
Qed.

Lemma marked_ac3: ac_mark h ac3 ow3.
Proof.
  unfold ac_mark. simpl. unfold loc_mark.
  unfold l3. right. exists j. exists a.
  split.
    simpl. tauto.
    unfold trans_mark. exists a.
      split.
        constructor. unfold ow3. simpl. tauto.
        unfold addr_mark. exists oa. unfold h.
        split. simpl. tauto.
          apply marked_oa.
Qed.

Lemma Aeq_addr_true:
  forall (A: Type) (a: addr) (x y: A), (if Aeq_addr a a then x else y) = x.
Proof.
  intros A a x y. destruct (Aeq_addr a a). trivial. apply ex_falso_quodlibet. apply n. trivial.
Qed.

Lemma Aeq_addr_false:
  forall (A: Type) (a b : addr) (x y : A), a <> b -> (if Aeq_addr a b then x else y) = y.
Proof.
  intros A a b x y. case (Aeq_addr a b); congruence.
Qed.


Lemma cf1_redsnd_cf2: cf1 =snd e1 => cf2.
Proof.
  unfold cf1. unfold cf2.
  apply redsnd_step with
  (a := a)(l := l1)(ms := mi)(mi := mbi)(mb := mb1)(mb' := mb2)(owns := ow2).
    simpl. tauto.
    unfold l1. simpl. tauto.
    simpl. unfold trans_owner_check.
      unfold h. simpl. rewrite Aeq_addr_true.
      rewrite taddr_eq_taddrn with (n := 2). simpl.
      rewrite Aeq_addr_true. unfold oa. simpl.
      rewrite Aeq_addr_true. destruct (Aeq_aid e1 e1); congruence.
      simpl. omega.
    simpl. tauto.
    unfold mb1. simpl. tauto.
    simpl. unfold h. unfold ow1. unfold trans_owner_update.
      unfold trans_owner_check. simpl. rewrite Aeq_addr_true.
      rewrite taddr_eq_taddrn with (n := 2). simpl.
      rewrite Aeq_addr_true. unfold oa. simpl.
      rewrite Aeq_addr_true. unfold ow2. trivial.
      simpl. omega.
    simpl. unfold mb2. unfold m. trivial.
    simpl. destruct (Aeq_mid mbi mbi).
      unfold mbs2. trivial.
      apply ex_falso_quodlibet. tauto.
Qed.

Lemma cf2_redrcv_cf3: cf2 =rcv e2 => cf3.
Proof.
  unfold cf2. unfold cf3.
  apply redrcv_step with (m := m)(l := l2)(l' := l3)(mb := mb2)(mb' := mb1)(mi := mbi)
  (owns := ow3)(i := j)(v := vnull).
    simpl. unfold ac2. tauto.
    simpl. tauto.
    unfold l2. simpl. tauto.
    unfold mb2. simpl. trivial.
    unfold mb2. simpl. trivial.
    simpl. unfold mb1. trivial.
    simpl. unfold trans_owner_update.
      rewrite taddr_eq_taddrn with (n := 2). simpl.
      rewrite Aeq_addr_true. unfold oa. simpl.
      rewrite Aeq_addr_true. unfold ow3. trivial.
      unfold h. simpl. rewrite Aeq_addr_true. simpl. omega.
      constructor.
      unfold l2. simpl. destruct (Aeq_vid j j).
        unfold l3. trivial.
        apply ex_falso_quodlibet. tauto.
      simpl.
        assert (e2 <> e1). unfold e2. unfold e1. intros K. inversion K.
        destruct (Aeq_aid e2 e1).
          apply ex_falso_quodlibet. tauto.
          destruct (Aeq_aid e2 e2).
            unfold acts2. unfold ac1. unfold ac3. unfold mbs1.
            destruct (Aeq_mid mbi mbi). trivial. apply ex_falso_quodlibet. tauto.
          tauto. 
Qed.


Lemma cf1_redt_cf3 : cf1 ==>* cf3.
Proof.
  apply redt_trans with (c2 := cf2).
  apply red_send with (e := e1). apply cf1_redsnd_cf2.
  apply redt_trans with (c2 := cf3).
  apply red_rcv with (e := e2). apply cf2_redrcv_cf3.
  apply redt_refl.
Qed.


Theorem live_mark : exists (c1 c2 : conf)(ac1 ac2 : actor),
  c1 ==>* c2 /\ set_In ac1 (acs c1) /\ ac_clean (shp c1) ac1 (ows c1)
    /\ Some ac2 = lookup_actor (acs c2) (fst ac1) /\ ac_mark (shp c2) ac2 (ows c2).
Proof.
  exists cf1. exists cf3. exists ac2. exists ac3.
  split.
    apply cf1_redt_cf3.
    split.
      unfold cf1. simpl. tauto.
      split.
        unfold cf1. simpl. apply clean_ac2.
        split.
          unfold cf3. unfold ac2. unfold acts2. simpl.
            destruct (Aeq_aid e2 e1).
              apply ex_falso_quodlibet. unfold e2 in e. unfold e1 in e. inversion e.
              destruct (Aeq_aid e2 e2). unfold ac3. trivial. tauto.
          unfold cf3. simpl. apply marked_ac3.
Qed.



End Liveness.
    
  









