(*
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: August 2013
  (c) INRIA
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import Recdef.
Require Import SiaamTable.
Require Import SiaamStructures.
Require Import SiaamUtils.
Require Import SiaamReach.
Require Import SiaamRules.




(* 
  A well-formed configuration is a configuration where:
   - the heap is well-formed, i.e. an object reference (addr) references
     a unique object in the heap, and is complete, 
     i.e. object reference fields of an object
     that appears in the heap also have an entry in the heap.
   - aids appearing in mailboxes are in actors.
   - all objects appearing in the heap have an owner.
   - all objects references in mailbox queues point to objects in the heap.
   - all object references appearing in mailbox queues, and object references
     of objects reachable from these, have a None owner.
*)

        
Definition consistent_mboxes_actors (mbs : mailboxes) (acts : actors) : Prop :=
  subset (aids_from_mboxes mbs) (aids_from_actors acts).

Definition consistent_heap_owners (h : heap) (owns : owners) : Prop :=
  subset (addrs_from_heap h) (addrs_from_owners owns).

Definition consistent_mboxes_heap (mbs : mailboxes) (h : heap) : Prop :=
  subset (addrs_from_mboxes mbs) (addrs_from_heap h).

Definition consistent_heap_actors (h : heap) (acs : actors) : Prop :=
  subset (addrs_from_actors acs) (addrs_from_heap h).


Definition consistent_mboxes_heap_owners 
                (mbs : mailboxes)(h : heap)(owns : owners) : Prop :=
  forall (a b : addr), 
      set_In a (addrs_from_mboxes mbs) ->
             reachable h a b ->
                    set_In (b,None) owns.

Definition wf_conf (c : conf) : Prop :=
  wfc_heap (shp c) 
  /\ wf_owners (ows c)
  /\ wf_mboxes (mbs c)
  /\ wf_actors (acs c)
  /\ consistent_heap_owners (shp c) (ows c)
  /\ consistent_mboxes_actors (mbs c) (acs c)
  /\ consistent_mboxes_heap (mbs c) (shp c)
  /\ consistent_heap_actors (shp c) (acs c).


Definition wfc_conf (c : conf): Prop := 
  wf_conf c /\ consistent_mboxes_heap_owners (mbs c) (shp c) (ows c).




Lemma addrs_from_actors_up : forall (acts : actors) (a: aid) (l1 l2 : locals),
  wf_actors acts ->
    action l1 l2 ->
    set_In (a,l1) acts ->
      subset (addrs_from_actors (up_actors a l2 acts)) (addrs_from_actors acts).
Proof.
  intros acts.
  induction acts as [ | (e,L) actss]; intros a l1 l2 W A S.
  Case "acts = nil".
    simpl. unfold subset. trivial.
  Case "acts = (e,L)::actss".
    simpl in S. inversion S as [EL | Sl1].
    SCase "L = l1".
      inversion EL.
      subst.
      simpl.
      destruct (Aeq_aid a a).
      SSCase "a =a".
        simpl.
        apply union_included_union.
        SSSCase "branche 1".
          action_cases (induction A) SSSSCase.
          SSSSCase "a_nwbool".
            subst.
            apply addrs_in_locals_nw.
               intros C. inversion C as [c HC]. inversion HC.
               trivial.
          SSSSCase "a_nwnat".
            subst.
            apply addrs_in_locals_nw.
               intros C. inversion C as [c HC]. inversion HC.
               trivial.
          SSSSCase "a_nwnull".
            subst.
            apply addrs_in_locals_nw.
               intros C. inversion C as [c HC]. inversion HC.
               trivial.
          SSSSCase "a_compat".
            subst. unfold addrs_in_actor. simpl.
            apply addrs_in_locals_up4.
            apply values_in_table_correct1 with (n := j). trivial.
          SSSSCase "a_forget".
            unfold addrs_in_actor. simpl.
            apply addrs_in_locals_sub. trivial.
        SSSCase "branche 2".
          destruct (set_In_dec Aeq_actor (a,l1) actss).
            apply IHactss with (l1 := l1).
              unfold wf_actors. unfold wf_actors in W. 
                apply wf_table_cons with (e:= (a,l1)). trivial.
               trivial.
               trivial.
            assert (N: ~ set_In a (indexes_in_table aid locals Aeq_aid actss)).
              intros C. apply n. apply indexes_in_table_correct2 in C.
              inversion C as [l3 Cl3].
              assert (EL3: l3 = l1).
                unfold wf_actors in W. unfold wf_table in W. apply W with (n:= a).
                simpl. right. trivial.
                simpl. left. trivial.
              subst. trivial.
            assert (EU : up_actors a l2 actss = actss).
              unfold up_actors.
              apply up_table_correct3.
              trivial.
            rewrite EU. 
            apply subset_refl.
      SSCase "a <> a".
        tauto.
    SCase "(a,l1) in actss".
      simpl.
      destruct (Aeq_aid a e).
      SSCase "a = e".
        subst. simpl.
        apply union_included_union.
        SSSCase "branche 1".
          assert (EL : L = l1).
            unfold wf_actors in W. unfold wf_table in W.
            apply W with (n := e).
            simpl. left. trivial.
            simpl. tauto.
          subst.
          action_cases (induction A) SSSSCase.
          SSSSCase "a_nwbool".
            subst.
            apply addrs_in_locals_nw.
               intros C. inversion C as [c HC]. inversion HC.
               trivial.
          SSSSCase "a_nwnat".
            subst.
            apply addrs_in_locals_nw.
               intros C. inversion C as [c HC]. inversion HC.
               trivial.
          SSSSCase "a_nwnull".
            subst.
            apply addrs_in_locals_nw.
               intros C. inversion C as [c HC]. inversion HC.
               trivial.
          SSSSCase "a_compat".
            subst. unfold addrs_in_actor. simpl.
            apply addrs_in_locals_up4.
            apply values_in_table_correct1 with (n := j). trivial.
          SSSSCase "a_forget".
            unfold addrs_in_actor. simpl.
            apply addrs_in_locals_sub. trivial.
        SSSCase "branche 2".
          apply IHactss with (l1 := l1).
            unfold wf_actors. unfold wf_actors in W. 
            apply wf_table_cons with (e:= (e,L)).
              trivial.
            trivial.
            trivial.
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss with (l1 := l1).
            unfold wf_actors. unfold wf_actors in W. 
            apply wf_table_cons with (e:= (e,L)).
              trivial.
            trivial.
            trivial.
Qed.
      

Lemma reds_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =s a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf.
  intros c1 c2 a W R.
  induction R.
  subst. simpl.
  split. tauto.
  split. tauto.
  split. tauto.
  split.
    unfold wf_actors. unfold up_actors.
    unfold wf_actors in W.
    apply wf_up_table. tauto.
  split. tauto.
  split.
    unfold consistent_mboxes_actors.
    unfold consistent_mboxes_actors in W.
    rewrite aids_from_actors_indexes.
    rewrite aids_from_actors_indexes in W.
    apply subset_trans with
      (y := indexes_in_table aid locals Aeq_aid (acs c1)).
    tauto.
    apply indexes_in_up_table1.
  split. tauto.
    unfold consistent_heap_actors.
    unfold consistent_heap_actors in W.
    apply subset_trans with (y := (addrs_from_actors (acs c1))).
      apply addrs_from_actors_up with (l1 := l).
        tauto. trivial. trivial.
      tauto.
Qed.
    


Lemma redfr_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =fr a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf. intros c1 c2 a W R.
  induction R.
  subst.
  simpl.
  split. tauto.
  split. tauto.
  split. tauto.
  split.
    unfold wf_actors.
    unfold wf_actors in W.
    apply wf_up_table. 
    tauto.
  split. tauto.
  split.
    unfold consistent_mboxes_actors.
    unfold consistent_mboxes_actors in W.
    rewrite aids_from_actors_indexes.
    apply subset_trans with (y := aids_from_actors (acs c1)).
      tauto.
      rewrite aids_from_actors_indexes. apply indexes_in_up_table1.
  split. tauto.
    unfold consistent_heap_actors.
    unfold consistent_heap_actors in W.
    value_cases (destruct v) Case.
    Case "vnull".
      apply subset_trans with (y := addrs_from_actors (acs c1)).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [c Cc]. inversion Cc.
      tauto.
    Case "vmark".
      apply subset_trans with (y := addrs_from_actors (acs c1)).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [c Cc]. inversion Cc.
      tauto.
    Case "vnat".
      apply subset_trans with (y := addrs_from_actors (acs c1)).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [c Cc]. inversion Cc.
      tauto.
   Case "vbool".
      apply subset_trans with (y := addrs_from_actors (acs c1)).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [c Cc]. inversion Cc.
      tauto.
    Case "vadd".
      apply subset_trans with 
        (y := set_add Aeq_addr a0 (addrs_from_actors (acs c1))).
        apply addrs_from_actors_up2.
         tauto.
         trivial.
        assert (A: set_In a0 (addrs_from_heap (shp c1))).
         unfold wfc_heap in W.
         unfold complete_heap in W.
         inversion W as [WHC [WFO [WFM [WFA [CHO [CMA [CMH CHA]]]]]]].
         inversion WHC as [WFH WCH].
         apply WCH with (b:= a0) in H2.
          trivial.
          apply addrs_in_fields_correct2 with (f := f).
          trivial.
         apply add_included_elim.
          trivial.
         tauto.
    Case "vaid".
      apply subset_trans with (y := addrs_from_actors (acs c1)).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [c Cc]. inversion Cc.
      tauto.
    Case "vmid".
      apply subset_trans with (y := addrs_from_actors (acs c1)).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [c Cc]. inversion Cc.
      tauto.
Qed. 

Lemma up_preserves_wfc_heap : forall (h : heap)(o : object)
                              (a :addr)(f : fid) (v w: value),
  set_In (a,o) h ->
    set_In (f, v) o ->
      wfc_heap h -> 
        (forall (e: addr), w = vadd e -> set_In e (addrs_from_heap h)) ->
          wfc_heap (up_heap a (up_object f w o) h).
Proof.
  unfold wfc_heap. 
  intros h o a f v w Ha Hf W V.
  inversion W as [WF CH].
  split.
  Case "WF".
    rewrite wf_heap_table.
    rewrite wf_heap_table in WF.
    apply wf_up_table.
    trivial.
  Case "CH".
    unfold complete_heap.
    unfold complete_heap in CH.
    intros s b o' H1 H2.
    assert (SU : subset (addrs_from_heap h)
                        (addrs_from_heap (up_heap a (up_object f w o) h))).
       rewrite addrs_from_heap_indexes.
       rewrite addrs_from_heap_indexes.
       apply indexes_in_up_table1.
    destruct (Aeq_addr s a).
    SCase "s = a".
      subst.
      assert (Eo : o' = up_object f w o).
        apply wf_up_table  with
           (Jeq_dec := Aeq_addr)
           (n := a)
           (a := (up_object f w o)) in WF.
        unfold wf_table in WF.
        apply WF with (n:= a).
          trivial.
          apply up_table_correct2.
          apply indexes_in_table_correct1 with (v := o).
          trivial.
      subst.
      apply addrs_in_fields_correct1 in H2.
      inversion H2 as [f0 Hb].
      destruct (Aeq_fid f0 f).
      SSCase "f = f0".
        subst.
        unfold up_object in Hb.
        apply up_table_correct6 in Hb.
        symmetry in Hb.
        apply V in Hb.
        apply SU.
        trivial.
      SSCase "f <> f0".
        unfold up_object in Hb.
        apply up_table_correct4 in Hb.
        apply addrs_in_fields_correct2 in Hb.
        apply CH with (b := b) in Ha.
        apply SU.
          trivial.
          trivial.
          auto.
    SCase "s <> a".
      assert (N: a <> s).
        intros C. apply n. auto.
      apply up_table_correct4 with
            (A := object)
            (Jeq_dec := Aeq_addr)
            (t := h)
            (v := (up_object f w o))
            (w := o') in N.
      apply SU.
      apply CH with (a := s)(o := o').
        trivial.
        trivial.
        trivial.
Qed.
    
  

Lemma redfw_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =fw a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf. intros c1 c2 a W R.
  induction R.
  subst.
  simpl.
  split.
    apply up_preserves_wfc_heap with (v := v).
    trivial.
    trivial.
    tauto.
    intros x X.
    unfold consistent_heap_actors in W.
    subst.
    inversion W as [WFC [WFO [ WFM [WFA [CHO [ CMA [CMH CAH]]]]]]].
    apply CAH.
    apply addrs_from_actors_correct2 with (e:= e)(l:= l).
      trivial.
      apply values_addrs_in_locals1. 
       apply values_in_table_correct1 with (n:= i). trivial.
  split. tauto.
  split. tauto.
  split. tauto.
  split. 
    unfold consistent_heap_owners.
    unfold consistent_heap_owners in W.
    apply subset_trans with
      (y := (addrs_from_heap (shp c1))).
      rewrite addrs_from_heap_indexes.
        rewrite addrs_from_heap_indexes.
        apply indexes_in_up_table2.
      tauto.
  split. tauto.
  split.
    unfold consistent_mboxes_heap.
      unfold consistent_mboxes_heap in W.
      apply subset_trans with
        (y := (addrs_from_heap (shp c1))).
        tauto.
        rewrite addrs_from_heap_indexes. 
          rewrite addrs_from_heap_indexes. apply indexes_in_up_table1.
   unfold consistent_heap_actors.
     unfold consistent_heap_actors in W.
      apply subset_trans with
        (y := (addrs_from_heap (shp c1))).
        tauto.
        rewrite addrs_from_heap_indexes. 
          rewrite addrs_from_heap_indexes. apply indexes_in_up_table1.
Qed.     

Lemma addrs_from_actors_add1 : forall (acts : actors)(a : actor),
  subset (addrs_from_actors (set_add Aeq_actor a acts))
         (set_union Aeq_addr (addrs_in_actor a) (addrs_from_actors acts)).
Proof.
  intros acts.
  induction acts as [ | e actss]; intros a.
  Case "acts = nil".
    simpl. apply subset_refl.
  Case "acts = e::actss".
    simpl. destruct (Aeq_actor).
    SCase "a = e".
      simpl. subst.
      apply union_included_union.
        apply subset_refl.
        apply union_included2.
    SCase "a <> e".
      simpl.
      apply subset_trans with
       (y := (set_union Aeq_addr 
                (set_union Aeq_addr (addrs_in_actor a)
                                    (addrs_in_actor e))
                (addrs_from_actors actss))).
       apply subset_trans with
        (y := (set_union Aeq_addr 
                (set_union Aeq_addr (addrs_in_actor e)
                                    (addrs_in_actor a))
                (addrs_from_actors actss))).
        apply subset_trans with
         (y := (set_union Aeq_addr (addrs_in_actor e)
                (set_union Aeq_addr 
                    (addrs_in_actor a)
                    (addrs_from_actors actss)))).
         apply union_included_union.
           apply subset_refl.
           apply IHactss.
           apply sub_union_assoc1.
           apply union_included_union.
             apply sub_union_com.
             apply subset_refl.
           apply sub_union_assoc2.
Qed.

Lemma addrs_from_actors_add2 : forall (acts : actors)(a : actor),
  subset (set_union Aeq_addr (addrs_in_actor a) (addrs_from_actors acts))
         (addrs_from_actors (set_add Aeq_actor a acts)).
Proof.
  intros acts.
  induction acts as [ | e actss]; intros a.
  Case "acts = nil".
    simpl. apply subset_refl.
  Case "acts = e::actss".
    simpl.
    destruct (Aeq_actor a e).
    SCase "a = e".
      subst. simpl.
      apply subset_trans with
        (y := set_union Aeq_addr
               (set_union Aeq_addr (addrs_in_actor e) (addrs_in_actor e))
               (addrs_from_actors actss)).
        apply sub_union_assoc1.
        apply union_included_union.
          unfold subset. intros x H. apply set_union_elim in H. tauto.
          apply subset_refl.
    SCase "a <> e".
      simpl.
        apply subset_trans with
            (y := set_union Aeq_addr (addrs_in_actor e)
                    (set_union Aeq_addr (addrs_in_actor a)
                    (addrs_from_actors actss))).
          apply subset_trans with
              (y := set_union Aeq_addr
                     (set_union Aeq_addr (addrs_in_actor e) (addrs_in_actor a))
                     (addrs_from_actors actss)).
            apply subset_trans with 
                (y := set_union Aeq_addr
                       (set_union Aeq_addr (addrs_in_actor a) (addrs_in_actor e))
                       (addrs_from_actors actss)).
              apply sub_union_assoc1.
              apply union_included_union.
                apply sub_union_com.
                apply subset_refl.
            apply sub_union_assoc2.
          apply union_included_union.
            apply subset_refl.
            apply IHactss.
Qed.
          
        
Lemma rednwa_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =nwa a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf. 
  intros c1 c2 a W R.
  induction R.
  subst. simpl.
  split. tauto.
  split. tauto.
  split. tauto.
  split.
    unfold wf_actors.
    apply not_in_wf_table1 with (Jeq_dec := Aeq_aid).
      unfold wf_actors in W.
       apply wf_up_table.
       tauto.
      intros C. apply H1. rewrite aids_from_actors_indexes.
        apply indexes_in_up_table22 with 
          (i := e)
          (v := (up_locals i (vaid ne) le)).
        trivial.
  split. tauto.
  split.
    unfold consistent_mboxes_actors.
    unfold consistent_mboxes_actors in W.
    apply subset_trans with 
      (y := (aids_from_actors (acs c1))).
      tauto.
      rewrite aids_from_actors_indexes. rewrite aids_from_actors_indexes.
      assert (A1: subset (indexes_in_table aid locals Aeq_aid
                           (up_actors e (up_locals i (vaid ne) le) (acs c1)))
                        (indexes_in_table aid locals Aeq_aid
                           (set_add (Aeq_tentry aid locals Aeq_aid Aeq_locals)
                           (ne, up_locals k (vmid m) lne)
                           (up_actors e (up_locals i (vaid ne) le) (acs c1))))).
        apply indexes_in_add_table1.
      assert (A2 : subset  (indexes_in_table aid locals Aeq_aid (acs c1))
                           (indexes_in_table aid locals Aeq_aid
                           (up_actors e (up_locals i (vaid ne) le) (acs c1)))).
        apply indexes_in_up_table1.
      apply subset_trans with 
        (y := (indexes_in_table aid locals Aeq_aid
                           (up_actors e (up_locals i (vaid ne) le) (acs c1)))).
      trivial.
      apply indexes_in_add_table1.
  split. tauto.
    unfold consistent_heap_actors.
    unfold consistent_heap_actors in W.
    remember ((ne, up_locals k (vmid m) lne)) as na.
    remember (up_actors e (up_locals i (vaid ne) le) (acs c1)) as uacs.
    apply subset_trans with
      (y := (addrs_from_actors (acs c1))).
      apply subset_trans with
      (y := set_union Aeq_addr (addrs_in_actor na)
                               (addrs_from_actors uacs)).
        apply addrs_from_actors_add1.
        apply subset_trans with
          (y := set_union Aeq_addr (addrs_from_actors (acs c1))
                                   (addrs_from_actors (acs c1))).
          apply union_included_union.
            subst. unfold addrs_in_actor. simpl.
              apply subset_trans with
              (y := addrs_in_locals lne).
                apply addrs_in_locals_up1.
                  intros C. inversion C as [u Cu]. inversion Cu.
                  apply subset_trans with (y := (addrs_in_locals le)).
                    apply addrs_from_locals_values. trivial.
                    apply addrs_from_actors_correct3 with (e:= e). trivial.
                rewrite Hequacs. apply addrs_from_actors_up1.
                  tauto. trivial.
                  intros C. inversion C as [u Cu]. inversion Cu.
       apply union_included1.
         apply subset_refl. apply subset_refl.
    tauto.
Qed.
                  
                    
Lemma aids_from_mboxes_add : forall (m : mailboxes) (mb :mbentry),
  subset (aids_from_mboxes (set_add Aeq_mbentry mb m))
         (set_add Aeq_aid (own (snd mb)) (aids_from_mboxes m)).
Proof.
  intros m.
  induction m as [ | mb1 ms]; intros mb.
  Case "m = nil".
    simpl. apply subset_refl.  
  Case "m = mb1 ::ms".
    simpl.
    destruct (Aeq_mbentry mb mb1).
    SCase "mb = mb1".
      subst. simpl. apply add_included. apply add_included_sub. apply subset_refl.
    SCase "mb <> mb1".
      simpl.
      apply subset_trans with
      (y := set_add Aeq_aid (own (snd mb1)) 
                            (set_add Aeq_aid (own (snd mb)) (aids_from_mboxes ms))).
        apply add_included. apply IHms.
        apply add_add_com. apply subset_refl.
Qed.

Lemma addrs_from_mboxes_add : forall (m : mailboxes) (mb : mbentry),
  subset (addrs_from_mboxes (set_add Aeq_mbentry mb m))
         (set_union Aeq_addr (addrs_from_mbox (snd mb)) (addrs_from_mboxes m)).
Proof.
  intros m. 
  induction m as [ | mb1 ms]; intros mb.
  Case "m = nil".
    simpl. apply subset_refl.
  Case "m = mb1 ::ms".
    simpl.
    destruct (Aeq_mbentry mb mb1).
    SCase" mb = mb1".
      subst. simpl. apply union_included_union.
        apply subset_refl.
        apply union_included2.
    SCase "mb <> mb1".
      simpl.
      apply subset_trans with
      (y := set_union Aeq_addr (addrs_from_mbox (snd mb1))
             (set_union Aeq_addr (addrs_from_mbox (snd mb)) (addrs_from_mboxes ms))).
       apply union_included_union.
         apply subset_refl.
         apply IHms.
       apply subset_trans with
       (y := set_union Aeq_addr (set_union Aeq_addr (addrs_from_mbox (snd mb))
                                                    (addrs_from_mbox (snd mb1)))
                                (addrs_from_mboxes ms)).
         apply subset_trans with
         (y := set_union Aeq_addr (set_union Aeq_addr (addrs_from_mbox (snd mb1))
                                                    (addrs_from_mbox (snd mb)))
                                (addrs_from_mboxes ms)).
           apply sub_union_assoc1.
           apply union_included_union.
             apply sub_union_com.
             apply subset_refl.
      apply sub_union_assoc2.
Qed.
           

Lemma rednwm_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =nwm a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf. intros C1 C2 a W R.
  induction R.
  subst.
  simpl.
  split. tauto.
  split. tauto.
  split.
    apply wf_add_table with (Jeq_dec := Aeq_mid).
      unfold wf_mboxes in W. tauto.
      simpl. rewrite mids_from_mboxes_indexes in H1. trivial.
  split. 
    unfold up_actors. apply wf_up_table. tauto.
  split. tauto.
  split.
    unfold consistent_mboxes_actors.
    apply subset_trans with
      (y := aids_from_actors (acs c1)).
      apply subset_trans with
        (y := set_add Aeq_aid e (aids_from_mboxes (mbs c1))).
        apply aids_from_mboxes_add. apply add_included_elim. 
          rewrite aids_from_actors_indexes. 
            apply indexes_in_table_correct1 with (v := l).
            trivial.
          tauto.
        rewrite aids_from_actors_indexes.
          rewrite aids_from_actors_indexes. apply indexes_in_up_table1.
  split.
    unfold consistent_mboxes_heap.
      apply subset_trans with
      (y := addrs_from_mboxes (mbs c1)).
        apply subset_trans with
        (y := (set_union Aeq_addr (addrs_from_mbox (snd (m, {| own := e; msgs := [] |})))
                                (addrs_from_mboxes (mbs c1)))).
          apply addrs_from_mboxes_add.
          simpl. unfold addrs_from_mbox. simpl. apply union_included1.
            unfold subset. intros a C. inversion C.
            apply subset_refl.
       tauto.
    unfold consistent_heap_actors.
    apply subset_trans with (y := (addrs_from_actors (acs c1))).
      apply addrs_from_actors_up1.
        tauto.
        trivial.
        intros C. inversion C as [u Cu]. inversion Cu.
      tauto.
Qed.
        
  
Lemma complete_heap_add : forall (h :heap)(a: addr) (o: object),
  complete_heap h ->
    ~ set_In a (addrs_from_heap h) ->
      subset (addrs_in_fields o) (addrs_from_heap h) ->
        complete_heap (set_add Aeq_sentry (a,o) h).
Proof.
  intros h.
  induction h as [ | (e, oe) hs]; intros a o CH N S.
  Case "h = nil".
    simpl.
    simpl in S.
    unfold complete_heap.
    intros c b oc Sc Sb.
    simpl in Sc.
    inversion Sc.
      inversion H. subst. simpl. right. unfold subset in S.
        apply S in Sb. inversion Sb.
      tauto.
  Case "h = (e,oe)::hs".
    simpl in S. simpl.
    destruct (Aeq_sentry (a,o) (e,oe)).
    SCase "(a,o) = (e,oe)".
      inversion e0. subst. trivial.
    SCase "(a,o) <> (e,oe)".
      unfold complete_heap.
      intros c b oc Sc Sb.
      simpl. apply set_add_intro.
      simpl in Sc.
      inversion Sc as [Eoc | Soc].
        inversion Eoc. subst. unfold complete_heap in CH.
          apply CH  with (a := c) in Sb. simpl in Sb. 
          apply set_add_elim in Sb.
          inversion Sb.
            left. trivial.
            right. 
              assert (A: subset (addrs_from_heap hs) 
                                (addrs_from_heap (set_add Aeq_sentry (a,o) hs))).
               rewrite addrs_from_heap_indexes.
               rewrite addrs_from_heap_indexes.
               apply indexes_in_add_table1.
              unfold subset in A.
              apply A in H.
              trivial.
        simpl. tauto.
      apply set_add_elim in Soc.
        inversion Soc as [E | H].
          inversion E. subst. unfold subset in S. apply S in Sb.
            apply set_add_elim in Sb. inversion Sb.
              tauto.
              right.
                assert (A: subset (addrs_from_heap hs)
                                  (addrs_from_heap (set_add Aeq_sentry (a,o) hs))).
                 rewrite addrs_from_heap_indexes.
                 rewrite addrs_from_heap_indexes.
                 apply indexes_in_add_table1.
                unfold subset in A.
                apply A in H.
                trivial.
          unfold complete_heap in CH. apply CH with (a := c) in Sb.
            simpl in Sb. apply set_add_elim in Sb.
            inversion Sb.
              tauto.
              right.
                assert (A: subset (addrs_from_heap hs)
                                  (addrs_from_heap (set_add Aeq_sentry (a,o) hs))).
                 rewrite addrs_from_heap_indexes.
                 rewrite addrs_from_heap_indexes.
                 apply indexes_in_add_table1.
                unfold subset in A.
                apply A in H0.
                trivial.
         simpl. tauto.
Qed.
                

Lemma rednwo_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =nwo a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf. intros C1 C2 a W R.
  induction R.
  subst.
  simpl.
  split.
    unfold wfc_heap. unfold wfc_heap in W.
    split.
      rewrite wf_heap_table.
        apply wf_add_table with (Jeq_dec := Aeq_addr).
          tauto.
          simpl. 
            intros C.
            apply H1.
            assert (A: subset (addrs_from_heap (shp c1))
                              (addrs_from_owners (ows c1))).
              tauto.
            unfold subset in A.
            apply A.
            rewrite addrs_from_heap_indexes. 
            trivial.
      apply complete_heap_add.
        tauto.
        assert (A: subset (addrs_from_heap (shp c1))
                              (addrs_from_owners (ows c1))).
          tauto.
         intros C. 
         apply H1. 
         unfold subset in A.
         apply A.
         rewrite addrs_from_heap_indexes. 
         trivial.
        rewrite addrs_from_heap_indexes in C.
        trivial.
        apply subset_trans with (y := (addrs_in_locals l)).
          apply sub_values_sub_addrs_in_locals. trivial.
          unfold consistent_heap_actors in W. 
            apply subset_trans with (y := addrs_from_actors (acs c1)).
            unfold subset. intros x S. 
            apply addrs_from_actors_correct2 with (e:= e)(l := l).
              trivial. trivial. tauto.
  split.
    unfold wf_owners.
    apply not_in_wf_table1 with (Jeq_dec := Aeq_addr).
    tauto.
    rewrite addrs_from_owners_indexes in H1.
      trivial.
  split.
    tauto.
  split.
    unfold wf_actors. 
    apply wf_up_table.
    tauto.
  split.
    unfold consistent_heap_owners.
    apply subset_trans with (y := set_add Aeq_addr a (addrs_from_heap (shp c1))).
      rewrite addrs_from_heap_indexes. rewrite addrs_from_heap_indexes.
        apply indexes_in_add_table2.
      apply subset_trans with (y := set_add Aeq_addr a (addrs_from_owners (ows c1))).
        apply add_included. tauto.
        rewrite addrs_from_owners_indexes.
          rewrite addrs_from_owners_indexes.
          apply indexes_in_add_table3.
  split.
    unfold consistent_mboxes_actors.
    apply subset_trans with (y := aids_from_actors (acs c1)).
      tauto.
      rewrite aids_from_actors_indexes.
      rewrite aids_from_actors_indexes.
      apply indexes_in_up_table1.
  split.
    unfold consistent_mboxes_heap.
      apply subset_trans with (y := set_add Aeq_addr a (addrs_from_heap (shp c1))).
        apply add_included_sub. tauto.
        rewrite addrs_from_heap_indexes.
          rewrite addrs_from_heap_indexes.
          apply indexes_in_add_table3.
    unfold consistent_heap_actors.
      apply subset_trans with (y := set_add Aeq_addr a (addrs_from_actors (acs c1))).
        apply addrs_from_actors_up2.
          tauto.
          trivial.
        apply subset_trans with (y := set_add Aeq_addr a (addrs_from_heap (shp c1))).
          apply add_included. tauto.
          rewrite addrs_from_heap_indexes. rewrite addrs_from_heap_indexes.
            apply indexes_in_add_table3.
Qed.
      

Lemma set_owner_update_wf :
  forall (l : set addr) (o : option aid)(ows owns: owners),
  wf_owners ows ->
    set_owner_update l o ows = owns -> 
      wf_owners owns.
Proof.
  intros l.
  induction l as [ | a ls]; intros o ows owns W H.
  Case "l = nil".
    simpl in H. subst. trivial.
  Case "l = a::ls".
    simpl in H.
    apply IHls with (o := o) (ows := up_owners a o ows).
      unfold wf_owners. apply wf_up_table. trivial.
      trivial.
Qed.

Lemma trans_owner_update_wf: 
  forall (h : heap)(o : option aid) (a: addr) (ows owns: owners),
  wf_owners ows ->
    trans_owner_update h ows o a = Some owns ->
      wf_owners owns.
Proof.
  intros h o a ows owns W T.
  unfold trans_owner_update in T.
    destruct (taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a])) as [adds | ].
    Case "Some adds".
      inversion T. apply set_owner_update_wf with (l := adds)(o := o)(ows := ows).
      trivial. trivial.
    Case "none".
      inversion T.
Qed.
   
Lemma set_owner_update_addrs :
  forall (l : set addr) (o : option aid)(ows: owners),
    subset (addrs_from_owners (set_owner_update l o ows))
           (addrs_from_owners ows)
    /\
    subset (addrs_from_owners ows)
           (addrs_from_owners (set_owner_update l o ows)).
Proof.
  intros l.
  induction l as [ | e ls]; intros o ows.
  Case "l = nil".
    simpl. split.
      apply subset_refl.
      apply subset_refl.
  Case "l = a :: ls".
    simpl.
    split.
      apply subset_trans with
      (y := addrs_from_owners (up_owners e o ows)).
        apply IHls. 
        rewrite addrs_from_owners_indexes.
          rewrite addrs_from_owners_indexes.
          apply indexes_in_up_table2.
      apply subset_trans with
      (y := addrs_from_owners (up_owners e o ows)).
        rewrite addrs_from_owners_indexes.
          rewrite addrs_from_owners_indexes.
          apply indexes_in_up_table1.
        apply IHls.
Qed.

Lemma trans_owner_update_addrs :
  forall (h : heap)(o : option aid) (a : addr) (ows owns : owners),
    trans_owner_update h ows o a = Some owns ->
      (subset (addrs_from_owners owns) (addrs_from_owners ows)
       /\ subset (addrs_from_owners ows) (addrs_from_owners owns)).
Proof.
  intros h o a ows owns T.
  unfold trans_owner_update in T.
  destruct (taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a])) as [adds | ].
  Case "Some adds".
      inversion T. apply set_owner_update_addrs. 
  Case "None".
      inversion T.
Qed.


Lemma trans_owner_update_addrs1:
  forall (h : heap)(o : option aid) (a : addr) (ows owns : owners),
    trans_owner_update h ows o a = Some owns ->
      subset (addrs_from_owners owns) (addrs_from_owners ows).
Proof.
  intros h o a ows owns T.
  unfold trans_owner_update in T.
  destruct (taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a])) as [adds | ].
  Case "Some adds".
      inversion T. apply set_owner_update_addrs. 
  Case "None".
      inversion T.
Qed.

Lemma trans_owner_update_addrs2:
  forall (h : heap)(o : option aid) (a : addr) (ows owns : owners),
    trans_owner_update h ows o a = Some owns ->
      subset (addrs_from_owners ows) (addrs_from_owners owns).
Proof.
  intros h o a ows owns T.
  unfold trans_owner_update in T.
  destruct (taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a])) as [adds | ].
  Case "Some adds".
      inversion T. apply set_owner_update_addrs. 
  Case "None".
      inversion T.
Qed.

Lemma aids_from_mboxes_up1 :
  forall (m : mailboxes)(mi : mid)(mb : mbox),
  subset (aids_from_mboxes (up_mboxes mi mb m))
         (set_add Aeq_aid (own mb) (aids_from_mboxes m)).
Proof.
  intros m.
  induction m as [ | (i,mbi) ms]; intros mi mb.
  Case "m = nil".
    simpl. unfold subset.
    intros a H. inversion H.
  Case "m = (i,mbi)::ms".
    simpl.
    destruct (Aeq_mid mi i).
    SCase "mi = i".
      simpl. 
      apply subset_trans with
      (y := set_add Aeq_aid (own mbi) (set_add Aeq_aid (own mb) (aids_from_mboxes ms))).
        apply add_included_sub. apply add_included_elim. 
          apply set_add_intro. left. trivial.
        apply IHms.
          apply add_add_com. apply subset_refl.
    SCase "mi <> i".
      simpl.
      apply subset_trans with
      (y := set_add Aeq_aid (own mbi) (set_add Aeq_aid (own mb) (aids_from_mboxes ms))).
        apply add_included. apply IHms.
        apply add_add_com. apply subset_refl.
Qed.


Lemma aids_from_mboxes_correct1: 
  forall (m :mailboxes)(mi :mid) (mb : mbox),
  set_In (mi, mb) m ->
    set_In (own mb) (aids_from_mboxes m).
Proof.
  intros m.
  induction m as [ | (i,mbi) ms]; intros mi mb H.
  Case "m = nil".
    inversion H.
  Case "m = (i,mbi)::ms".
    simpl. simpl in H.
    inversion H as [E | S].
      inversion E. subst. apply set_add_intro. left. trivial.
      apply set_add_intro. right. apply IHms with (mi := mi). trivial.
Qed.

Lemma addrs_from_mboxes_up : forall (m : mailboxes) (mi : mid) (mb : mbox),
  subset (addrs_from_mboxes (up_mboxes mi mb m))
         (set_union Aeq_addr (addrs_from_mbox mb) (addrs_from_mboxes m)).
Proof.
  intros m.
  induction m as [ | (i,mbi) ms]; intros mi mb.
  Case "m = nil".
    simpl. unfold subset. intros a H. inversion H.
  Case "m = (i,mbi)::ms".
    simpl.
    destruct (Aeq_mid mi i).
    SCase "mi = i".
      simpl. 
      apply subset_trans with
      (y := set_union Aeq_addr (addrs_from_mbox mbi)
              (set_union Aeq_addr (addrs_from_mbox mb) (addrs_from_mboxes ms))).
        apply union_included1.
        apply subset_trans with
        (y := set_union Aeq_addr (addrs_from_mbox mb)
                     (set_union Aeq_addr (addrs_from_mbox mbi) (addrs_from_mboxes ms))).
           apply union_included.
           apply subset_trans with
           (y := set_union Aeq_addr 
                     (set_union Aeq_addr (addrs_from_mbox mb) (addrs_from_mbox mbi))
                     (addrs_from_mboxes ms)).
             apply sub_union_assoc1.
             apply subset_trans with
             (y := set_union Aeq_addr 
                     (set_union Aeq_addr (addrs_from_mbox mbi) (addrs_from_mbox mb))
                     (addrs_from_mboxes ms)).
               apply union_included_union.
                 apply sub_union_com.
                 apply subset_refl.
               apply sub_union_assoc2.
            apply subset_trans with
            (y := set_union Aeq_addr (addrs_from_mbox mbi)
                                 (addrs_from_mboxes (up_mboxes mi mb ms))).
               apply union_included2.
               apply union_included_union.
                 apply subset_refl.
                 apply IHms.
        apply subset_trans with
        (y := set_union Aeq_addr 
                 (set_union Aeq_addr (addrs_from_mbox mbi)(addrs_from_mbox mb))
                 (addrs_from_mboxes ms)).
           apply sub_union_assoc1.
           apply subset_trans with
           (y := set_union Aeq_addr 
                 (set_union Aeq_addr (addrs_from_mbox mb)(addrs_from_mbox mbi))
                 (addrs_from_mboxes ms)).
             apply union_included_union.
               apply sub_union_com.
               apply subset_refl.
           apply sub_union_assoc2.
    SCase "mi <> i".
      simpl.
      apply subset_trans with
      (y := set_union Aeq_addr (addrs_from_mbox mbi)
              (set_union Aeq_addr (addrs_from_mbox mb) (addrs_from_mboxes ms))).
        apply union_included_union.
          apply subset_refl.
          apply IHms.
      apply subset_trans with
      (y := set_union Aeq_addr 
                 (set_union Aeq_addr (addrs_from_mbox mbi)(addrs_from_mbox mb))
                 (addrs_from_mboxes ms)).
        apply sub_union_assoc1.
          apply subset_trans with
          (y := set_union Aeq_addr 
               (set_union Aeq_addr (addrs_from_mbox mb)(addrs_from_mbox mbi))
               (addrs_from_mboxes ms)).
             apply union_included_union.
               apply sub_union_com.
               apply subset_refl.
             apply sub_union_assoc2.
Qed.
               
Lemma addrs_from_mbox_cons1 : forall (mb : mbox)(i : msgid)(a : addr),
  subset (set_add Aeq_addr a (addrs_from_mbox mb))
         (addrs_from_mbox {| own := own mb ; msgs := (i,a):: msgs mb |}).
Proof.
  intros mb i a.
  unfold addrs_from_mbox at 2. simpl.
  unfold addrs_from_mbox.
  unfold subset.
  intros x H.
  apply set_add_elim in H.
  inversion H as [E | S].
    subst. simpl. left. trivial.
    simpl. right. trivial.
Qed.

Lemma addrs_from_mbox_cons2 : forall (mb : mbox)(i : msgid)(a : addr),
  subset (addrs_from_mbox {| own := own mb ; msgs := (i,a):: msgs mb |})
         (set_add Aeq_addr a (addrs_from_mbox mb)).
Proof.
  intros mb i a.
  unfold addrs_from_mbox.
  simpl.
  unfold subset.
  intros x H.
  apply set_add_intro.
  simpl in H.
  inversion H; auto.
Qed.

Lemma addrs_from_mboxes_correct1: forall (m : mailboxes)(i : mid) (mb : mbox),
  set_In (i,mb) m ->
    subset (addrs_from_mbox mb) (addrs_from_mboxes m).
Proof.
  intros m.
  induction m as [ | (j,mbj) ms]; intros i mb.
  Case "m = nil".
    intros H. inversion H.
  Case "m = (j,mbj) :: ms".
    intros H.
    simpl.
    simpl in H.
    inversion H as [E | S].
    SCase "E".
      inversion E. subst. apply union_included.
    SCase "S".
      apply subset_trans with
      (y := set_union Aeq_addr (addrs_from_mbox mbj) (addrs_from_mbox mb)).
        apply union_included2.
        apply union_included_union.
          apply subset_refl.
          apply IHms with (i := i). trivial.
Qed.
      

Lemma redsnd_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =snd a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf.
  intros c1 c2 a W R.
  induction R. subst. simpl.
  split. tauto.
  split. 
    apply trans_owner_update_wf with (h := (shp c1))(o := None)(a:= a)(ows := (ows c1)).
    tauto.
    auto.
  split.
    unfold wf_mboxes. apply wf_up_table. tauto.
  split. tauto.
  split.
    unfold consistent_heap_owners.
    apply subset_trans with (y := addrs_from_owners (ows c1)).
      tauto.
      apply trans_owner_update_addrs2 with (h := (shp c1))
                                           (o := None)
                                           (a := a).
        auto.
  split.
    unfold consistent_mboxes_actors.
      apply subset_trans with
      (y := set_add Aeq_aid (own {| own := own mb; msgs := (ms, a) :: msgs mb |}) 
                            (aids_from_mboxes (mbs c1))).
        apply aids_from_mboxes_up1.
        simpl. apply subset_trans with
          (y := (aids_from_mboxes (mbs c1))).  
          apply add_included_elim. apply aids_from_mboxes_correct1 with (mi := mi).
            trivial.
          apply subset_refl.
      tauto.
  split. 
     unfold consistent_mboxes_heap.
       apply subset_trans with
       (y := set_union Aeq_addr 
                    (addrs_from_mbox {| own := own mb; msgs := (ms, a) :: msgs mb |})
                    (addrs_from_mboxes (mbs c1))).
         apply addrs_from_mboxes_up.
         apply subset_trans with
         (y := set_add Aeq_addr a (addrs_from_mboxes (mbs c1))).
           apply union_included1.
             apply subset_trans with (y := set_add Aeq_addr a (addrs_from_mbox mb)).
               apply addrs_from_mbox_cons2.
               apply add_included. 
                 apply addrs_from_mboxes_correct1 with (i := mi). trivial.
               apply add_included_sub. apply subset_refl.
           apply add_included_elim.
             apply value_in_locals_addr in H0. apply addrs_from_actors_correct3 in H.
               unfold consistent_heap_actors in W.
               assert (A: subset (addrs_from_actors (acs c1)) (addrs_from_heap (shp c1))).
                 tauto.
               unfold subset in A.
               apply A.
               unfold subset in H.
               apply H.
               trivial.
        tauto.
    tauto.
Qed.


Definition addrs_from_queue (q : queue) : set addr :=
  fold_right (fun (m : message)(adds : set addr) => 
                  set_add Aeq_addr (snd m) adds)
    nil q.

Lemma addrs_from_queue_cons : forall (q : queue) (i : msgid) (a : addr),
  addrs_from_queue ((i,a)::q) = set_add Aeq_addr a (addrs_from_queue q).
Proof.
  intros q.
  induction q as [ | (j,b) qs]; intros i a.
  Case "q = nil".
    simpl. trivial.
  Case "q = (j,b)::qs".
    simpl. trivial.
Qed.


Lemma addrs_from_queue_remove : forall (q : queue),
  subset (addrs_from_queue (removelast q)) (addrs_from_queue q).
Proof.
  intros q.
  induction q as [ | (i,a) qs].
  Case "q = nil".
    simpl. apply subset_refl.
  Case "q = (i,a)::qs".
    simpl.
    destruct qs as [ | (j,b) qss].
      simpl. unfold subset. intros x H. inversion H.
      rewrite addrs_from_queue_cons.
      apply add_included.
      apply IHqs.
Qed.

Lemma addrs_from_mbox_from_queue1 : forall (mb : mbox),
  subset (addrs_from_mbox mb) (addrs_from_queue (msgs mb)).
Proof.
  intros mb.
  destruct mb as [o q].
  simpl.
  induction q as [ | (i,a) qs].
  Case "q = nil".
    simpl. unfold addrs_from_mbox. simpl. apply subset_refl.
  Case "q = (i,a)::qs".
    unfold addrs_from_mbox. simpl.
    unfold addrs_from_mbox in IHqs. simpl in IHqs.
    unfold subset.
    intros x H.
    apply set_add_intro.
    simpl in H.
    inversion H as [E | S].
      left. auto.
      right. apply IHqs. trivial.
Qed.

Lemma addrs_from_mbox_from_queue2 : forall (mb : mbox),
  subset (addrs_from_queue (msgs mb)) (addrs_from_mbox mb).
Proof.
  intros mb.
  destruct mb as [o q].
  simpl.
  unfold addrs_from_mbox. simpl.
  induction q as [ | (i,a) qs].
  Case "q = nil".
    simpl. apply subset_refl.
  Case "q = (i,a)::qs".
    simpl.
    unfold subset.
    intros x H.
    simpl.
    apply set_add_elim in H.
    inversion H as [E | S].
      left. auto.
      right. apply IHqs. trivial.
Qed.

Lemma addrs_from_mbox_remove : forall (mb : mbox),
  subset (addrs_from_mbox {| own := own mb ; msgs := removelast (msgs mb) |})
         (addrs_from_mbox mb).
Proof.
  intros mb.
  apply subset_trans with 
  (y := addrs_from_queue (msgs {| own := own mb; msgs := removelast (msgs mb) |})).
    apply addrs_from_mbox_from_queue1.
    simpl.
      apply subset_trans with (y := addrs_from_queue (msgs mb)).
        apply addrs_from_queue_remove.
        apply addrs_from_mbox_from_queue2.
Qed.


Lemma addr_lasto_in_queue : forall (q : queue)(m : message),
  lasto q = Some m -> set_In (snd m) (addrs_from_queue q).
Proof.
  intros q.
  induction q as [ | mg qs]; intros m H.
  Case "q = nil".
    simpl in H. inversion H.
  Case "q = mg::qs".
    simpl in H.
    simpl.
    apply set_add_intro.
    destruct qs as [ | mh qss].
      left. inversion H. trivial.
      right. apply IHqs. trivial.
Qed.


Lemma redrcv_preserves_wf : forall (c1 c2 : conf)(a: aid),
  wf_conf c1 -> c1 =rcv a => c2 -> wf_conf c2.
Proof.
  unfold wf_conf. intros c1 c2 a W R.
  induction R.
  subst.
  simpl.
  split. tauto.
  split.
    apply trans_owner_update_wf with 
           (h := shp c1)(o := Some (own mb))(a := snd m)(ows := ows c1).
      tauto.
      auto.
  split.
    unfold wf_mboxes. apply wf_up_table. tauto.
  split.
    unfold wf_actors. apply wf_up_table. tauto.
  split.
    unfold consistent_heap_owners.
    apply subset_trans with (y := addrs_from_owners (ows c1)).
      tauto.
      apply trans_owner_update_addrs2 with
                 (h := shp c1)(o := Some (own mb))(a := snd m).
       auto.
  split.
    unfold consistent_mboxes_actors.
    apply subset_trans with
    (y := aids_from_actors (acs c1)).
      apply subset_trans with 
      (y := set_add Aeq_aid (own {| own := own mb; msgs := removelast (msgs mb) |}) 
                            (aids_from_mboxes (mbs c1))).
        apply aids_from_mboxes_up1.
        simpl.
          apply subset_trans with (y := aids_from_mboxes (mbs c1)).
            apply add_included_elim.
              apply aids_from_mboxes_correct1 with (mi := mi). trivial.
              apply subset_refl.
          tauto.
     rewrite aids_from_actors_indexes.
       rewrite aids_from_actors_indexes. apply indexes_in_up_table1.
  split.
    unfold consistent_mboxes_heap.
    apply subset_trans with
    (y := set_union Aeq_addr 
          (addrs_from_mbox {| own := own mb; msgs := removelast (msgs mb) |}) 
          (addrs_from_mboxes (mbs c1))).
      apply addrs_from_mboxes_up.
      apply subset_trans with (y := set_union Aeq_addr (addrs_from_mbox mb)
                                                       (addrs_from_mboxes (mbs c1))).
        apply union_included_union.
          apply addrs_from_mbox_remove.
          apply subset_refl.
        apply subset_trans with (y := addrs_from_mboxes (mbs c1)).
          apply union_included1.
            apply addrs_from_mboxes_correct1 with (i := mi).
              trivial.
              apply subset_refl.
          tauto.
    unfold consistent_heap_actors.
    apply subset_trans with
      (y := set_add Aeq_addr (snd m) (addrs_from_actors (acs c1))).
      apply addrs_from_actors_up2.
        tauto.
        trivial.
      apply add_included_elim.
        unfold consistent_mboxes_heap in W.
        assert (A : subset (addrs_from_mboxes (mbs c1)) (addrs_from_heap (shp c1))).
          tauto.
        apply A.
        assert (B : subset (addrs_from_mbox mb) (addrs_from_mboxes (mbs c1))).
          apply addrs_from_mboxes_correct1 with (i := mi). trivial.
        apply B.
        assert (C : subset (addrs_from_queue (msgs mb)) (addrs_from_mbox mb)).
          apply addrs_from_mbox_from_queue2.
        apply C.
          apply addr_lasto_in_queue. auto.
      tauto.
Qed.
          

Theorem red_preserves_wf : forall (c1 c2 : conf),
  c1 ==> c2 -> wf_conf c1 -> wf_conf c2.
Proof.
  intros c1 c2 R.
  red_cases (induction R) Case ; intros W.
  Case "red_silent".
    apply reds_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_fread".
    apply redfr_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_fwrite".
    apply redfw_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_newact".
    apply rednwa_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_newmbox".
    apply rednwm_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_newobj".
    apply rednwo_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_send".
    apply redsnd_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
  Case "red_rcv".
    apply redrcv_preserves_wf with (c1 := c1) (a := e). trivial. trivial.
Qed.








