(*
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: June 2013
  (c) INRIA
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import Recdef.
Require Import SiaamStructures.
Require Import SiaamUtils.
Require Import SiaamTable.


(*
  Definitions used to verify the correctness 
  of the defined functions and algorithms.
*)

Definition ofield (h : heap) (s a : addr) : Prop :=
  (exists (o : object), set_In (s,o) h /\ (exists (f : fid), In (f, vadd a) o)).

Inductive reachable : heap -> addr -> addr -> Prop :=
  | r_self : forall (h : heap) (a : addr), 
             In a (addrs_from_heap h) -> reachable h a a
  | r_field : forall (h : heap) (a b c: addr), 
             reachable h a b ->
                  ofield h b c ->
                     reachable h a c.

Definition s_reachable (h : heap) (ass : set addr) (a : addr): Prop :=
  exists (b : addr),  In b ass /\ reachable h b a.


Lemma reachable_trans : forall (h : heap) (a b c: addr),
  reachable h a b -> reachable h b c -> reachable h a c.
Proof.
  intros h a b c Rab Rbc. generalize dependent a.
  induction Rbc.
  Case "r_self".
    intros b Rba. trivial.
  Case "r-field".
    intros d Rda. apply r_field with (b:= b).
    apply IHRbc. trivial.
    trivial.
Qed.


Inductive ireachable : heap -> addr -> addr -> Prop :=
  | i_self : forall (h : heap) (a : addr),
                 In a (addrs_from_heap h) -> ireachable h a a
  | i_field : forall (h : heap) (a b c : addr),
                 In a (addrs_from_heap h) -> 
                  ofield h a b ->
                   ireachable h b c ->
                     ireachable h a c.

Lemma ireachable_reachable : forall (h : heap) (a b : addr),
  ireachable h a b -> reachable h a b.
Proof.
  intros h a b Iab.
  induction Iab.
  Case "i_self".
    constructor. trivial.
  Case "i_field".
    assert (Rab: reachable h a b).
      apply r_field with (b := a).
      constructor. trivial.
      unfold ofield in H. 
      trivial. 
    apply reachable_trans with (b := b).
    trivial.
    trivial.
Qed.


Lemma ireachable_trans : forall (h : heap) (a b c : addr),
 ireachable h a b -> ireachable h b c -> ireachable h a c.
Proof.
  intros h a b c Iab. generalize dependent c.
  induction Iab.
  Case "i_self".
    intros c Iac. trivial.
  Case "i_field".
    intros d Idc.
    apply IHIab in Idc.
    apply i_field with (b:= b).
    trivial. trivial. trivial. 
Qed.


Lemma reachable_ireachable : forall (h : heap) (a b : addr),
  complete_heap h -> reachable h a b -> ireachable h a b.
Proof.
  intros h a b W Rab. induction Rab.
  Case "r_self".
    constructor. trivial.
  Case "r_field".
    assert (Ibc : ireachable h b c).
      apply i_field with (b := c).
      unfold ofield in H. inversion H as [o [I E]].
      apply addrs_from_heap_correct2.
      exists o. trivial.
      trivial. constructor.
      unfold complete_heap in W.
      unfold ofield in H. inversion H as [o [I E]].
      inversion E as [f F].
      apply addrs_in_fields_correct2 in F.
      apply W with (a := b) in F.
      trivial. trivial.
    apply ireachable_trans with (b := b).
    apply IHRab. trivial.
    trivial.
Qed.

Lemma reachable_iff_ireachable : forall (h : heap) (a b : addr),
  complete_heap h -> (reachable h a b <-> ireachable h a b).
Proof.
  intros h a b. split.
  apply reachable_ireachable. trivial.
  apply ireachable_reachable.
Qed.
    

Lemma reachable_in_heap : forall (h : heap) (a b : addr),
  complete_heap h -> 
     reachable h a b -> In a (addrs_from_heap h) /\ In b (addrs_from_heap h).
Proof.
  intros h a b W R.
  induction R.
  Case "r_self".
    tauto.
  Case "r_field".
    split.
    tauto.
    unfold ofield in H.
    unfold complete_heap in W.
    inversion H as [o [I E]].
    inversion E as [f F].
    apply addrs_in_fields_correct2 in F.
    apply W with (a := b) in F. trivial. trivial.
Qed.

Lemma s_reachable_in_heap : forall (h : heap) (roots : set addr) (a : addr),
  complete_heap h -> s_reachable h roots a -> In a (addrs_from_heap h).
Proof.
  intros h roots a W S.
  unfold s_reachable in S. inversion S as [b [Ib Rba]].
  apply reachable_in_heap in Rba. tauto. trivial.
Qed.


(*
   We define an equivalent of the transitive_owner_check function in the paper.
   Note that transitive_owner_check does NOT fail when encountering an actor id or 
   a mailbox id. This means that, contrary to the paper, one can safely send 
   an actor id in a message. Note also that this does not invalidate the isolation 
   result for the paper since we are being here more liberal with what can be sent.
*)





Fixpoint addrs_from_roots (h : heap) (roots: list addr) : option (set addr) :=
  match roots with
  | nil => Some nil
  | a :: rs => match lookup_heap h a with
               | None => None
               | Some o => 
                    match addrs_from_roots h rs with
                    | None => None
                    | Some ls => Some (set_union Aeq_addr (addrs_in_fields o) ls)
                    end
               end
  end.




Function taddr (h : heap) (roots unreached : set addr) 
               {measure length unreached}: option (set addr) :=
  match (addrs_from_roots h roots) with
  | None => None
  | Some adds => let newroots := (set_inter Aeq_addr adds unreached) in
              match newroots with 
              | nil => Some roots
              | u :: us   => taddr h (set_union Aeq_addr roots newroots)
                               (set_diff Aeq_addr unreached newroots)
              end
  end.
Proof.
  intros h roots unreached adds A a l I.
  rewrite <- I.
  apply length_set_diff_not_empty.
  apply inter_not_empty_com.
  apply inter_inter_not_empty.
  rewrite I.
  intros C.
  inversion C.
Qed.




Fixpoint taddrn (h : heap) (roots unreached : set addr)
                (n : nat) : option (set addr) :=
  match n with
  | 0 => Some roots
  | S m =>   match (addrs_from_roots h roots) with
            | None => None
            | Some adds => let newroots := (set_inter Aeq_addr adds unreached) in
                           match newroots with 
                           | nil => Some roots
                           | u :: us => taddrn h (set_union Aeq_addr roots newroots)
                                             (set_diff Aeq_addr unreached newroots)
                                             m
                           end
            end
  end.


Lemma taddr_eq_taddrn : forall (h : heap) (roots unreached : set addr) (n : nat),
  length unreached < n ->
    taddr h roots unreached = taddrn h roots unreached n.
Proof.
  intros h roots unreached.
  functional induction (taddr h roots unreached); intros n L.
  Case "addrs from roots h = None".
    destruct n as [ | m].
    SCase "n = 0".
      simpl. inversion L.
    SCase "n = m+1".
      simpl. rewrite e. trivial.
  Case "addrs from roots h = Some adds && inter adds unreached = nil".
    destruct n as [ | m].
    SCase "n = 0".
      inversion L.
    SCase "n = m+1".
      simpl. rewrite e. rewrite e0. trivial.
  Case "addrs from roots h = Some adds && inter adds unreached <> nil".
    destruct n as [ | m].
    SCase "n = 0".
      inversion L.
    SCase "n = m+1".
      simpl. rewrite e. rewrite e0. rewrite <- e0.
      apply IHo. 
      assert (LL :
                length (set_diff Aeq_addr unreached (u :: us)) 
                < length unreached ).
          rewrite <- e0.
          apply length_set_diff_not_empty.
          apply inter_not_empty_com.
          apply inter_inter_not_empty.
          rewrite e0. 
          intros C.  inversion C.
      rewrite e0. omega.
Qed.
                




Lemma lookup_correct1: forall (h : heap) (a: addr) (o : object),
  lookup_heap h a = Some o -> In (a, o) h.
Proof.
  intros h a o L. 
  induction h as [ | (e,o') ss].
  Case "h = nil".
    simpl in L. inversion L.
  Case "h = (e,o')::ss".
    simpl in L.
    destruct (Aeq_addr a e).
    SCase "a = e".
      inversion L. subst. simpl. tauto.
    SCase "a <> e".
      simpl. right.
      apply IHss. trivial.
Qed.


Lemma lookup_correct2 : forall (h : heap) (a : addr) (o : object),
  lookup_heap h a = Some o -> set_In a (addrs_from_heap h).
Proof.
  intros h. 
  induction h as [ | (e,o') ss]; intros a o H.
  Case "h =  nil".
    simpl. simpl in H. inversion H.
  Case "h = (e,o')::ss".
    simpl. simpl in H.
    destruct (Aeq_addr a e).
    SCase "a = e".
      apply set_add_intro.
      tauto.
    SCase "a <> e".
      apply set_add_intro.
      right.
      apply IHss with (o := o). trivial.
Qed.



Lemma addrs_in_fields_reachable : forall (h : heap) (a b : addr)(o : object),
  complete_heap h ->
   lookup_heap h a = Some o ->
    In b (addrs_in_fields o) -> 
      reachable h a b.
Proof.
  intros h a b o CH L I.
  assert (F: ofield h a b).
    unfold ofield. 
    apply lookup_correct1 in L. 
      exists o. split.
      trivial.
      apply addrs_in_fields_correct1. trivial.
  apply r_field with (b := a).
  constructor. 
  apply lookup_correct2 with (o := o). trivial.
  trivial.
Qed.


Lemma addrs_in_fields_complete : forall (h : heap) (a b: addr) (o : object),
  wf_heap h ->
   lookup_heap h a = Some o ->
     ofield h a b -> 
        set_In b (addrs_in_fields o).
Proof.
  intros h a b o WF HL HO.
  unfold ofield in HO.
  inversion HO as [o' [I F]].
  inversion F as [f E].
  assert (Eqo : o = o').
    unfold wf_heap in WF. unfold wf_heap in WF. apply WF with (a := a).
    apply lookup_correct1. trivial. trivial.
  subst. rename o' into o.
  apply addrs_in_fields_correct2 with (f := f). trivial.
Qed.




Lemma addrs_from_roots_reachable : 
  forall (h : heap) (roots : set addr) (l : set addr)(a : addr),
    complete_heap h ->
      addrs_from_roots h roots = Some l -> 
         In a l -> s_reachable h roots a.
Proof.
  intros h roots. generalize dependent h. 
  induction roots as [ | r rs]; intros h l a CH H S.
  Case "roots = nil".
    simpl in H. inversion H. subst. inversion S.
  Case "roots = r :: rs".
    simpl in H. remember (lookup_heap h r) as lk.
    destruct lk.
    SCase "lk = Some o".
      remember (addrs_from_roots h rs) as addhrs. destruct addhrs as [l' |].
      SSCase "addhrs = Some l'".
        inversion H. rewrite <- H1 in S.
        apply set_union_elim with (Aeq_dec := Aeq_addr) in S. 
        destruct S as [SL | SR].
        SSSCase "a in o fields".
          unfold s_reachable. exists r. split.
            apply in_eq.
            apply addrs_in_fields_reachable with (o := o). 
            trivial. auto. trivial.
        SSSCase "a in l'".
          symmetry in Heqaddhrs. 
          apply IHrs with (a := a) in Heqaddhrs.
          unfold s_reachable.
          unfold s_reachable in Heqaddhrs.
          inversion Heqaddhrs as [b].
          exists b. split.
            apply in_cons. tauto.
            tauto.
          trivial.
          trivial.
      SSCase "addhrs = None".
        inversion H.
    SCase "lk = None".
        inversion H.
Qed.



Lemma addrs_from_roots_in_heap : 
  forall (h : heap) (roots : set addr) (l : set addr),
    complete_heap h -> 
      addrs_from_roots h roots = Some l ->
         subset l (addrs_from_heap h).
Proof.
  intros h roots l W A.
  unfold subset. intros a I.
  apply s_reachable_in_heap with (roots := roots).
  trivial.
  apply addrs_from_roots_reachable with (l := l).
  trivial. trivial. trivial.
Qed.
  


Lemma addrs_from_roots_complete : 
  forall (h : heap) (roots : set addr) (l : list addr)(s a : addr),
    wf_heap h ->
     addrs_from_roots h roots = Some l ->
      set_In s roots -> 
        ofield h s a -> 
          In a l.
Proof.
  intros h roots. 
  generalize dependent h.
  induction roots as [ | r rs]; intros h l s a W A S H.
  Case "roots = nil".
    inversion S.
  Case "roots = r ::rs".
    simpl in S.
    inversion S as [Hr | Hs].
    SCase "r = s".
      subst. simpl in A. 
      remember (lookup_heap h s) as lko.
      destruct (lko).
      SSCase "lko = Some o".
        remember (addrs_from_roots h rs) as addhrs.
        destruct addhrs.
        SSSCase "addhrs = Some l0".
          inversion A.
          assert (Ha : In a (addrs_in_fields o)).
            apply addrs_in_fields_complete with (h := h) (a := s).
            trivial.
            auto.
            trivial.
          apply set_union_intro. left. trivial.
        SSSCase "addhrs = None".
          inversion A.
      SSCase "lko = None".
        inversion A.
    SCase "s in rs".
      simpl in A. destruct (lookup_heap h r).
      SSCase "Some".
        remember (addrs_from_roots h rs) as addhrs.
        destruct addhrs.
        SSSCase "Some".
          inversion A.
          apply set_union_intro. right.
          apply IHrs with (h:= h)(s := s).
          trivial.
          auto.
          trivial.
          trivial.
        SSSCase "None".
          inversion A.
      SSCase "None".
        inversion A.
Qed.


Lemma taddr_reachable : 
   forall (h : heap) (roots unreached result : set addr) (a : addr),
    complete_heap h ->
     subset roots (addrs_from_heap h) -> 
       subset unreached (addrs_from_heap h) ->
           taddr h roots unreached = Some result ->
             set_In a result -> 
               (s_reachable h roots a).
Proof.
  intros h roots unreached. 
  functional induction (taddr h roots unreached); intros result a W I J T S.
  Case "addrs_from_roots h roots = None".
    inversion T.
  Case "addrs_from_roots h roots = Some ls and set_inter ls unreached = nil".
    inversion T. subst. unfold s_reachable. exists a. split.
    tauto. 
    constructor. unfold subset in I. apply I in S. trivial.
  Case "addrs_from_roots h roots = Some adds and set_inter adds unreached <> nil".
    remember (set_inter Aeq_addr adds unreached) as newroots.
    apply IHo with (a := a) in T. 
    unfold s_reachable in T. inversion T as [b].
    assert (Rb: s_reachable h roots b).
      inversion H as [Ib Rba].
      apply set_union_elim in Ib.
      inversion Ib as [Ibroots | Ibnewroots].
      SCase "b in roots".
        unfold s_reachable. exists b. split. 
        trivial. constructor. apply I. trivial.
      SCase "b in newroots".
        rewrite Heqnewroots in Ibnewroots.
        apply set_inter_elim1 in Ibnewroots.
        apply addrs_from_roots_reachable with (l := adds).
        trivial. trivial. trivial.
    unfold s_reachable in Rb.
    inversion Rb as [c].
    unfold s_reachable. exists c. split.
    tauto. apply reachable_trans with (b:= b). tauto. tauto.
    trivial. apply union_included1. trivial.
    rewrite Heqnewroots. apply inter_included2. trivial.
    apply diff_included. trivial.
    trivial.
Qed.


Lemma taddr_sub : forall (h : heap) (roots unr s : set addr),
  taddr h roots unr = Some s -> subset roots s.
Proof.
  intros h roots unr.
  functional induction (taddr h roots unr); intros s T.
  Case "addrs from roots h = None".
     inversion T.
  Case "addrs from roots h = Some s && inter adds unr = nil".
    inversion T. apply subset_refl.
  Case "addrs from roots h = Some s && inter adds unr <> nil".
    apply IHo in T. 
    apply subset_trans 
         with (y := set_union Aeq_addr roots (set_inter Aeq_addr adds unreached)).
    apply union_included.
    trivial.
Qed.
    


Lemma taddrn_complete : forall (h : heap) (roots unr s : set addr) 
                               (a : addr) (n : nat),
  wfc_heap h -> 
    s_reachable h roots a ->
     subset (set_union Aeq_addr roots unr) (addrs_from_heap h) -> 
      subset (addrs_from_heap h) (set_union Aeq_addr roots unr) ->
       set_inter Aeq_addr roots unr = nil ->
         length unr < n -> 
           taddrn h roots unr n = Some s ->
             set_In a s.
Proof.
  intros h roots unr s a n W Ra.
  unfold s_reachable in Ra. inversion Ra as [b [Ib Rba]].
  clear Ra. generalize dependent roots. generalize dependent unr.
  generalize dependent s. generalize dependent n.
  apply reachable_iff_ireachable in Rba.
  induction Rba; intros n s unr roots I HV HU HI L T.
  Case "r_self".
    assert (S: subset roots s). 
       apply taddr_sub with (h := h) (unr := unr).
       rewrite taddr_eq_taddrn with (n := n). trivial.
       trivial.
    destruct n as [ | m].
    SCase "n = 0".
      inversion L.
    SCase "n = m+1".
      apply S. trivial.
  Case "r_field".
    destruct n as [| m].
    SCase "n = 0".
      inversion L.
    SCase "n = m+1".
      assert (T1: taddrn h roots unr (S m) = Some s). trivial.
      simpl in T1.
      remember (addrs_from_roots h roots) as addhrs.
      destruct addhrs as [adds |].
      SSCase "addhrs = Some adds".
        remember (set_inter Aeq_addr adds unr) as inter.
        destruct inter as [ | u us].
        SSSCase "inter = nil".
          inversion T1. subst. clear T1.
          assert (Ibadds : In b adds).
            apply addrs_from_roots_complete with (h := h) (roots := s) (s:= a).
            unfold wfc_heap in W. tauto.
            auto. trivial. trivial.
          assert (HS: subset adds (set_diff Aeq_addr (addrs_from_heap h) unr)).
            apply included_diff.
            unfold subset. intros e Ie.
            apply addrs_from_roots_reachable with (h := h) (roots := s) in Ie.
            apply s_reachable_in_heap with (roots := s).
            trivial. unfold wfc_heap in W. tauto. auto.
            unfold wfc_heap in W. tauto. auto. auto.
          assert (SS : subset adds s).
            apply included_diff_empty 
              with (Aeq_dec := Aeq_addr) (t := (addrs_from_heap h))
                                           (y := unr).
            apply included_diff.
            apply addrs_from_roots_in_heap with (roots := s).
            unfold wfc_heap in W. tauto. auto. auto.
            trivial.
            apply inter_empty_com. trivial.
          apply IHRba with (n:= S m)(unr := unr)(roots := s).
          trivial. unfold subset in SS. apply SS in Ibadds. trivial.
          trivial. trivial. trivial. trivial.
          trivial.
        SSSCase "inter = u ::us".
          assert (Ibadds : In b adds).
            apply addrs_from_roots_complete with (h := h) (roots := roots) (s := a).
            unfold wfc_heap in W. tauto. auto. trivial. trivial.
          assert (Ibor : In b (set_inter Aeq_addr adds unr) \/ In b roots).
            destruct (set_In_dec Aeq_addr b unr).
            SSSSCase "b in unr".
              left. apply set_inter_intro. trivial. trivial.
            SSSSCase "b not in unr".
              right. 
              assert (Ibh: In b (addrs_from_heap h)).
                apply s_reachable_in_heap with (roots := roots). 
                unfold wfc_heap in W. tauto.
                apply addrs_from_roots_reachable with (l := adds).
                unfold wfc_heap in W. tauto.
                auto. trivial.
              apply inter_union_elim with (Aeq_dec := Aeq_addr) (y := unr).
              apply included_trans_element with (x := (addrs_from_heap h)).
              apply Aeq_addr. trivial. apply HU. trivial.
          inversion Ibor as [Ibnewroots | Ibroots].
          SSSSCase "b in newroots".
            apply IHRba with (n := m)(unr := (set_diff Aeq_addr unr (u :: us)))
                           (roots := set_union Aeq_addr roots (u :: us)).
            trivial.
            apply set_union_intro2. rewrite Heqinter. trivial.
            apply included_diff_union. apply HV.
            rewrite Heqinter. apply inter_included2.
            apply subset_trans with (y := (set_union Aeq_addr roots unr)).
            apply union_included2. trivial.
            apply union_diff_included. trivial.
            rewrite Heqinter.  apply inter_included2.
            apply subset_trans with (y := (set_union Aeq_addr roots unr)).
            apply union_included2. apply HV.
            apply inter_diff_empty with (t:= set_union Aeq_addr roots unr).
            trivial. rewrite Heqinter. 
            apply subset_trans with 
              (y := (set_union Aeq_addr roots (set_inter Aeq_addr adds unr))).
            apply union_included2. apply union_included_union.
            apply subset_refl. apply inter_included2. apply subset_refl.
            apply subset_refl.
            apply length_diff_not_empty_decreases. trivial.
            rewrite Heqinter. apply inter_not_empty_com. 
            apply inter_inter_not_empty. rewrite <- Heqinter.
            intros C. inversion C.
            trivial.
          SSSSCase "b in roots".
            apply IHRba with (n := S m)(unr := unr) (roots := roots).
            trivial. trivial. trivial. trivial. trivial. trivial. trivial.
      SSCase "addhrs = None".
        inversion T1.
        unfold wfc_heap in W. tauto.
Qed.
  
      
Lemma taddr_complete_corollary : forall (h : heap) (a b: addr) 
                                       (s : set addr)(n : nat),
  wfc_heap h ->
    taddrn h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) n = Some s ->
      length (addrs_from_heap h) < n ->
        reachable h a b -> 
          set_In b s.
Proof.
  intros h a b s n W T L R.
  apply taddrn_complete with (h := h)
                             (roots := [a])
                             (unr := set_diff Aeq_addr (addrs_from_heap h) [a])
                             (n := n).
  trivial.
  unfold s_reachable. exists a. split. simpl. left. trivial. trivial.
  apply union_included1.
    unfold subset. intros e Ie. simpl in Ie. inversion Ie.
      subst. apply reachable_in_heap in R. tauto.
      unfold wfc_heap in W. tauto. 
    inversion H.
    unfold subset. intros e Ie. 
      apply set_diff_elim1 with (Aeq_dec := Aeq_addr) (y := [a]).
      trivial.
    unfold subset. intros e Ie. 
      apply set_union_intro. destruct (Aeq_addr e a).
      left. simpl. left. auto.
      right. apply set_diff_intro. trivial. intros C. apply n0. simpl in C.
      inversion C. auto. inversion H.
    apply empty_iff_no_elements. apply Aeq_addr. intros e C. 
      destruct (Aeq_addr e a).
      apply set_inter_elim2 in C. apply set_diff_elim2 in C. 
         apply C. simpl. left. auto.
      apply set_inter_elim1 in C. simpl in C. inversion C. 
        apply n0. auto. inversion H.
    apply lt_trans with (m := length (addrs_from_heap h)). 
      apply length_set_diff_not_empty. intros C. 
      rewrite empty_iff_no_elements in C.
      assert (HT: In a (addrs_from_heap h)). apply reachable_in_heap in R. 
        tauto. unfold wfc_heap in W. tauto.
      assert (HF : not (In a (set_inter Aeq_addr (addrs_from_heap h) [a]))).
        apply C.  
      apply HF. apply set_inter_intro. trivial. simpl. 
      left. trivial. apply Aeq_addr. trivial. trivial.
Qed.


Lemma taddr_complete : forall (h : heap) (a b : addr) (s : set addr),
  wfc_heap h ->
    taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) = Some s ->
      reachable h a b ->
        set_In b s.
Proof.
  intros h a b s W T R.
  rewrite taddr_eq_taddrn with (n := S (length (addrs_from_heap h))) in T.
  apply taddr_complete_corollary with (h := h) (a := a)
                                 (n := S (length (addrs_from_heap h) )).
  trivial. trivial. omega. trivial. 
  apply lt_trans with (m := length (addrs_from_heap h)).
  apply length_set_diff_not_empty. intros C. 
  apply inter_empty_com in C. simpl in C.
  assert (Iah: In a (addrs_from_heap h)).
    apply reachable_in_heap in R. tauto. unfold wfc_heap in W. tauto.
  apply set_mem_correct2 with (Aeq_dec := Aeq_addr) in Iah.
  rewrite Iah in C. inversion C.
  omega.
Qed.



Theorem taddr_sound_complete : forall (h : heap) (a b : addr) (s : set addr),
  wfc_heap h -> 
    set_In a (addrs_from_heap h) ->
    taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) = Some s ->
      (reachable h a b <-> set_In b s).
Proof.
  intros h a b s W T. split.
  Case "->".
    apply taddr_complete. trivial. trivial.
  Case "<-".
    intros Ib.
    assert (SR : s_reachable h [a] b).
      apply taddr_reachable 
           with (unreached := (set_diff Aeq_addr (addrs_from_heap h) [a]))
                (result := s).
      unfold wfc_heap in W. tauto.
      unfold subset. intros e Ie. inversion Ie. subst. trivial.
      inversion H0.
      apply diff_included. apply subset_refl.
      trivial.
      trivial.
    unfold s_reachable in SR.
    inversion SR as [e He].
    inversion He as [Ha R].
    inversion Ha. subst. tauto. inversion H0.
Qed.
    




