(*
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: August 2013
  (c) INRIA
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import Recdef.
Require Import SiaamStructures.
Require Import SiaamUtils.
Require Import SiaamTable.
Require Import SiaamReach.

(*
  Auxiliary functions.
*)



Fixpoint check_owner (toc : list addr) (ows : owners) (o : option aid) : bool :=
  match toc with
  | nil => true
  | a :: ts => match lookup_owner ows a with
               | None => false
               | Some None => match o with
                              | None => check_owner ts ows o
                              | Some _ => false
                              end
               | Some (Some e) => match o with 
                                  | None => false
                                  | Some u => if Aeq_aid u e 
                                              then check_owner ts ows o 
                                              else false
                                  end
               end
  end.



Definition trans_owner_check (h : heap) 
                             (ows: owners) 
                             (o : option aid) 
                             (a : addr): bool :=
  match taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) with
  | None => false
  | Some ls => check_owner ls ows o
  end.


Fixpoint set_owner_update (l : set addr) (o : option aid) 
                          (ows : owners) : owners :=
  match l with
  | nil => ows
  | a :: ls => set_owner_update  ls o (up_owners a o ows)
  end.

Definition trans_owner_update (h : heap) (ows : owners) 
                              (o : option aid) (a : addr) : option owners :=
  match taddr h [a] (set_diff Aeq_addr (addrs_from_heap h) [a]) with
  | None => None
  | Some adds => Some (set_owner_update adds o ows)
  end.



(*
  Rules for local actor actions (no global side-effect actions).
*)

Inductive action : locals -> locals -> Prop :=
  | a_nwbool : forall (l l' : locals) (i : vid)(b : bool),
        not (set_In i (vids_from_locals l)) -> 
           l' = set_add Aeq_lentry (i, vbool b) l -> 
              action l l'
  | a_nwnat : forall (l l' : locals) (i : vid)(n : nat),
        not (set_In i (vids_from_locals l)) -> 
           l' = set_add Aeq_lentry (i, vnat n) l -> 
              action l l'
  | a_nwnull : forall (l l' : locals) (i : vid),
        not (set_In i (vids_from_locals l)) -> 
           l' = set_add Aeq_lentry (i, vnull) l -> 
              action l l'
  | a_compat : forall (l l' : locals) (i j : vid) (vi vj : value),
        set_In (i, vi) l ->
        set_In (j, vj) l ->
         v_compat vi vj ->
           l' = up_locals i vj l ->
             action l l'
  | a_forget : forall (l l' : locals),
        subset l' l -> action l l'.


Tactic Notation "action_cases" tactic(first) ident(c) :=
  first;
  [ Case_aux c "a_nwbool"
  | Case_aux c "a_nwnat" 
  | Case_aux c "a_nwnull" 
  | Case_aux c "a_compat" 
  | Case_aux c "a_forget" ].


(*
  SIAAM rules.
*)


Reserved Notation " t '=s' a '=>' t' " (at level 40).

Inductive reds : conf -> aid -> conf -> Prop :=
  | reds_step : forall (c1 c2 : conf)(a : aid)(l l' : locals), 
      set_In (a,l) (acs c1) -> 
         action l l' -> 
           c2 = mkcf (up_actors a l' (acs c1)) (ows c1) (mbs c1) (shp c1) ->
             c1 =s a => c2

  where " t '=s' a '=>' t' " := (reds t a t').


Reserved Notation " t '=fr' a '=>' t' " (at level 40).

Inductive redfr : conf -> aid -> conf -> Prop :=
  | redfr_step : forall (c1 c2 : conf)(e : aid)
                        (l l' : locals)(i j : vid)
                        (v w : value) (a: addr)(o : object)(f: fid), 
      set_In (e,l) (acs c1) -> 
       set_In (i, w) l -> 
        set_In (j,vadd a) l ->
         set_In (a,o) (shp c1) ->
          set_In (f,v) o ->
           set_In (a, Some e) (ows c1) ->
            v_compat w v ->
             l' = up_locals i v l ->
              c2 = mkcf (up_actors e l' (acs c1)) (ows c1) (mbs c1) (shp c1) ->
               c1 =fr e => c2

  where " t '=fr' a '=>' t' " := (redfr t a t').


Reserved Notation " t '=fw' a '=>' t' " (at level 40).

Inductive redfw : conf -> aid -> conf -> Prop :=
  | redfw_step : forall (c1 c2 : conf)(e : aid)
                        (l l' : locals)(i j : vid)
                        (v w : value) (a: addr)(o : object)(f: fid), 
      set_In (e,l) (acs c1) -> 
       set_In (i, w) l -> 
        set_In (j,vadd a) l ->
         set_In (a,o) (shp c1) ->
          set_In (f,v) o ->
           set_In (a, Some e) (ows c1) ->
            v_compat v w ->
             c2 = mkcf (acs c1) (ows c1) (mbs c1) 
                       (up_heap a (up_object f w o) (shp c1)) ->
              c1 =fw e => c2

  where " t '=fw' a '=>' t' " := (redfw t a t').


Reserved Notation " t '=nwa' a '=>' t' " (at level 40).

Inductive rednwa : conf -> aid -> conf -> Prop :=
  | rednwa_step : forall (c1 c2 : conf)(e ne: aid) (m : mid)(mb : mbox)
                         (le lne: locals)(i k: vid)(w : value), 
      set_In (e,le) (acs c1) -> 
       set_In (i, w) le -> 
        not (set_In ne (aids_from_actors (acs c1))) ->
         subset (values_from_locals lne) (values_from_locals le) ->
          not (set_In k (vids_from_locals lne)) ->
           set_In (m, mb) (mbs c1) ->
            v_compat w (vaid ne) -> 
             c2 = mkcf (set_add Aeq_actor  (ne, (up_locals k (vmid m) lne))
                             (up_actors e (up_locals i (vaid ne) le) (acs c1)))
                       (ows c1)
                       (mbs c1)
                       (shp c1) ->     
              c1 =nwa e => c2

  where " t '=nwa' a '=>' t' " := (rednwa t a t').


Reserved Notation " t '=nwm' a '=>' t' " (at level 40).

Inductive rednwm : conf -> aid -> conf -> Prop :=
  | rednwm_step : forall (c1 c2 : conf)(e : aid) (m : mid)
                         (l l': locals)(i: vid)(v: value), 
      set_In (e,l) (acs c1) -> 
       set_In (i, v) l -> 
        not (set_In m  (mids_from_mboxes (mbs c1))) ->
          v_compat v (vmid m) -> 
            l' = up_locals i (vmid m) l ->
             c2 = mkcf (up_actors e l' (acs c1))
                       (ows c1)
                       (set_add Aeq_mbentry (m, mkmb e []) (mbs c1))
                       (shp c1) ->     
              c1 =nwm e => c2

  where " t '=nwm' a '=>' t' " := (rednwm t a t').


Reserved Notation " t '=nwo' a '=>' t' " (at level 40).

Inductive rednwo : conf -> aid -> conf -> Prop :=
  | rednwo_step : forall (c1 c2 : conf)(e : aid) (a : addr)(o:object)
                         (l l': locals)(i: vid)(v : value), 
      set_In (e,l) (acs c1) -> 
       set_In (i, v) l -> 
        not (set_In a  (addrs_from_owners (ows c1))) ->
         subset (values_in_table fid value Aeq_value o)(values_from_locals l) ->
          v_compat v (vadd a) -> 
            l' = up_locals i (vadd a) l ->
             c2 = mkcf (up_actors e l' (acs c1))
                       (set_add Aeq_oentry (a,Some e) (ows c1))
                       (mbs c1)
                       (set_add Aeq_sentry (a,o) (shp c1)) ->     
              c1 =nwo e => c2

  where " t '=nwo' a '=>' t' " := (rednwo t a t').


Reserved Notation " t '=snd' a '=>' t' " (at level 40).

Inductive redsnd : conf -> aid -> conf -> Prop :=
  | redsnd_step : forall (c1 c2 : conf)(e : aid) (a : addr)
                         (l : locals) (ms: msgid)(mi: mid)
                         (mb mb': mbox) (owns : owners), 
      set_In (e,l) (acs c1) -> 
       set_In (vadd a) (values_from_locals l) -> 
        trans_owner_check (shp c1) (ows c1) (Some e) a = true ->
         set_In (mi,mb) (mbs c1) ->
          not (set_In ms (msgids_from_mbox mb)) ->
           Some owns = trans_owner_update (shp c1) (ows c1) None a ->
            mb' = mkmb (own mb) ((ms,a)::(msgs mb)) -> 
             c2 = mkcf (acs c1)
                       owns
                       (up_mboxes mi mb' (mbs c1))
                       (shp c1) ->     
              c1 =snd e => c2

  where " t '=snd' a '=>' t' " := (redsnd t a t').




Reserved Notation " t '=rcv' a '=>' t' " (at level 40).

Inductive redrcv : conf -> aid -> conf -> Prop :=
  | redrcv_step : forall (c1 c2 : conf) (e : aid) (m : message)
                         (l l': locals) (mb mb': mbox) (mi : mid)
                         (owns : owners)(i : vid)(v : value),
     set_In (e,l) (acs c1) -> 
      set_In (mi,mb) (mbs c1) ->
       set_In (i,v) l ->
       (own mb) = e -> 
         Some m = lasto (msgs mb) ->
          mb' = mkmb (own mb) (removelast (msgs mb)) ->
           Some owns = trans_owner_update (shp c1) (ows c1) (Some e) (snd m) ->
            v_compat v (vadd (snd m)) -> 
             l' = up_locals i (vadd (snd m)) l ->
              c2 = mkcf (up_actors e l' (acs c1))
                        owns
                        (up_mboxes mi mb' (mbs c1))
                        (shp c1) ->
               c1 =rcv e => c2
 
  where " t '=rcv' a '=>' t' " := (redrcv t a t').
           
        



Reserved Notation " t '==>' t' " (at level 40).

Inductive red : conf -> conf -> Prop := 
  | red_silent : forall (c1 c2 : conf)(e : aid),
               c1 =s e => c2 -> c1 ==> c2
  | red_fread : forall (c1 c2 : conf)(e : aid), c1 =fr e => c2  -> c1 ==> c2
  | red_fwrite : forall (c1 c2 : conf)(e : aid), c1 =fw e => c2  -> c1 ==> c2
  | red_newact : forall (c1 c2 : conf)(e : aid), c1 =nwa e => c2 -> c1 ==> c2
  | red_newmbox : forall (c1 c2 : conf)(e : aid), c1 =nwm e => c2 -> c1 ==> c2
  | red_newobj : forall (c1 c2 : conf)(e : aid), c1 =nwo e => c2 -> c1 ==> c2
  | red_send : forall (c1 c2 : conf)(e : aid), c1 =snd e => c2 -> c1 ==> c2
  | red_rcv : forall (c1 c2 : conf)(e : aid), c1 =rcv e => c2 -> c1 ==> c2


  where " t '==>' t' " := (red t t').

Tactic Notation "red_cases" tactic(first) ident(c) :=
  first;
  [ Case_aux c "red_silent"
  | Case_aux c "red_fread" 
  | Case_aux c "red_fwrite" 
  | Case_aux c "red_newact" 
  | Case_aux c "red_newmbox" 
  | Case_aux c "red_newobj"
  | Case_aux c "red_send"
  | Case_aux c "red_rcv" ].



Reserved Notation " t1 '==>*' t2 " (at level 40).

Inductive redt : conf -> conf -> Prop :=
  | redt_refl : forall (c : conf), c ==>* c 
  | redt_trans : forall (c1 c2 c3 : conf),
                  c1 ==> c2 ->
                  c2 ==>* c3 ->
                    c1 ==>* c3

    where " t1 '==>*' t2 " := (redt t1 t2).








