(**
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: June 2013
  (c) INRIA
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import SiaamUtils.
Require Import SiaamTable.




(**
  Basic data types for the SIAAM specification:
  - addr : Object references are called addresses and are representd by nats
  - fid : Field names, or field indexes, are represented by nats
  - vid : Local variable names -- vid -- are represented by nats
  - aid : Actor names are represented by nats
  - mid : Mailbox ids are represented by nats
  - msgid : Message ids are represented by nats
*) 

Inductive addr := mkaddr : nat -> addr.
Inductive fid := mkfid : nat -> fid.
Inductive vid := mkvid : nat -> vid.
Inductive aid := mkaid : nat -> aid.
Inductive mid := mkmid : nat -> mid.
Inductive msgid := mkmsgid : nat -> msgid.


(* 
  Values are restricted to be booleans, nats, object references, actor ids or
   mailbox ids. 
*)

Inductive value : Type :=
  | vnull : value
  | vmark : value
  | vnat : nat -> value
  | vbool: bool -> value
  | vadd: addr -> value
  | vaid: aid -> value
  | vmid : mid -> value.


Inductive v_compat : value -> value -> Prop :=
  | v_nulladd : forall (a : addr), v_compat vnull (vadd a)
  | v_nullaid : forall (a : aid), v_compat vnull (vaid a)
  | v_nullmid : forall (a : mid), v_compat vnull (vmid a)
  | v_nullmark : v_compat vnull vmark
  | v_nulradd : forall (a : addr), v_compat (vadd a) vnull
  | v_nulraid : forall (a : aid), v_compat (vaid a) vnull
  | v_nulrmid : forall (a : mid), v_compat (vmid a) vnull
  | v_nat : forall (n m : nat), v_compat (vnat n) (vnat m)
  | v_bool : forall (a b : bool), v_compat (vbool a) (vbool b)
  | v_add : forall (a b : addr), v_compat (vadd a) (vadd b)
  | v_aid : forall (a b : aid), v_compat (vaid a) (vaid b)
  | v_mid : forall (a b : mid), v_compat (vmid a) (vmid b).



Tactic Notation "value_cases" tactic(first) ident(c) :=
  first;
  [ Case_aux c "vnull"
  | Case_aux c "vmark" 
  | Case_aux c "vnat" 
  | Case_aux c "vbool" 
  | Case_aux c "vadd" 
  | Case_aux c "vaid" 
  | Case_aux c "vmid" 
].


(* 
  An object is a set of pairs <field indexes, values>. 
*)


Definition fentry := tentry fid value.
Definition object := table fid value.


(* 
  A heap is  a set of addresses paired with objects, 
  ie a set of pairs (object reference, object) 
*)

Definition sentry := tentry addr object.
Definition heap := table addr object.


(* 
  A set of local variables is a list of variable names paired with values 
*)

Definition lentry := tentry vid value.
Definition locals := table vid value.


(* 
   A message queue is a list of object references.
   A message is a pair (msgid, addr) 
*)

Definition message := prod msgid addr.
Definition queue := list message.


(* 
  A mailbox is a record (own: a, msgs: q), 
  where:
  - a is an actor id (the owner of the mailbox),
  - q is a queue of messages. 
*)

Record mbox : Type := mkmb { own : aid ; 
                             msgs : queue}.


(* 
  Mailboxes is a list of mailboxes in a configuration. 
*)

Definition mbentry := tentry mid mbox.
Definition mailboxes := table mid mbox.


(* 
  Owners is a set of pairs (object reference, actor id). A pair (r, a) 
  in owners means the object references by r is owned by actor a. 
*)

Definition oentry := tentry addr (option aid).
Definition owners := table addr (option aid).

(* 
  An actor is a pair (aid, locals), where aid is the actor id and locals are
  a set of local actor variables. 
  Actors is a set of actor id and represents the set of active actors 
  in a configuration. 
*)


Definition actor := tentry aid locals.
Definition actors := table aid locals.


(* 
  A system configuration, or configuration for short, is a tuple
   (actors, owners, mailboxes, heap). 
*)

Record conf : Type := mkcf { acs : actors; 
                             ows : owners ; 
                             mbs : mailboxes ;
                             shp : heap }.





(*
  Basic facts about the SIAAM data types
*)


Lemma Aeq_bool : forall (b1 b2 : bool), {b1 = b2} + {b1 <> b2}.
Proof.
  intros b1 b2. destruct b1; destruct b2 ; solve [ left ; trivial 
                                    | right; intros H ; inversion H].
Qed.

Lemma Aeq_nat : forall (n1 n2 : nat), {n1 = n2} + {n1 <> n2}.
Proof.
  intros n1 n2. destruct (eq_nat_decide n1 n2) as [E | N].
  left. apply eq_nat_eq. trivial.
  right. intros H. apply N. apply eq_eq_nat. trivial.
Qed.

Lemma Aeq_addr : forall (a1 a2 : addr), {a1 = a2} + {a1 <> a2}.
Proof.
  intros a1 a2.
  destruct a1 as [n1].
  destruct a2 as [n2].
  destruct (Aeq_nat n1 n2).
  auto. right. intros C. apply n. inversion C. trivial.
Qed.
 

Lemma Aeq_aid : forall (a1 a2 : aid), {a1 = a2} + {a1 <> a2}.
Proof.
  intros a1 a2.
  destruct a1 as [n1].
  destruct a2 as [n2].
  destruct (Aeq_nat n1 n2).
  auto. right. intros C. apply n. inversion C. trivial.
Qed.

Lemma Aeq_oaid : forall (a1 a2 : option aid), {a1 = a2} + {a1 <> a2}.
Proof.
  intros a1 a2. 
  destruct a1 as [ e1 | ].
    destruct a2 as [ e2 | ].
      destruct (Aeq_aid e1 e2).
        subst. left. trivial.
        right. intros C. apply n. inversion C. trivial.
      right. intros C. inversion C.
    destruct a2 as [e2 | ].
      right. intros C.  inversion C.
      left. trivial.
Qed.
      
Lemma Aeq_mid : forall (m1 m2 : mid), {m1 = m2} + {m1 <> m2}.
Proof.
  intros a1 a2.
  destruct a1 as [n1].
  destruct a2 as [n2].
  destruct (Aeq_nat n1 n2).
  auto. right. intros C. apply n. inversion C. trivial.
Qed.

Lemma Aeq_vid : forall (v1 v2 : vid), {v1 = v2} + {v1 <> v2}.
Proof.
  intros a1 a2.
  destruct a1 as [n1].
  destruct a2 as [n2].
  destruct (Aeq_nat n1 n2).
  auto. right. intros C. apply n. inversion C. trivial.
Qed.


Lemma Aeq_fid : forall (f1 f2 : fid), {f1 = f2} + {f1 <> f2}.
Proof.
  intros a1 a2.
  destruct a1 as [n1].
  destruct a2 as [n2].
  destruct (Aeq_nat n1 n2).
  auto. right. intros C. apply n. inversion C. trivial.
Qed.

Lemma Aeq_msgid : forall (m1 m2 : msgid), {m1 = m2} + {m1 <> m2}.
Proof.
  intros a1 a2.
  destruct a1 as [n1].
  destruct a2 as [n2].
  destruct (Aeq_nat n1 n2).
  auto. right. intros C. apply n. inversion C. trivial.
Qed.


Lemma Aeq_value : forall (v1 v2: value), {v1 = v2} + {v1 <> v2}.
Proof.
  intros v1 v2. destruct v1.
    destruct v2; solve [left; trivial | right;  intros H; inversion H].
    destruct v2; solve [right ; intros C ; inversion C | left; trivial].
    destruct v2 ; solve[
      right; intros H; inversion H
      | destruct (Aeq_nat n n0) as [E | N]; 
        [left; rewrite E; trivial
         | right; intros H; inversion H; apply N; trivial]].
    destruct v2 ; solve [right ; intros H; inversion H
           | destruct (Aeq_bool b b0);
               [left; rewrite e; trivial
               | right; intros H; inversion H; apply n; trivial]].
    destruct v2 ; solve[
      right; intros H; inversion H
      | destruct (Aeq_addr a a0) as [E | N]; 
        [left; rewrite E; trivial
         | right; intros H; inversion H; apply N; trivial]].
    destruct v2 ; solve[
      right; intros H; inversion H
      | destruct (Aeq_aid a a0) as [E | N]; 
        [left; rewrite E; trivial
         | right; intros H; inversion H; apply N; trivial]].
    destruct v2 ; solve[
      right; intros H; inversion H
      | destruct (Aeq_mid m m0) as [E | N]; 
        [left; rewrite E; trivial
         | right; intros H; inversion H; apply N; trivial]].
Qed.


Lemma Aeq_fentry : forall (e1 e2 : fentry), {e1 = e2} + {e1 <> e2}.
Proof.
  intros e1 e2. destruct e1 as [a1 v1]. destruct e2 as [a2 v2]. 
  destruct (Aeq_value v1 v2).
    destruct (Aeq_fid a1 a2).
      subst. left. trivial.
      right. intros C. inversion C. apply n. trivial.
    right. intros C. inversion C. apply n. trivial.
Qed.


Lemma Aeq_object : forall (o1 o2 : object), {o1 = o2} + {o1 <> o2}.
Proof.
  intros o1 o2. apply list_eq_dec. apply Aeq_fentry.
Qed.


Lemma Aeq_sentry : forall (e1 e2 : sentry), {e1 = e2} + {e1 <> e2}.
Proof.
  intros e1 e2. destruct e1 as [a1 o1]. destruct e2 as [a2 o2].
  destruct (Aeq_addr a1 a2) as [E | N].
    destruct (Aeq_object o1 o2) as [EO | NO].
      left. subst. trivial.
      right. intros C. inversion C. apply NO. trivial.
    right. intros C. inversion C. apply N. trivial.
Qed.

Lemma Aeq_heap : forall (h1 h2 : heap), {h1 = h2} + {h1 <> h2}.
Proof.
  intros h1 h2. apply list_eq_dec. apply Aeq_sentry.
Qed.


Lemma Aeq_lentry : forall (e1 e2 : lentry), {e1 = e2} + {e1 <> e2}.
Proof.
  intros e1 e2. destruct e1 as [a1 o1]. destruct e2 as [a2 o2].
  destruct (Aeq_vid a1 a2) as [E | N].
    destruct (Aeq_value o1 o2) as [EO | NO].
      left. subst. trivial.
      right. intros C. inversion C. apply NO. trivial.
    right. intros C. inversion C. apply N. trivial.
Qed.

Lemma Aeq_locals : forall (l1 l2 : locals), {l1 = l2} + {l1 <> l2}.
Proof.
  intros l1 l2. apply list_eq_dec. apply Aeq_lentry.
Qed. 

Lemma Aeq_actor : forall (a1 a2 : actor),  {a1 = a2} + {a1 <> a2}.
Proof.
  intros a1 a2. destruct a1 as [i1 l1]. destruct a2 as [i2 l2].
  destruct (Aeq_aid i1 i2) as [E | N].
    destruct (Aeq_locals l1 l2) as [EL | NL].
      subst. left. trivial.
      right. intros C. apply NL. inversion C. trivial.
    right. intros C. inversion C. apply N. trivial.
Qed.
  


Lemma Aeq_message : forall (m1 m2 : message), {m1 = m2} + {m1 <> m2}.
Proof.
  intros m1 m2. destruct m1 as [i1 a1]. destruct m2 as [i2 a2].
  destruct (Aeq_msgid i1 i2) as [E | N].
    destruct (Aeq_addr a1 a2) as [En | Nn].
      left. subst. trivial.
      right. intros C. inversion C. apply Nn. trivial.
    right. intros C. inversion C. apply N. trivial.
Qed.


Lemma Aeq_queue : forall (q1 q2 : queue), {q1 = q2} + {q1 <> q2}.
Proof.
  intros q1 q2. apply list_eq_dec. apply Aeq_message.
Qed.


Lemma Aeq_mbox : forall (m1 m2 : mbox), {m1 = m2} + {m1 <> m2}.
Proof.
  intros m1 m2.
  destruct m1 as [ o1 q1 ].
  destruct m2 as [ o2 q2 ].
    destruct (Aeq_aid o1 o2) as [Eo | No].
      destruct (Aeq_queue q1 q2) as [Eq | Nq].
        left. subst. trivial.
        right. intros C. inversion C. apply Nq. trivial.
      right.  intros C. inversion C. apply No. trivial.
Qed.

Lemma Aeq_oentry : forall (o1 o2 : oentry), {o1 = o2} + {o1 <> o2}.
Proof.
  intros o1 o2. destruct o1 as [a i]. destruct o2 as [b j].
  destruct (Aeq_addr a b) as [E | N].
    destruct i as [e1 | ].
      destruct j as [e2 |].
        destruct (Aeq_aid e1 e2) as [E' | N'].
          left. subst. trivial.
          right . intros C. inversion C. apply N'. trivial.
        right. intros C. inversion C.
      destruct j as [e2 | ].
        right. intros C. inversion C.
        left. subst. trivial.
  right. intros C. inversion C. apply N. trivial.
Qed.

Lemma Aeq_mbentry : forall (m1 m2 : mbentry), {m1 = m2} + {m1 <> m2}.
Proof.
  intros m1 m2.
  destruct m1 as [i1 mb1].
  destruct m2 as [i2 mb2].
  destruct (Aeq_mid i1 i2).
    destruct (Aeq_mbox mb1 mb2).
      subst. left. trivial.
      right. intros C. inversion C. apply n. trivial.
    right. intros C. inversion C. apply n. trivial.
Qed.

(*************************************************************
  Operations and predicates on SIAAM structures.
**************************************************************)

(*
  Update functions for various structures.
*)
 

Definition up_locals (i : vid) (v : value) (l : locals): locals :=
  up_table vid value Aeq_vid i v l.


Definition up_actors (a : aid) (l : locals) (acs: actors) : actors :=
  up_table aid locals Aeq_aid a l acs.


Definition up_object (f : fid)(v : value) (o : object): object :=
  up_table fid value Aeq_fid f v o.


Definition up_heap (a: addr) (o : object) (s : heap) : heap :=
  up_table addr object Aeq_addr a o s.


Definition up_mboxes (i : mid) (m : mbox) (mbs : mailboxes) : mailboxes :=
  up_table mid mbox Aeq_mid i m mbs.


Definition up_owners (a: addr) (o : option aid) (owns : owners) : owners :=
  up_table addr (option aid) Aeq_addr a o owns.



(*
  vids_from_locals extracts the list of variable ids (vid) from a locals set.
*)

Definition vids_from_locals (l : locals) : set vid :=
  fold_right (fun (e : lentry) (vids : set vid) => 
                  set_add Aeq_vid (fst e) vids)
             nil l.


Lemma vids_from_locals_correct1 : forall (l : locals) (i : vid) (v: value),
  set_In (i,v) l -> set_In i (vids_from_locals l).
Proof.
  intros l. induction l as [ | e ls]; intros i v H.
  Case "l = []".
    inversion H.
  Case "l = e :: ls".
    destruct (Aeq_lentry (i,v) e) as [Y | N].
    SCase "e = (i,v)".
      unfold vids_from_locals. simpl.
      rewrite <- Y. simpl. apply set_add_intro. left. trivial.
    SCase "e <> (i,v)".
      simpl in H. inversion H.
      apply ex_falso_quodlibet. apply N. auto.
      unfold vids_from_locals. simpl.
        destruct e as [j w]. simpl.
        apply set_add_intro. destruct (Aeq_vid i j) as [Eij | Nij].
        SSCase "i = j".
          left. trivial.
        SSCase "i <> j".
          right. apply IHls with (v := v). trivial.
Qed.

Lemma vids_from_locals_correct2 : forall (l : locals) (i : vid),
  set_In i (vids_from_locals l) -> exists (v : value), set_In (i,v) l.
Proof.
  intros l. induction l as [ | e ls]; intros i H.
  Case "l = nil".
    simpl in H. inversion H.
  Case "l = e:: ls".
    destruct e as [ j w ]. simpl in H.
    apply set_add_elim in H. inversion H as [E | S].
      subst. exists w. simpl. left. trivial.
      apply IHls in S. inversion S as [v].
       exists v. simpl. right. trivial.
Qed.
    
  
Lemma vids_from_locals_indexes : forall (l :locals),
   vids_from_locals l = indexes_in_table vid value Aeq_vid l.
Proof.
  intros l. induction l as [ | (i,v) ls].
  Case "l = nil".
    simpl. trivial.
  Case "l = (i,v) :: ls".
    simpl. rewrite IHls. trivial.
Qed.



(*
  values_from_locals extracts the list of values from a locals set.
*)

Definition values_from_locals (l : locals) : set value :=
  fold_right (fun (e : lentry) (vals : set value) =>
                  set_add Aeq_value (snd e) vals)
             nil l.


Lemma values_from_locals_values : forall (l : locals),
  values_from_locals l = values_in_table vid value Aeq_value l.
Proof.
  intros l.
  induction l as [ | (i,v) ls].
  Case "l = nil".
    simpl. trivial.
  Case "l = (i,v)::ls".
    simpl. rewrite IHls. trivial.
Qed.



(*
  addrs_from_owners extracts the list of object references (addr) from an owners set.
*)

Definition addrs_from_owners (o : owners) : set addr :=
  fold_right (fun (p : addr * (option aid)) (adds : set addr) =>
                  set_add Aeq_addr (fst p) adds)
             nil o.


Lemma addrs_from_owners_indexes : forall (o : owners),
  addrs_from_owners o = indexes_in_table addr (option aid) Aeq_addr o.
Proof.
  intros o.
  induction o as [ | (a, e) os].
  Case "o = nil".
    simpl. trivial.
  Case "o = e::os".
    simpl. rewrite IHos. trivial.
Qed.

(*
  aids_from_owners extracts the list of actor ids (aid) from a owners set.
*)

Definition aids_from_owners (o : owners) : set aid :=
   fold_right (fun (p : addr * (option aid)) (aids: set aid) => 
                   match (snd p) with
                   | None => aids 
                   | Some a => set_add Aeq_aid a aids
                   end) 
              nil o.


(*
  aids_from_actors extracts the list of actor ids from a set of actors.
*)

Definition aids_from_actors (acs : actors) : set aid :=
  fold_right (fun (a : actor)(aids : list aid) => set_add Aeq_aid (fst a) aids)
             nil acs.

Lemma aids_from_actors_indexes : forall (acs : actors),
  aids_from_actors acs = indexes_in_table aid locals Aeq_aid acs.
Proof.
  intros acs.
  induction acs as [ | (i,l) acss].
  Case "acs = nil".
    simpl. trivial.
  Case "acs = (i,l)::acss".
    simpl. rewrite IHacss. trivial.
Qed.
    

(*
  aids_from_boxes extracts the list of actor ids from a set of mailboxes.
*)

Definition aids_from_mboxes (mbxs : mailboxes) : set aid :=
   fold_right 
    (fun (p : mid * mbox) (aids : list aid) => set_add Aeq_aid (own (snd p)) aids) 
    nil mbxs.


(*
  msgids_from_mbox extracts the list of message ids (msgids) from a mailbox.
*)

Definition msgids_from_mbox (mb : mbox): set msgid :=
   fold_right 
     (fun (m : message) (msgids : list msgid) => set_add Aeq_msgid (fst m) msgids)
     nil (msgs mb).

(*
  mids_from_mboxes extracts the list of mailbox ids (mids) from a set of mailboxes.
*)

Definition mids_from_mboxes (mbxs : mailboxes) : set mid :=
   fold_right (fun (p : mid * mbox) (mids : list mid) => set_add Aeq_mid (fst p) mids)
              nil mbxs.


Lemma mids_from_mboxes_indexes : forall (mbxs : mailboxes),
  mids_from_mboxes mbxs = indexes_in_table mid mbox Aeq_mid mbxs.
Proof.
  intros mbxs.
  induction mbxs as [ | (m,mb) mbxss].
  Case "mbxs = nil".
    simpl. trivial.
  Case "mbxs = (m,mb)::mbxss".
    simpl. rewrite IHmbxss. trivial.
Qed.

(*
  addrs_from_heap extracts the list of object references (addr) appearing in
  (the store of) a heap.
*)

Definition addrs_from_heap (h : heap) : set addr :=
   fold_right 
    (fun (p : addr * object) (addrs: list addr) => set_add Aeq_addr (fst p) addrs) 
    nil h.


Lemma addrs_from_heap_indexes : forall (h: heap),
  addrs_from_heap h = indexes_in_table addr object Aeq_addr h.
Proof.
  intros h.
  induction h as [ | (a,o) hs].
  Case "h = nil".
    simpl. trivial.
  Case "h = (a,o)::hs".
    simpl.
    rewrite <- IHhs.
    trivial.
Qed.

Lemma addrs_from_heap_correct1 : forall (h : heap) (a : addr),
  set_In a (addrs_from_heap h) -> exists (o : object), set_In (a, o) h.
Proof.
  intros h.
  induction h as [ | (e,o) ss]; intros a Ia.
  Case "s = nil".
    simpl in Ia. inversion Ia.
  Case "s = (e,o)::ss".
    simpl. destruct (Aeq_addr e a) as [Tea | Fea].
    SCase "e = a".
      subst. exists o. left. trivial.
    SCase "e <> a".
      simpl in Ia. fold addrs_from_heap in Ia.
      unfold addrs_from_heap in IHss. simpl in IHss.
      apply set_add_elim in Ia.
      inversion Ia as [F | T].
        apply ex_falso_quodlibet. apply Fea. auto.
        apply IHss in T. inversion T as [o'].
          exists o'. right. trivial.
Qed.

Lemma addrs_from_heap_correct2 : forall (h : heap) (a : addr),
  (exists (o : object), set_In (a, o) h) -> set_In a (addrs_from_heap h).
Proof.
  intros h. 
  induction h as [ | (e,o) ss]; intros a E.
  Case "s = nil".
    simpl. simpl in E. inversion E. trivial.
  Case "s = (e,o)::ss".
    simpl. destruct (Aeq_addr e a) as [Y | N].
    SCase "E: e = a".
      apply set_add_intro. left. auto.
    SCase "N: e <> a".
      apply set_add_intro.
      right. apply IHss. simpl in E.
      inversion E as [o' [F | T]].
      inversion F. apply ex_falso_quodlibet. apply N. trivial.
      exists o'. simpl. trivial.
Qed.
      


(*
  addrs_from_mbox extracts the list of object references (addr) appearing in
  (the message queue of) a mailbox.
*)
Definition addrs_from_mbox (mb : mbox) : set addr :=
  map (fun msg : message => snd msg) (msgs mb).

(*
  addrs_from_mboxes extracts the list of object references (addr) appearing in
  a set of mailboxes.
*)

Definition addrs_from_mboxes (m : mailboxes) : set addr :=
  fold_right (fun (p : mid * mbox) (addrs : list addr) => 
                  set_union Aeq_addr (addrs_from_mbox (snd p))  addrs) 
             nil m.


Fixpoint in_heap (r : addr) (s : heap) : bool :=
  match s with
  | nil => false
  | (e,o) :: ss => if (Aeq_addr e r) then true else in_heap r ss
  end.


Fixpoint in_owners (r : addr) (ows : owners) : bool :=
  match ows with
  | nil => false
  | (e,ow) :: owss => if (Aeq_addr e r) then true else in_owners r owss
  end.


(* 
   Considering that the present setting is essentially untyped, 
   we have no need for the addr-fields function specified in the paper. 
   We define a function addrs_in_fields that plays a similar role to 
   non-null-addres in the paper since it returns a list
   of (possibly duplicated) object references contained in object field. 
 *)

Fixpoint addrs_in_fields (o : object) : set addr :=
  match o with
  | nil => nil
  | (i,vnull) :: fs => addrs_in_fields fs
  | (i,vmark) :: fs => addrs_in_fields fs
  | (i,vnat n):: fs => addrs_in_fields fs
  | (i,vbool b):: fs => addrs_in_fields fs
  | (i,vaid a) :: fs => addrs_in_fields fs
  | (i,vmid m) :: fs => addrs_in_fields fs
  | (i,vadd a) :: fs => set_add Aeq_addr a (addrs_in_fields fs)
  end.


Lemma addrs_in_fields_correct1 : forall (o : object) (a : addr),
  In a (addrs_in_fields o) -> 
    exists (f : fid), In (f, vadd a) o.
Proof.
  intros o. induction o as [ | (i,v) o']; intros a S.
  Case "o = nil".
    simpl in S. inversion S.
  Case "o = (i,v)::o'".
    simpl in S. destruct v.
    SCase "v = null".
      simpl.  apply IHo' in S. destruct S as [f SE]. exists f.
      destruct (Aeq_fentry (f, vadd a) (i, vnull)).
      left. auto.
      right. trivial.
    SCase "v = vmark".
      simpl. apply IHo' in S. destruct S as [f SE]. exists f.
      destruct (Aeq_fentry (f,vadd a) (i,vmark)).
      left. auto.
      right. trivial.
    SCase "v = nat".
      simpl. apply IHo' in S. destruct S as [f SE]. exists f.
      right. trivial.
    SCase "v = bool".
      simpl. apply IHo' in S. destruct S as [f SE]. exists f.
      right. trivial.
    SCase "v = addr".
      simpl. remember (Aeq_addr a a0) as eaa. destruct eaa.
      SSCase "a = a0".
        subst. exists i. left. trivial.
      SSCase "a <> a0".
        assert (HS : In a (addrs_in_fields o')).
          apply set_add_elim2 with (Aeq_dec := Aeq_addr) (b := a0).
          trivial. trivial.
        apply IHo' in HS. inversion HS as [f SE].
        exists f. right. trivial.
    SCase "v = aid".
      simpl. apply IHo' in S. inversion S as [f SE]. exists f.
      right. trivial.
    SCase "v = mid".
      simpl. apply IHo' in S. inversion S as [f SE]. exists f.
      right. trivial.
Qed.

Lemma addrs_in_fields_correct2 : forall (o : object) (a: addr) (f: fid),
  In (f, vadd a) o -> In a (addrs_in_fields o).
Proof.
  intros o. induction o as [ | (i,v) os]; intros a f H.
  Case "o = nil".
    simpl in H. inversion H.
  Case "o = (i,v)::os".
    simpl. destruct v.
    SCase "v : null".
      simpl in H. inversion H as [H1 | H2].
        inversion H1.
        apply IHos with (f := f). trivial.
    SCase "v : mark".
      simpl in H. inversion H as [H1 | H2].
        inversion H1.
        apply IHos with (f := f). trivial.
    SCase "v : nat".
      simpl in H. inversion H as [H1 | H2].
        inversion H1. 
        apply IHos with (f := f). trivial.
    SCase "v : bool".
      simpl in H. inversion H as [H1 | H2].
        inversion H1.
        apply IHos with (f := f). trivial.
    SCase "v : addr".
      simpl. destruct (Aeq_addr a a0).
      SSCase "a = a0".
        subst. rename a0 into a.
        apply set_add_intro2. 
        trivial.
      SSCase "a <> a0".
        apply set_add_intro.
        right.
        simpl in H. inversion H as [H1 | H2].
          inversion H1. apply ex_falso_quodlibet. apply n. auto.
        apply IHos with (f := f). trivial. 
    SCase "v : aid".
      simpl in H. inversion H as [H1 | H2].
        inversion H1.
        apply IHos with (f := f). trivial.
    SCase "v : mid".
      simpl in H. inversion H as [H1 | H2].
        inversion H1.
        apply IHos with (f := f). trivial.
Qed.


(*
  addrs_in_actor extracts the list of object references contained in local variables
  of an actor.
*)

Fixpoint addrs_in_locals (l : locals) : set addr :=
    match l with
  | nil => nil
  | (i,vnull) :: ls => addrs_in_locals ls
  | (i,vmark) :: ls => addrs_in_locals ls
  | (i,vnat n):: ls => addrs_in_locals ls
  | (i,vbool b):: ls => addrs_in_locals ls
  | (i,vaid a) :: ls => addrs_in_locals ls
  | (i,vmid m) :: ls => addrs_in_locals ls
  | (i,vadd a) :: ls => set_add Aeq_addr a (addrs_in_locals ls)
  end.


Definition addrs_in_actor (a : actor) : set addr := addrs_in_locals (snd a).


(*
  addrs_from_actors extracts the list of object references contained in local
  actor variables.
*)

Definition addrs_from_actors (acs : actors) : set addr :=
  fold_right (fun (a : actor) (addrs : list addr) => 
                    set_union Aeq_addr (addrs_in_actor a) addrs)
             nil acs.





Lemma addrs_in_locals_nw : forall (l : locals)(i : vid) (v : value),
   not (exists (a : addr), v = vadd a) ->
    not (set_In i (vids_from_locals l)) ->
     subset (addrs_in_locals (set_add Aeq_lentry (i,v) l)) (addrs_in_locals l).
Proof.
  intros l.
  induction l as [ | (j,w) ls]; intros i v E S.
  Case "l = nil".
    simpl. 
    value_cases (destruct v) SCase; try (unfold subset; trivial).
      SCase "vadd".
        apply ex_falso_quodlibet. apply E. exists a. trivial.
  Case "l = ((j,w):: ls".
    simpl.
    assert (N: i <> j).
      SCase "Proof of N".
        intros C. apply S. simpl.
        apply set_add_intro. left. trivial.
    destruct (Aeq_lentry (i,v)(j,w)).
      apply ex_falso_quodlibet. apply N. inversion e. trivial.
      simpl.
      value_cases (destruct w) SCase; 
        try (apply IHls; trivial; intros C; apply S; 
             simpl; apply set_add_intro; right; trivial).
      SCase "vadd".
        apply add_included. apply IHls. trivial.
        intros C. apply S. simpl. apply set_add_intro. right. trivial.
Qed.


Lemma addrs_in_locals_up1 : forall (l : locals)(i : vid) (w : value),
    not (exists (a : addr), w = vadd a) ->
      subset (addrs_in_locals (up_locals i w l)) (addrs_in_locals l).
Proof.
  intros l.
  induction l as [ | (k,u) ls]; intros i w S.
  Case "l = nil".
    simpl. unfold subset. trivial.
  Case "l = (k,u)::ls".
    simpl.
    remember (Aeq_vid i k) as Aik.
    destruct Aik.
    SCase "i = k".
      value_cases (destruct u) SSCase; try(simpl;
                      destruct w ; try (apply IHls ; trivial);
                      apply ex_falso_quodlibet; apply S; exists a; trivial;
                      apply ex_falso_quodlibet; apply S; inversion S; inversion S).
       SSCase "vadd".
        simpl.
         destruct w ; try (apply add_included_sub ; apply IHls ; trivial).
          SSCase "vadd".
            apply ex_falso_quodlibet. apply S. exists a0. trivial.
       SSCase "vaid".
        simpl.
         value_cases (destruct w) SSSCase; try (apply IHls ; trivial).
         SSSCase "vadd".
           apply ex_falso_quodlibet. apply S. exists a0. trivial.
    SCase "i <> k".
      simpl. value_cases (destruct u) SSCase; try(apply IHls; trivial).
      SSCase "vadd".
        apply add_included. apply IHls. trivial.
Qed.


Lemma addrs_in_locals_up2 : forall (l : locals)(i: vid) (a :addr),
  subset (addrs_in_locals (up_locals i (vadd a) l))
         (set_add Aeq_addr a (addrs_in_locals l)).
Proof.
  intros l.
  induction l as [ | (j,w) ls]; intros i a.
  Case "l = nil".
    simpl. unfold subset. intros b C. inversion C.
  Case "l = (j,w)::ls".
    simpl.
    remember (Aeq_vid i j) as Aij.
    destruct Aij.
    SCase "i = j".
      subst. simpl.
      destruct w.
        SSCase "vnull".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply IHls.
        SSCase "vmark".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply IHls.
        SSCase "vnat".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply IHls.
        SSCase "vbool".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply IHls.
        SSCase "vadd".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply add_add_com.
          apply add_included_sub.
          apply IHls.
        SSCase "vaid".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply IHls.
        SSCase "vmid".
          apply add_included_elim.
          apply set_add_intro. left. trivial.
          apply IHls.
    SCase "i <> j".
      simpl.
      destruct w; try (apply IHls).
      SSCase "vadd".
        apply add_add_com.
        apply add_included.
        apply IHls.
Qed.


Lemma addrs_in_locals_up3 : forall (l : locals)(i : vid) (a : addr),
     set_In a (addrs_in_locals l) ->
      subset (addrs_in_locals (up_locals i (vadd a) l)) (addrs_in_locals l).
Proof.
  intros l i a H.
  apply in_add_in_sub with (Aeq_dec := Aeq_addr)(a:= a).
    trivial.
    apply addrs_in_locals_up2.
Qed.


Lemma values_addrs_in_locals1 : forall (l : locals)(a : addr),
  set_In (vadd a) (values_in_table vid value Aeq_value l) ->
    set_In a (addrs_in_locals l).
Proof.
  intros l. 
  induction l as [ | (i,v) ls]; intros a H.
  Case "l = nil".
    simpl in H. tauto.
  Case "l = (i,v)::ls".
    simpl.
    value_cases (destruct v) SCase.
    SCase "vnull".
      simpl in H. apply set_add_elim in H.
      inversion H as [E | S].
        inversion E.
        apply IHls. trivial.
    SCase "vmark".
      simpl in H. apply set_add_elim in H.
      inversion H as [E | S].
        inversion E.
        apply IHls. trivial.
   SCase "vnat".
      simpl in H. apply set_add_elim in H.
      inversion H as [E | S].
        inversion E.
        apply IHls. trivial.
   SCase "vbool".
      simpl in H. apply set_add_elim in H.
      inversion H as [E | S].
        inversion E.
        apply IHls. trivial.
   SCase "vadd".
     apply set_add_intro.
     simpl in H. apply set_add_elim in H.
     inversion H as [E | S].
       inversion E. left. trivial.
       right. apply IHls. trivial.
   SCase "vaid".
      simpl in H. apply set_add_elim in H.
      inversion H as [E | S].
        inversion E.
        apply IHls. trivial.
  SCase "vmid".
      simpl in H. apply set_add_elim in H.
      inversion H as [E | S].
        inversion E.
        apply IHls. trivial.
Qed.


Lemma addrs_in_locals_up4 : forall (l : locals) (i : vid) (v: value),
  set_In v (values_in_table vid value Aeq_value l) ->
   subset (addrs_in_locals (up_locals i v l)) (addrs_in_locals l).
Proof.
  intros l i v H.
  value_cases (destruct v) Case.
  Case "vnull".
    apply addrs_in_locals_up1.  intros C. inversion C as [w Cw]. inversion Cw.
  Case "vmark".
    apply addrs_in_locals_up1;  intros C; inversion C as [w Cw]; inversion Cw.
  Case "vnat".
    apply addrs_in_locals_up1;  intros C; inversion C as [w Cw]; inversion Cw.
  Case "vbool".
    apply addrs_in_locals_up1;  intros C; inversion C as [w Cw]; inversion Cw.
  Case "vadd".
    apply addrs_in_locals_up3. apply values_addrs_in_locals1. trivial.
  Case "vaid".
    apply addrs_in_locals_up1;  intros C; inversion C as [w Cw]; inversion Cw.
  Case "vmid".
    apply addrs_in_locals_up1;  intros C; inversion C as [w Cw]; inversion Cw.
Qed.


Lemma addrs_in_locals_sub : forall (l1 l2 : locals),
  subset l2 l1 ->
    subset (addrs_in_locals l2) (addrs_in_locals l1).
Proof.
  intros l1 l2. 
  generalize dependent l1.
  induction l2 as [ | (n,v) ls]; intros l1 H.
  Case "l2 = nil".
    simpl. unfold subset.
    intros a Ia. inversion Ia.
  Case "l2 = (n,v)::ls".
    simpl.
    value_cases (destruct v) SCase.
    SCase "vnull".
      apply IHls. apply cons_subset with (a := (n,vnull)). apply Aeq_lentry.
      trivial.
    SCase "vmark".
      apply IHls. apply cons_subset with (a := (n,vmark)). apply Aeq_lentry.
      trivial.
    SCase "vnat".
      apply IHls. apply cons_subset with (a := (n,vnat n0)). apply Aeq_lentry.
      trivial.
    SCase "vbool".
      apply IHls. apply cons_subset with (a := (n,vbool b)). apply Aeq_lentry.
      trivial.
    SCase "vadd".
      apply in_add_in_sub with (Aeq_dec := Aeq_addr)(a := a).
      assert (A: set_In (n,vadd a) l1).
        SSCase "Proof of A".
          unfold subset in H. apply H. simpl. left. trivial.
      apply values_addrs_in_locals1.
      apply values_in_table_correct1 with (n:= n). trivial.
      apply add_included. 
      apply IHls.
      apply cons_subset with (a0 := (n,vadd a)). apply Aeq_lentry. trivial.
    SCase "vaid".
      apply IHls. apply cons_subset with (a0 := (n,vaid a)). apply Aeq_lentry.
      trivial.
    SCase "vmid".
      apply IHls. apply cons_subset with (a := (n,vmid m)). apply Aeq_lentry.
      trivial.
Qed.

Lemma addr_in_locals_value : forall (l : locals)(a : addr),
  set_In a (addrs_in_locals l) -> set_In (vadd a) (values_from_locals l).
Proof.
  intros l.
  induction l as [ | (i,v) ls]; intros a H.
  Case "l = nil".
    simpl in H. tauto.
  Case "l = (i,v)::ls".
    simpl.
    apply set_add_intro.
    simpl in H.
    value_cases (destruct v) SCase;
      try (right ; apply IHls ; trivial; fail).
      SCase "vadd".
        apply set_add_elim in H.
        inversion H as [E | S].
          subst. left. trivial.
          right. apply IHls. trivial.
Qed.

Lemma value_in_locals_addr : forall (l : locals) (a: addr),
  set_In (vadd a) (values_from_locals l) -> set_In a (addrs_in_locals l).
Proof.
  intros l.
  induction l as [ | (i,v) ls]; intros a H.
  Case "l = nil".
    simpl in H. tauto.
  Case "l = (i,v)::ls".
    simpl.
    simpl in H.
    apply set_add_elim in H.
    inversion H as [E | S].
    SCase "E".
      rewrite <- E.
      apply set_add_intro. tauto.
    SCase "S".
    value_cases (destruct v) SSCase;
      try (apply IHls ; trivial; fail).
      SSCase "vadd".
        apply set_add_intro.
          right. apply IHls. trivial.
Qed.

Lemma addrs_from_locals_values : forall (l1 l2 : locals),
  subset (values_from_locals l1) (values_from_locals l2) ->
    subset (addrs_in_locals l1) (addrs_in_locals l2).
Proof.
  intros l1 l2 H.
  unfold subset.
  intros a S.
  apply value_in_locals_addr.
  apply addr_in_locals_value in S.
  apply H.
  trivial.
Qed.


Lemma addrs_from_actors_correct1: forall (acts : actors)(a: addr),
  set_In a (addrs_from_actors acts) -> 
    exists (e:aid)(l:locals), set_In (e,l) acts /\ set_In a (addrs_in_locals l).
Proof.
  intros acts.
  induction acts as [ | (g,lg) actss] ; intros a H.
  Case "acts = nil".
    simpl in H. tauto.
  Case "acts = (g,lg)::actss".
    simpl in H.
    apply set_union_elim in H.
    inversion H as [Hg | HI].
    SCase "Hg".
      exists g. exists lg.
      split. simpl. left. trivial.
      unfold addrs_in_actor in Hg. simpl in Hg. trivial.
    SCase "HI".
      apply IHactss in HI.
      inversion HI as [e [l P]].
      exists e. exists l. 
      split. simpl. right. tauto.
        tauto.
Qed.

Lemma addrs_from_actors_correct2 : 
 forall (acts : actors)(a: addr)(e : aid) (l: locals),
  set_In (e,l) acts ->
    set_In a (addrs_in_locals l) ->
      set_In a (addrs_from_actors acts).
Proof.
  intros acts.
  induction acts as [ | (g,lg) actss]; intros a e l H G.
  Case "acts = nil".
    inversion H.
  Case "acts = (g,lg)::actss".
    simpl.
    apply set_union_intro.
    simpl in H.
    inversion H as [E | S].
      SCase "E".
        inversion E. subst.
        left. unfold addrs_in_actor. simpl. trivial.
      SCase "S".
        right. apply IHactss with (e:= e) (l := l). trivial. trivial.
Qed.

Lemma addrs_from_actors_correct3: forall (acts : actors)(e : aid) (l : locals),
  set_In (e,l) acts ->
    subset (addrs_in_locals l) (addrs_from_actors acts).
Proof.
  intros acts.
  induction acts as [ | (g,lg) actss]; intros e l H.
  Case "acts = nil".
    inversion H.
  Case "acts = (g,lg)::actss".
    simpl.
    unfold addrs_in_actor.
    simpl.
    simpl in H.
    inversion H as [E | S].
      inversion E. subst. apply union_included. 
      apply IHactss in S.
      apply subset_trans with 
      (y := set_union Aeq_addr (addrs_in_locals lg)(addrs_in_locals l)).
        apply union_included2.
        apply union_included_union.
          apply subset_refl.
          trivial.
Qed.


Lemma sub_values_sub_addrs_in_locals : forall (l : locals)(o : object),
  subset (values_in_table fid value Aeq_value o) (values_from_locals l) ->
   subset (addrs_in_fields o) (addrs_in_locals l).
Proof.
  intros l o H.
  unfold subset.
  intros a S.
  apply value_in_locals_addr.
  apply addrs_in_fields_correct1 in S.
  apply H.
  inversion S as [f I].
  apply values_in_table_correct1 with (n := f).
  trivial.
Qed.

(*
  lookup_heap retrieves the object associated with an object reference in a store.
*)

Fixpoint lookup_heap (s : heap) (r : addr) : option object :=
  match s with
  | nil => None
  | (e,o):: ss => if (Aeq_addr r e) then Some o else lookup_heap ss r
  end.


Fixpoint lookup_owner (ows : owners) (r : addr) : option (option aid) :=
  match ows with
  | nil => None
  | (e,o) :: owss => if (Aeq_addr e r) then Some o else lookup_owner owss r
  end.


Lemma lookup_owner_tlookup :
  forall (owns : owners)(a : addr),
    lookup_owner owns a = 
      tlookup addr (option aid) Aeq_addr a owns.
Proof.
  intros owns.
  induction owns as [ | (u, ou) ownss]; intros a.
  Case "owns = nil".
    simpl. trivial.
  Case "owns = (u,ou)::ownss".
    simpl.
    destruct (Aeq_addr a u).
    SCase "a = u".
      subst. destruct (Aeq_addr u u). trivial. 
      apply ex_falso_quodlibet. apply n. trivial.
    SCase "a <> u".
      destruct (Aeq_addr u a).
      apply ex_falso_quodlibet.  apply n.  auto. 
      apply IHownss.
Qed.

Fixpoint lookup_actor (acts : actors) (a : aid) : option actor :=
  match acts with
  | nil => None
  | (e,l):: actss => if (Aeq_aid a e) then Some (e,l) else lookup_actor actss a
  end.


Lemma lookup_actor_up : forall (acts : actors) (a : aid) (l : locals),
  set_In a (aids_from_actors acts) ->
    lookup_actor (up_actors a l acts) a = Some (a,l).
Proof.
  intros acts.
  induction acts as [ | (u,lu) actss]; intros a l S.
  Case "acts = nil".
    simpl. simpl in S. tauto.
  Case "acts = (u,lu) :: actss".
    simpl. simpl in S. apply set_add_elim in S.
    inversion S as [E | I].
    SCase "E".
      subst. destruct (Aeq_aid u u).
        simpl. destruct (Aeq_aid u u). trivial.
        apply ex_falso_quodlibet. apply n. trivial.
        apply ex_falso_quodlibet. apply n. trivial.
    SCase "I".
      destruct (Aeq_aid a u).
      SSCase "a = u".
        simpl. destruct (Aeq_aid a a). trivial.
        apply ex_falso_quodlibet. apply n. trivial.
      SSCase "a <> u".
        simpl. destruct (Aeq_aid a u). subst. 
        apply ex_falso_quodlibet. apply n. trivial.
        apply IHactss. trivial.
Qed.


(*
  A well-formed heap is a heap where any object reference (addr) r has only a single
  entry in the heap's store.
*)

Definition wf_heap (s : heap) : Prop := 
    forall (a: addr) (o1 o2 : object),
     set_In (a,o1) s -> set_In (a, o2) s -> o1 = o2.

Lemma wf_heap_table : forall (s : heap),
  wf_heap s <-> wf_table addr object s.
Proof.
  intros s.
  split ; unfold wf_table; unfold wf_heap; trivial.
Qed.

Definition complete_heap (h : heap) : Prop :=
  forall (a b: addr) (o : object),
    set_In (a,o) h -> 
      set_In b (addrs_in_fields o) ->
        set_In b (addrs_from_heap h).

Definition wfc_heap (h : heap) : Prop :=
  wf_heap h /\ complete_heap h.


Definition wf_locals (l : locals) : Prop := wf_table vid value l.

Definition wf_actor (a : actor) : Prop := wf_locals (snd a).

Definition wf_actors (acts : actors) : Prop := wf_table aid locals acts.

Definition wfc_actors (acts : actors) : Prop :=
  wf_actors acts /\
    (forall (a : actor), set_In a acts -> wf_actor a).

Definition wf_owners (owns : owners) : Prop := wf_table addr (option aid) owns.

Definition wf_object (o : object) : Prop := wf_table fid value o.

Definition wfcc_heap (h : heap) : Prop :=
  wfc_heap h /\
  (forall (a: addr) (o: object), set_In (a,o) h -> wf_object o).

Definition wf_queue (q : queue) : Prop :=
  forall (m: msgid)(a1 a2: addr),
    set_In (m,a1) q -> set_In (m,a2) q -> a1 = a2.

Definition wf_mbox (m : mbox) : Prop := wf_queue (msgs m).

Definition wf_mboxes (mbxs : mailboxes) : Prop := wf_table mid mbox mbxs.

Definition wfc_mboxes (mbxs : mailboxes) : Prop :=
  wf_mboxes mbxs /\ 
    (forall (p : mid * mbox), set_In p mbxs -> wf_mbox (snd p)).


Lemma addrs_from_actors_up1 : forall (acts : actors)(a :aid)
                               (l : locals) (i : vid) (v: value),
  wf_actors acts ->
   set_In (a,l) acts -> 
    not (exists (a : addr), v = vadd a) ->
     subset (addrs_from_actors (up_actors a (up_locals i v l) acts))
            (addrs_from_actors acts).
Proof.
  intros acts.
  induction acts as [ | (e,l') actss]; intros a l i v W S E.
  Case "acts = nil".
    simpl. apply subset_refl.
  Case "acts = (e,l')::actss".
    value_cases (destruct v) SCase.
    SCase "vnull".
      simpl. destruct (Aeq_aid a e).
      SSCase "a = e".
        subst.
        assert (L: l' = l).
          unfold wf_actors in W. unfold wf_table in W.
          apply W with (n:= e).
          simpl. tauto.
          simpl. inversion S.
            tauto. tauto.
        subst. simpl.
        apply union_included_union.
          unfold addrs_in_actor. simpl. apply addrs_in_locals_up1.
            intros C. inversion C as [c Cc]. inversion Cc.
          destruct (set_In_dec Aeq_actor (e,l) actss).
          SSSCase "(e,l) in actss".
            apply IHactss.
            unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons with (e := (e,l)). trivial.
            trivial.
            trivial.
          SSSCase "(e,l) not in actss".
            unfold up_actors. rewrite up_table_correct3. apply subset_refl.
            intros C. apply n. apply indexes_in_table_correct2 in C.
            inversion C as [k Ck].
            assert (K : k = l).
              unfold wf_actors in W. unfold wf_table in W.
              apply W with (n := e).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.  
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss.
            unfold wf_actors. unfold wf_actors in W. 
              apply wf_table_cons with (e:= (e,l')).
              trivial.
            simpl in S. inversion S as [Se | Sl].
              inversion Se. apply ex_falso_quodlibet. apply n. auto.
              trivial.
            trivial.
    SCase "vmark".
      simpl. destruct (Aeq_aid a e).
      SSCase "a = e".
        subst.
        assert (L: l' = l).
          unfold wf_actors in W. unfold wf_table in W.
          apply W with (n:= e).
          simpl. tauto.
          simpl. inversion S.
            tauto. tauto.
        subst. simpl.
        apply union_included_union.
          unfold addrs_in_actor. simpl. apply addrs_in_locals_up1.
            intros C. inversion C as [c Cc]. inversion Cc.
          destruct (set_In_dec Aeq_actor (e,l) actss).
          SSSCase "(e,l) in actss".
            apply IHactss.
            unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons with (e := (e,l)). trivial.
            trivial.
            trivial.
          SSSCase "(e,l) not in actss".
            unfold up_actors. rewrite up_table_correct3. apply subset_refl.
            intros C. apply n. apply indexes_in_table_correct2 in C.
            inversion C as [k Ck].
            assert (K : k = l).
              unfold wf_actors in W. unfold wf_table in W.
              apply W with (n := e).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.  
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss.
            unfold wf_actors. unfold wf_actors in W. 
              apply wf_table_cons with (e:= (e,l')).
              trivial.
            simpl in S. inversion S as [Se | Sl].
              inversion Se. apply ex_falso_quodlibet. apply n. auto.
              trivial.
            trivial.
    SCase "vnat".
      simpl. destruct (Aeq_aid a e).
      SSCase "a = e".
        subst.
        assert (L: l' = l).
          unfold wf_actors in W. unfold wf_table in W.
          apply W with (n:= e).
          simpl. tauto.
          simpl. inversion S.
            tauto. tauto.
        subst. simpl.
        apply union_included_union.
          unfold addrs_in_actor. simpl. apply addrs_in_locals_up1.
            intros C. inversion C as [c Cc]. inversion Cc.
          destruct (set_In_dec Aeq_actor (e,l) actss).
          SSSCase "(e,l) in actss".
            apply IHactss.
            unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons with (e := (e,l)). trivial.
            trivial.
            trivial.
          SSSCase "(e,l) not in actss".
            unfold up_actors. rewrite up_table_correct3. apply subset_refl.
            intros C. apply n0. apply indexes_in_table_correct2 in C.
            inversion C as [k Ck].
            assert (K : k = l).
              unfold wf_actors in W. unfold wf_table in W.
              apply W with (n := e).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.  
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss.
            unfold wf_actors. unfold wf_actors in W. 
              apply wf_table_cons with (e:= (e,l')).
              trivial.
            simpl in S. inversion S as [Se | Sl].
              inversion Se. apply ex_falso_quodlibet. apply n0. auto.
              trivial.
            trivial.
    SCase "vbool".
      simpl. destruct (Aeq_aid a e).
      SSCase "a = e".
        subst.
        assert (L: l' = l).
          unfold wf_actors in W. unfold wf_table in W.
          apply W with (n:= e).
          simpl. tauto.
          simpl. inversion S.
            tauto. tauto.
        subst. simpl.
        apply union_included_union.
          unfold addrs_in_actor. simpl. apply addrs_in_locals_up1.
            intros C. inversion C as [c Cc]. inversion Cc.
          destruct (set_In_dec Aeq_actor (e,l) actss).
          SSSCase "(e,l) in actss".
            apply IHactss.
            unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons with (e := (e,l)). trivial.
            trivial.
            trivial.
          SSSCase "(e,l) not in actss".
            unfold up_actors. rewrite up_table_correct3. apply subset_refl.
            intros C. apply n. apply indexes_in_table_correct2 in C.
            inversion C as [k Ck].
            assert (K : k = l).
              unfold wf_actors in W. unfold wf_table in W.
              apply W with (n := e).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.  
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss.
            unfold wf_actors. unfold wf_actors in W. 
              apply wf_table_cons with (e:= (e,l')).
              trivial.
            simpl in S. inversion S as [Se | Sl].
              inversion Se. apply ex_falso_quodlibet. apply n. auto.
              trivial.
            trivial.
    SCase "vadd".
      apply ex_falso_quodlibet. apply E. exists a0. trivial.
    SCase "vaid".
      simpl. destruct (Aeq_aid a e).
      SSCase "a = e".
        subst.
        assert (L: l' = l).
          unfold wf_actors in W. unfold wf_table in W.
          apply W with (n:= e).
          simpl. tauto.
          simpl. inversion S.
            tauto. tauto.
        subst. simpl.
        apply union_included_union.
          unfold addrs_in_actor. simpl. apply addrs_in_locals_up1.
            intros C. inversion C as [c Cc]. inversion Cc.
          destruct (set_In_dec Aeq_actor (e,l) actss).
          SSSCase "(e,l) in actss".
            apply IHactss.
            unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons with (e := (e,l)). trivial.
            trivial.
            trivial.
          SSSCase "(e,l) not in actss".
            unfold up_actors. rewrite up_table_correct3. apply subset_refl.
            intros C. apply n. apply indexes_in_table_correct2 in C.
            inversion C as [k Ck].
            assert (K : k = l).
              unfold wf_actors in W. unfold wf_table in W.
              apply W with (n := e).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.  
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss.
            unfold wf_actors. unfold wf_actors in W. 
              apply wf_table_cons with (e:= (e,l')).
              trivial.
            simpl in S. inversion S as [Se | Sl].
              inversion Se. apply ex_falso_quodlibet. apply n. auto.
              trivial.
            trivial.
    SCase "vmid".
      simpl. destruct (Aeq_aid a e).
      SSCase "a = e".
        subst.
        assert (L: l' = l).
          unfold wf_actors in W. unfold wf_table in W.
          apply W with (n:= e).
          simpl. tauto.
          simpl. inversion S.
            tauto. tauto.
        subst. simpl.
        apply union_included_union.
          unfold addrs_in_actor. simpl. apply addrs_in_locals_up1.
            intros C. inversion C as [c Cc]. inversion Cc.
          destruct (set_In_dec Aeq_actor (e,l) actss).
          SSSCase "(e,l) in actss".
            apply IHactss.
            unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons with (e := (e,l)). trivial.
            trivial.
            trivial.
          SSSCase "(e,l) not in actss".
            unfold up_actors. rewrite up_table_correct3. apply subset_refl.
            intros C. apply n. apply indexes_in_table_correct2 in C.
            inversion C as [k Ck].
            assert (K : k = l).
              unfold wf_actors in W. unfold wf_table in W.
              apply W with (n := e).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.  
      SSCase "a <> e".
        simpl.
        apply union_included_union.
          apply subset_refl.
          apply IHactss.
            unfold wf_actors. unfold wf_actors in W. 
              apply wf_table_cons with (e:= (e,l')).
              trivial.
            simpl in S. inversion S as [Se | Sl].
              inversion Se. apply ex_falso_quodlibet. apply n. auto.
              trivial.
            trivial.
Qed.
   
Lemma addrs_from_actors_up2 : forall (acts : actors)(e: aid)
                                (l : locals) (i : vid) (a : addr),
  wf_actors acts ->
   set_In (e,l) acts -> 
    subset (addrs_from_actors (up_actors e (up_locals i (vadd a) l) acts))
           (set_add Aeq_addr a (addrs_from_actors acts)).
Proof.
  intros acts.
  induction acts as [ | (g,lg) actss]; intros e l i a W S.
  Case "acts = nil".
    inversion S.
  Case "acts = (g;lg)::actss".
    simpl.
    destruct (Aeq_aid e g).
    SCase "e = g".
      subst. simpl.
      assert (L : lg = l).
        unfold wf_actors in W. unfold wf_table in W.
        apply W with (n:= g).
          simpl. tauto.
          trivial.
      subst.
      apply subset_trans with 
       (y := set_union Aeq_addr (set_add Aeq_addr a (addrs_in_actor (g, l)))
                                (set_add Aeq_addr a (addrs_from_actors actss))).
      apply union_included_union.
        unfold addrs_in_actor. simpl. apply addrs_in_locals_up2.
      destruct (set_In_dec Aeq_actor (g,l) actss).
      SSCase "(g,l) in actss".
        apply IHactss.
          unfold wf_actors. unfold wf_actors in W.
           apply wf_table_cons with (e := (g,l)).
           trivial.
          trivial.
      SSCase "(g,l) not in actss".
        unfold up_actors.
        rewrite up_table_correct3.
          apply add_included_sub. apply subset_refl.
          intros C. apply n. apply indexes_in_table_correct2 in C.
            inversion C as [l' Cl'].
            unfold wf_actors in W. unfold wf_table in W.
            assert (L : l' = l).
              apply W with (n := g).
              simpl. tauto.
              simpl. tauto.
            subst. trivial.
      apply union_add_add_union.
    SCase "e <> g".
      simpl.
      apply subset_trans with
       (y := set_union Aeq_addr (set_add Aeq_addr a (addrs_in_actor (g,lg)))
                                (set_add Aeq_addr a (addrs_from_actors actss))).
       SSCase "branche 1".
         apply union_included_union.
           apply add_included_sub. apply subset_refl.
           apply IHactss.
             unfold wf_actors. unfold wf_actors in W.
              apply wf_table_cons in W. trivial.
             simpl in S. inversion S as [N | Y].
               inversion N. apply ex_falso_quodlibet. apply n. auto.
               trivial.
       SSCase "branche 2".
         apply union_add_add_union.
Qed.
    

Lemma lookup_actor_up_not : forall (acts : actors) (a b: aid) (l l': locals),
  wf_actors acts ->
   a <> b ->
    set_In (b,l') acts ->
      Some (b,l') = lookup_actor (up_actors a l acts) b.
Proof.
  intros acts.
  induction acts as [ | (u,lu) actss]; intros a b l l' W N S.
  Case "acts = nil".
    inversion S.
  Case "acts = (u,lu)::actss".
    simpl in S.
    inversion S as [E | I].
    SCase "E".
      inversion E. subst.
      simpl. destruct (Aeq_aid a b).
        apply ex_falso_quodlibet. apply N. trivial.
        simpl. destruct (Aeq_aid b b). trivial.
          apply ex_falso_quodlibet. apply n0. trivial.
    SCase "I".
      simpl. destruct (Aeq_aid a u).
        simpl. destruct (Aeq_aid b a). subst.
          apply ex_falso_quodlibet. apply N. trivial.
          apply IHactss. 
            unfold wf_actors. apply wf_table_cons with (e:= (u,lu)). trivial.
            auto.
            trivial.
        simpl. destruct (Aeq_aid b u).
          assert (A : lu = l').
            unfold wf_actors in W. unfold wf_table in W.
            apply W with (n := b). subst.
            simpl. tauto. subst. simpl. tauto.
          subst. trivial.
          apply IHactss.
            unfold wf_actors. apply wf_table_cons with (e:= (u,lu)). trivial.
            trivial.
            trivial.
Qed.

Lemma lookup_actor_elim : forall (acts : actors) (a e: aid) (l: locals),
  Some (e,l) = lookup_actor acts a -> e = a /\ set_In (e,l) acts.
Proof.
  intros acts.
  induction acts as [ | (u,lu) actss]; intros a e l H.
  Case " acts = nil".
    simpl. simpl in H. inversion H.
  Case " acts = (u,lu::actss".
    simpl. simpl in H.
    destruct (Aeq_aid a u).
    SCase "a = u".
      subst. inversion H. subst. tauto.
    SCase "a <> u".
      apply IHactss in H. tauto.
Qed.

Lemma lookup_actor_intro : forall (acts : actors)(a: aid)(l : locals),
  wf_actors  acts -> 
    set_In (a,l) acts -> 
      lookup_actor acts a = Some (a,l).
Proof.
  intros acts.
  induction acts as [ | (u,lu) actss]; intros a l W H.
  Case "acts = nil".
    inversion H.
  Case "acts = (u,lu)::actss".
    simpl.
    simpl in H.
    inversion H as [E | S].
      inversion E. subst. 
      destruct (Aeq_aid a a).
          SCase "a = u".
            trivial.
            apply ex_falso_quodlibet. apply n. trivial.
          SCase "a <> u".
            destruct (Aeq_aid a u).
              subst. assert (A: lu = l).
                unfold wf_actors in W. unfold wf_table in W. 
                apply W with (n := u). simpl. tauto.
                simpl. tauto.
              subst. trivial.
            apply IHactss.
              unfold wf_actors. apply wf_table_cons with (e := (u,lu)). trivial.
              trivial.
Qed.

