(**
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: June 2013
  (c) INRIA
*)


(**
  A simple abstraction for finite dictionary.
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import SiaamUtils.

Section Table.

Variable J : Type.
Variable A : Type.
Hypothesis Jeq_dec : forall (n m : J), {n = m} + {n <> m}.
Hypothesis Aeq_dec : forall (n m : A), {n = m} + {n <> m}.

(* 
  An entry is pair <index, value>.
*)

Definition tentry := prod J A.

(* 
  A table is a list of entries.
*)

Definition table  := set tentry.

(* 
  A well-formed table has only one entry value per index.
*)

Definition wf_table (t : table) : Prop :=
  forall (n : J) (v1 v2 : A),
   set_In (n,v1) t -> set_In (n,v2) t -> v1 = v2.


Lemma wf_table_cons : forall (t : table) (e : tentry),
  wf_table (e::t) -> wf_table t.
Proof.
  intros t e. unfold wf_table. intros H.
  intros n v1 v2 I1 I2.
  apply H with (n := n).
  simpl. right. trivial.
  simpl. right. trivial.
Qed.

Lemma Aeq_tentry : forall (e1 e2 : tentry),
  {e1 = e2} + {e1 <> e2}.
Proof.
  intros e1 e2.
  destruct e1 as [ n1 v1].
  destruct e2 as [n2 v2].
  destruct (Jeq_dec n1 n2).
  Case "n1 = n2".
    subst. destruct (Aeq_dec v1 v2).
      subst. left. trivial.
      right. intros C. apply n. inversion C. trivial.
  Case "n1 <> n2".
    right. intros C. apply n. inversion C. trivial.
Qed.


(*
  Operation on tables: update and lookup.
*)

Fixpoint up_table (n : J) (v : A)(t : table): table :=
  match t with
  | [] => []
  | (m,w)::ts => if Jeq_dec n m
                 then (n, v) :: (up_table n v ts) 
                 else (m,w):: (up_table n v ts)
  end.


Fixpoint tlookup (n : J) (t : table) : option A :=
  match t with
  | [] => None
  | (m,v):: ts => if Jeq_dec n m
                  then Some v
                  else tlookup n ts
  end.

(*
  Lemmas concerning indexes.
*)

Fixpoint indexes_in_table (t : table) : set J :=
  match t with
  | [] => []
  | (n,v)::ts => set_add Jeq_dec n(indexes_in_table ts)
  end.

Fixpoint values_in_table (t : table) : set A :=
  match t with
  | [] => []
  | (n,v)::ts => set_add Aeq_dec v (values_in_table ts)
  end.


Lemma indexes_in_table_correct1 : forall (t : table)(n : J) (v : A),
  set_In (n,v) t -> set_In n (indexes_in_table t).
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n v H.
  Case "t = nil".
    simpl. trivial.
  Case "t = (m,w)::ts".
    simpl in H.
    inversion H as [E | S].
      inversion E. subst. simpl. apply set_add_intro. left. trivial.
      simpl. apply set_add_intro. right. apply (IHts n v). trivial.
Qed.

Lemma indexes_in_table_correct2 : forall ( t: table) (n : J),
  set_In n (indexes_in_table t) -> exists (v: A), set_In (n,v) t.
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n H.
  Case "t = nil".
    simpl in H. tauto.
  Case "t = (m,w)::ts".
    simpl in H.
    apply set_add_elim in H.
    inversion H as [E | S].
      subst. exists w. simpl. left. trivial.
      simpl. apply IHts in S. inversion S as [v Sv].
       exists v. right. trivial.
Qed.


Lemma indexes_in_up_table1: forall (t : table)(i : J) (v : A),
  subset (indexes_in_table t) (indexes_in_table (up_table i v t)).
Proof.
  intros t.
  induction t as [ | (j,w) ts]; intros i v.
  Case "t = nil".
    simpl. unfold subset. trivial.
  Case "t = (j,w)::ts".
    simpl.
    remember (Jeq_dec i j) as eij.
    destruct eij.
    SCase "i = j".
      subst. simpl.
      unfold subset.
      intros e H.
      apply set_add_intro.
      apply set_add_elim in H.
      inversion H as [E | S].
        left. trivial.
        right. apply IHts. trivial.
    SCase "i <> j".
      simpl.
      unfold subset.
      intros e H.
      apply set_add_intro.
      apply set_add_elim in H.
      inversion H as [E | S].
        left. trivial.
        right. apply IHts. trivial.
Qed.

Lemma indexes_in_up_table11 : forall (t : table) (i j: J) (v: A),
  set_In j (indexes_in_table t) ->
    set_In j (indexes_in_table (up_table i v t)).
Proof.
  intros t i j v H.
  assert (S : subset (indexes_in_table t) (indexes_in_table (up_table i v t))).
    apply indexes_in_up_table1.
  unfold subset in S.
  apply S.
  trivial.
Qed.

Lemma indexes_in_up_table2: forall (t : table)(i : J) (v : A),
  subset (indexes_in_table (up_table i v t)) (indexes_in_table t) .
Proof.
  intros t.
  induction t as [ | (j,w) ts]; intros i v.
  Case "t = nil".
    simpl. unfold subset. trivial.
  Case "t = (j,w)::ts".
    simpl.
    remember (Jeq_dec i j) as eij.
    destruct eij.
    SCase "i = j".
      subst. simpl.
      unfold subset.
      intros e H.
      apply set_add_intro.
      apply set_add_elim in H.
      inversion H as [E | S].
        left. trivial.
        right. unfold subset in IHts. apply IHts with (i:= j)(v := v). trivial.
    SCase "i <> j".
      simpl.
      unfold subset.
      intros e H.
      apply set_add_intro.
      apply set_add_elim in H.
      inversion H as [E | S].
        left. trivial.
        right. unfold subset in IHts. apply IHts with (i:= i)(v := v). trivial.
Qed.

Lemma indexes_in_up_table22 : forall (t: table)(i j: J)(v: A),
    set_In j (indexes_in_table (up_table i v t)) ->
      set_In j (indexes_in_table t).
Proof.
  intros t i j v H.
  assert (S : subset (indexes_in_table (up_table i v t)) (indexes_in_table t)).
    apply indexes_in_up_table2.
  unfold subset in S.
  apply S.
  trivial.
Qed.

Lemma indexes_in_add_table1 : 
  forall (t : table)
         (H: forall (e f : tentry), {e = f} + {e <> f})
         (n : J) (v:A),
  subset (indexes_in_table t)
         (indexes_in_table (set_add H (n,v) t)).
Proof.
  intros t.
  unfold subset.
  induction t as [ | (m,w) ts]; intros H n v.
  Case "t = nil".
    simpl. intros a F. tauto.
  Case "t = (m,w):: ts".
    intros a HS.
    simpl. 
    simpl in HS.
    apply set_add_elim in HS.
    inversion HS as [E | S].
    SCase "E".
      subst.
      destruct (H (n,v)(m,w)).
        inversion e. subst. simpl. apply set_add_intro. left. trivial.
        simpl. apply set_add_intro. left. trivial.
    SCase "S".
      destruct (H (n,v)(m,w)).
        inversion e.  subst. simpl. apply set_add_intro. right. trivial.
        simpl. apply set_add_intro.
          right. apply IHts. trivial.
Qed.
  
Lemma indexes_in_add_table2 : 
  forall (t : table)
         (H: forall (e f : tentry), {e = f} + {e <> f})
         (n : J) (v:A),
  subset (indexes_in_table (set_add H (n,v) t))
         (set_add Jeq_dec n (indexes_in_table t)).
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros H n v.
  Case " t = nil".
    simpl. apply subset_refl.
  Case " t = (m,w):: ts".
    simpl.
    destruct (H (n,v) (m,w)).
    SCase "(n,v) = (m,w)".
      inversion e. subst. simpl.
      apply add_included.
      apply add_included_sub.
      apply subset_refl.
    SCase "(n,v) <> (m,w)".
      simpl.
      apply subset_trans with 
      (y := set_add Jeq_dec m (set_add Jeq_dec n (indexes_in_table ts))).
        apply add_included. apply IHts.
        apply add_add_com. apply subset_refl.
Qed.

Lemma indexes_in_add_table3 : 
  forall (t : table)
         (H: forall (e f : tentry), {e = f} + {e <> f})
         (n : J) (v:A),
  subset (set_add Jeq_dec n (indexes_in_table t))
         (indexes_in_table (set_add H (n,v) t)).
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros H n v.
  Case "t = nil".
    simpl. apply subset_refl.
  Case "t = (m,w)::ts".
    simpl.
    destruct (H (n,v) (m,w)).
    SCase "(n,v) = (m,w)".
      inversion e. subst. simpl.
      apply subset_trans with (y := set_add Jeq_dec m (indexes_in_table ts)).
        unfold subset. intros a Ia.
        apply set_add_intro.
        apply set_add_elim in Ia.
        inversion Ia as [Ea | Sa].
          subst. tauto.
          apply set_add_elim in Sa. trivial.
        apply subset_refl.
    SCase "(n,v <> (m,w)".
      simpl.
      apply subset_trans with 
      (y := set_add Jeq_dec m (set_add Jeq_dec n (indexes_in_table ts))).
        apply add_add_com. apply subset_refl.
        apply add_included. apply IHts.
Qed.
        


Lemma indexes_in_add_table : forall (t : table) (n : J) (v:A),
  subset (indexes_in_table t)
         (indexes_in_table (set_add Aeq_tentry (n,v) t)).
Proof.
  intros t.
  unfold subset.
  induction t as [ | (m,w) ts]; intros n v.
  Case "t = nil".
    simpl. intros a F. tauto.
  Case "t = (m,w):: ts".
    intros a H.
    simpl. 
    simpl in H.
    apply set_add_elim in H.
    inversion H as [E | S].
    SCase "E".
      subst.
      destruct (Aeq_tentry (n,v)(m,w)).
        inversion e. subst. simpl. apply set_add_intro. left. trivial.
        simpl. apply set_add_intro. left. trivial.
    SCase "S".
      destruct (Aeq_tentry (n,v)(m,w)).
        inversion e.  subst. simpl. apply set_add_intro. right. trivial.
        simpl. apply set_add_intro.
          right. apply IHts. trivial.
Qed.
  

(*
  Lemmas concerning lookup.
*)

Lemma tlookup_correct1 : forall (t : table)(n : J)(v : A),
  tlookup n t = Some v -> set_In (n,v) t.
Proof.
  intros t. induction t as [ | e ts]; intros n v H.
  Case "t = nil".
    simpl. simpl in H. inversion H.
  Case "t = e::ts".
    destruct e as [m w].
    simpl in H.
    remember (Jeq_dec n m) as b.
    destruct b.
    SCase "n = m".
      inversion H. subst. 
      simpl. left. trivial.
    SCase "n <> m".
      simpl. right.
      apply IHts. trivial.
Qed.

Lemma tlookup_correct2 : forall (t : table)(n : J)(v:A),
  wf_table t -> set_In (n,v) t -> tlookup n t = Some v.
Proof.
  intros t. induction t as [ | (m,w) ts]; intros n v W I.
  Case "t = nil".
    simpl in I. tauto.
  Case "t = (m,w)::ts".
    simpl. remember (Jeq_dec n m) as b.
    destruct b.
    SCase "n = m".
      subst.
      assert (E: w = v).
        unfold wf_table in W. apply W with (n:= m).
        simpl. left. trivial.
        trivial.
      subst. trivial.
    SCase "n<> m".
      apply IHts.
      apply wf_table_cons with (e:= (m,w)). trivial.
      simpl in I. inversion I as [E | In].
        inversion E. apply ex_falso_quodlibet. apply n0. auto.
        trivial.
Qed.
    
Lemma tlookup_correct3 : forall (t: table)(n : J)(v : A),
  tlookup n t = Some v -> set_In n (indexes_in_table t).
Proof.
  intros t. 
  induction t as [ | (m,w) ts]; intros n v H.
  Case "t = nil".
    simpl in H. inversion H.
  Case "t = (m,w)::ts".
    simpl. simpl in H.
    remember (Jeq_dec n m) as b.
    destruct b.
    SCase "n = m".
      apply set_add_intro. left. trivial.
    SCase "n <> m".
      apply set_add_intro. right. apply IHts with (v := v). trivial.
Qed.

Lemma tlookup_correct4: forall (t: table) (n : J),
  set_In n (indexes_in_table t) -> exists (v: A), tlookup n t = Some v.
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n H.
  Case "t = nil".
    simpl in H. tauto.
  Case "t = (m,w)::ts".
    simpl in H.
    simpl.
    remember (Jeq_dec n m) as b.
    destruct b.
    SCase "n = m".
      exists w. trivial.
    SCase "n <> m".
      apply set_add_elim in H. inversion H as [E | I].
        apply ex_falso_quodlibet. apply n0. auto.
        apply IHts. trivial.
Qed.

Lemma tlookup_correct5 : forall (t: table) (n : J),
  tlookup n t = None -> not (exists (v: A), set_In (n,v) t).
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n T.
  Case "t = nil".
    simpl. simpl in T. intros K. inversion K. trivial.
  Case " t = (m,w)::ts".
    intros K. simpl in T.
    inversion K as [v S].
    simpl in S.
    inversion S as [E | I].
      inversion E. subst. destruct (Jeq_dec n n).
        inversion T.
        apply n0. trivial.
      destruct (Jeq_dec n m).
        subst. inversion T.
        apply IHts in T. apply T. exists v. trivial.
Qed.

Lemma in_table_dec : forall (t : table) (n : J), 
{exists (v : A), set_In (n,v) t} + { not (exists (v: A), set_In (n,v) t)}.
Proof.
  intros t n.
  remember (tlookup n t) as L.
  destruct L.
    left. exists a. apply tlookup_correct1. auto.
    right. apply tlookup_correct5. auto.
Qed.


(*
  Lemmas concerning update.
*)

Lemma up_table_correct1 : forall (t : table)(n : J) (v : A),
  set_In n (indexes_in_table t) ->
    tlookup n (up_table n v t) = Some v.
Proof.
  intros t. 
  induction t as [ | (m,w) ts]; intros n v H.
  Case "t = nil".
    simpl in H. tauto.
  Case "t = (m,w)::ts".
    simpl in H. simpl.
    remember (Jeq_dec n m) as b.
    destruct b.
    SCase "n = m".
      simpl. subst. rewrite <- Heqb. trivial.
    SCase "n <> m".
      apply set_add_elim in H.
      inversion H as [E | I].
        apply ex_falso_quodlibet. apply n0. auto.
        simpl. rewrite <- Heqb. apply IHts. trivial.
Qed.

Lemma up_table_correct2 : forall (t : table) (n : J) (v : A),
  set_In n (indexes_in_table t) -> 
    set_In (n,v) (up_table n v t).
Proof.
  intros t n v H.
  apply tlookup_correct1.
  apply up_table_correct1.
  trivial.
Qed.

Lemma up_table_correct3 : forall (t : table) (n : J) (v : A),
  not (set_In n (indexes_in_table t)) -> up_table n v t = t.
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n v H.
  Case "t = nil".
    simpl. trivial.
  Case "t = (m,w)::ts".
    simpl.
    remember (Jeq_dec n m) as b.
    destruct b.
    SCase "n = m".
      simpl in H. apply ex_falso_quodlibet.
      apply H. subst. apply set_add_intro. left. trivial.
    SCase "n <> m".
      simpl in H.
      rewrite IHts. trivial.
      intros C.
      apply H. apply set_add_intro. right. trivial.
Qed.


Lemma up_table_correct4 : forall (t : table) (n m : J) (v w: A),
  n <> m -> set_In (m,w) (up_table n v t) -> set_In (m,w) t.
Proof.
  intros t.
  induction t as [ | (k,u) ts]; intros n m v w N I.
  Case "t = nil".
    simpl in I. inversion I.
  Case "t = (k,u)::ts".
    simpl.
    simpl in I.
    remember (Jeq_dec n k) as jnk.
    destruct jnk.
    SCase "n = k".
      subst. simpl in I.
      inversion I as [E | S].
        inversion E. subst. apply ex_falso_quodlibet. apply N. trivial.
        right. apply IHts with (n := k)(v := v). trivial. trivial.
    SCase "n <> k".
      simpl in I.
      inversion I as [E | S].
        left. trivial.
        right. apply IHts with (n := n)(v := v). trivial. trivial.
Qed.

Lemma up_table_correct5 : forall (t : table) (n : J) (v w: A),
  set_In (n,w) (up_table n v t) -> set_In n (indexes_in_table t).
Proof.
  intros t.
  induction t as [ | (m,u) ts]; intros n v w H.
  Case "t = nil".
    simpl in H. inversion H.
  Case "t = (m,u)::ts".
    simpl. apply set_add_intro.
    simpl in H.
    remember (Jeq_dec n m) as jnm.
    destruct jnm.
    SCase "n = m".
      left. trivial.
    SCase "n <> m".
      right. simpl in H.
      inversion H as [E | S].
        inversion E. subst.  apply ex_falso_quodlibet. apply n0. trivial.
        apply IHts with (v:= v)(w:= w). trivial.
Qed.

Lemma up_table_correct6 : forall (t: table) (n : J) (v w : A),
  set_In (n,w) (up_table n v t) -> w = v.
Proof.
  intros t.
  induction t as [ | (m,u) ts]; intros n v w H.
  Case "t = nil".
    simpl in H. inversion H.
  Case "t = (m,u)::ts".
    simpl in H.
    remember (Jeq_dec n m) as jnm.
    destruct jnm.
    SCase "n = m".
      simpl in H.
      inversion H as [E | S].
        inversion E. auto.
        apply IHts with (n:= n). trivial.
    SCase "n <> m".
      simpl in H.
      inversion H as [E | S].
        inversion E. subst.  apply ex_falso_quodlibet. apply n0. trivial.
        apply IHts with (n := n). trivial.
Qed.


Lemma up_table_correct7 : forall (t : table) (n : J) (v : A),
  wf_table t -> set_In (n,v) t -> up_table n v t = t.
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n v W H.
  Case "t = nil".
    simpl. trivial.
  Case "t = (m,w)::ts".
    simpl in H.
    inversion H as [E | S].
    SCase "(m,w) = (n,v)".
      inversion E. subst.
      simpl. destruct (Jeq_dec n n).
        assert (Its : {set_In (n,v) ts} + {~ set_In (n,v) ts}).
          apply set_In_dec. intros x y.
          destruct x as [i1 v1].
          destruct y as [i2 v2].
          destruct (Jeq_dec i1 i2).
            destruct (Aeq_dec v1 v2).
              subst. left. trivial.
                     right. intros C. apply n0. inversion C. trivial.
            right. intros C. apply n0. inversion C. trivial.
       destruct Its.
         SSCase "(n,v) in ts".
           rewrite IHts. trivial. apply (wf_table_cons ts (n,v)). trivial. trivial.
         SSCase "(n,v) not in ts".
           assert (Nts: ~ set_In n (indexes_in_table ts)).
             intros C. apply n0. apply indexes_in_table_correct2 in C.
             inversion C as [w Sw].
             assert (SSw : set_In (n,w) ((n,v)::ts)).
               simpl. destruct (Aeq_dec v w).
                 subst. left. trivial.
                 right. trivial.
             unfold wf_table in W.
             apply (W n w v) in SSw. subst. trivial. simpl. left. trivial.
           apply up_table_correct3 with (v:= v) in Nts.
           rewrite Nts. trivial.
           apply ex_falso_quodlibet. apply n0. trivial.
    SCase "(n,v) in ts".
      simpl.
      remember (Jeq_dec n m) as Jnm.
      destruct Jnm.
      SSCase "n = m".
        subst.
        assert (Evw : w = v).
          unfold wf_table in W. apply W with (n := m).
          simpl. left. trivial.
          simpl. right. trivial.
        subst. rewrite IHts. trivial. apply wf_table_cons with (e := (m,v)).
          trivial. trivial.
      SSCase "n <> m".
        rewrite IHts. trivial. apply wf_table_cons with (e := (m,w)). trivial.
        trivial.
Qed.
          

Lemma up_table_correct8 : forall (t: table)(n m : J) (v w : A),
  n <> m ->
    set_In (m,w) t ->
      set_In (m,w) (up_table n v t).
Proof.
  intros t.
  induction t as [ | (k,u) ts]; intros n m v w N H.
  Case "t = nil".
    inversion H.
  Case "t = (k,u)::ts".
    simpl in H.
    simpl.
    inversion H as [E | S].
    SCase "E".
      inversion E. subst.
      destruct (Jeq_dec n m).
      SSCase "n = m".
        apply ex_falso_quodlibet. apply N. trivial.
      SSCase "n <> m".
        simpl. left. trivial.
    SCase "S".
      destruct (Jeq_dec n k).
        simpl. right. apply IHts. trivial. trivial.
        simpl. right. apply IHts. trivial. trivial.
Qed.

Lemma up_table_elim : forall (t : table) (n m : J) (v w : A),
  set_In (m,w) (up_table n v t) -> (m,w) = (n,v) \/ set_In (m,w) t.
Proof.
  intros t n m v w H.
  destruct (Jeq_dec m n).
  Case " m= n".
    subst. left. apply up_table_correct6 in H. subst. trivial.
  Case "m <> n".
    right. apply up_table_correct4 with (n := n) (v := v).
      auto. trivial.
Qed.


(* 
  Lemmas concerning values.
*)

Lemma values_in_table_correct1 : forall (t: table) (n : J) (v : A),
  set_In (n,v) t -> set_In v (values_in_table t).
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n v H.
  Case "t = nil".
    inversion H.
  Case "t = (m,w)::ts".
    simpl.
    apply set_add_intro.
    simpl in H.
    inversion H as [E | S].
      inversion E. left. trivial.
      right. apply IHts with (n := n). trivial.
Qed.

Lemma values_in_table_correct2: forall (t: table) (v: A),
  set_In v (values_in_table t) -> exists (n: J), set_In (n,v) t.
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros v H.
  Case "t = nil".
    simpl in H. tauto.
  Case "t = (m,w)::ts".
    simpl.
    simpl in H.
    apply set_add_elim in H.
    inversion H as [E | S].
      subst. exists m. left. trivial.
      apply IHts in S. inversion S as [n Sn]. exists n. right. trivial.
Qed.

(*
  Lemmas concerning well-formedness.
*)

Lemma not_in_wf_table_not_index : forall (t : table) (n : J) (v : A),
  wf_table ((n,v)::t) ->
    ~ set_In (n,v) t ->
      ~ set_In n (indexes_in_table t).
Proof.
  intros t.
  induction t as [ | (m,w) ts]; intros n v W H.
  Case "t = nil".
    simpl. intros C. trivial.
  Case "t = (m,w)::ts".
    intros C. apply H.
    apply indexes_in_table_correct2 in C.
    inversion C as [u Cu].
    assert (Euv : u = v).
      unfold wf_table in W.
      apply W with (n0 := n).
      simpl in Cu.
      inversion Cu as [E | S].
        inversion E. subst. simpl. right. left. trivial.
          simpl. right. right. trivial.
          simpl. left. trivial.
    subst. trivial.
Qed.

Lemma wf_up_table : forall (t : table) (n : J) (a : A),
  wf_table t -> wf_table (up_table n a t).
Proof.
  intros t n a.
  unfold wf_table.
  intros W.
  intros m w1 w2 I1 I2.
  remember (Jeq_dec n m) as jnm.
  destruct jnm.
  Case "n = m".
    subst. apply up_table_correct6 in I1.
    apply up_table_correct6 in I2. subst. trivial.
  Case "n <> m".
    apply W with (n := m).
      apply up_table_correct4 with (n := n)(v := a). trivial. trivial.
      apply up_table_correct4 with (n := n)(v := a). trivial. trivial.
Qed.


Lemma wf_add_table : forall (t: table) 
                            (H: forall (e1 e2 : tentry),{e1 = e2} + {e1 <> e2}) 
                            (e : tentry),
  wf_table t -> 
   ~(set_In (fst e) (indexes_in_table t)) -> 
     wf_table (set_add H e t).
Proof.
  intros t.
  induction t as [ | a ts]; intros H e W S.
  Case "t = nil".
    simpl. unfold wf_table.
    intros n v1 v2 H1 H2.
    simpl in H1.
    simpl in H2.
    destruct e as [ m w].
    inversion H1 as [E1 | F].
      inversion E1. inversion H2 as [E2 | F2].
        inversion E2. subst. trivial.
        tauto.
    tauto.
  Case "t = a::ts".
    simpl. 
    destruct (H e a).
    SCase "e = a".
      trivial.
    SCase "e <> a".
      assert (W1 : wf_table (a :: ts)). trivial.
      apply wf_table_cons in W.
      apply IHts with (H := H) (e:= e) in W .
      unfold wf_table.
      intros m w1 w2 H1 H2.
      simpl in H1.
      simpl in H2.
      inversion H1 as [E1 | S1].
        SSCase "E1".
          inversion H2 as [E2 | S2].
            destruct a as [k u]. inversion E1. inversion E2. subst. trivial.
            apply set_add_elim in S2. inversion S2 as [Ee | Se].
              apply ex_falso_quodlibet. apply S. rewrite <- Ee.
                rewrite E1. simpl. apply set_add_intro. left.  trivial.
              unfold wf_table in W1. apply W1 with (n := m).
                rewrite <- E1. simpl. left. trivial.
                simpl. right. trivial.
        SSCase "S1".
          apply set_add_elim in S1. inversion S1 as [Ee | Se].
            inversion H2 as [E2 | S2].
              apply ex_falso_quodlibet. apply S.
                rewrite <- Ee. rewrite E2. simpl. apply set_add_intro. left. trivial.
              unfold wf_table in W. apply W with (n := m).
                rewrite Ee. apply set_add_intro. left. trivial.
                trivial.
            inversion H2 as [E2 | S2].
              unfold wf_table in W1. apply W1 with (n := m). simpl. right. trivial.
                rewrite <- E2. simpl. left. trivial.
              unfold wf_table in W. apply W with (n:= m).
                apply set_add_intro. right. trivial. trivial.
      intros C. apply S. simpl. destruct a as [na va].
      apply set_add_intro. right. trivial.
Qed.

Lemma not_in_wf_table : forall (t : table)(n: J) (v: A),
  wf_table t -> 
    ~ (set_In n (indexes_in_table t)) -> 
      wf_table (set_add Aeq_tentry (n,v) t).
Proof.
  unfold wf_table.
  intros t.
  induction t as [ | (m,w) ts]; intros n v W H.
  Case "t = nil".
    simpl. intros m w1 w2.
    intros H1 H2.
    inversion H1 as [E1 | F1].
      inversion H2 as [E2 | F2].
        inversion E1. inversion E2. subst. trivial.
        tauto.
      tauto.
  Case "t = (m,w)::ts".
    intros k v1 v2 H1 H2.
    apply set_add_elim in H1.
    apply set_add_elim in H2.
    inversion H1 as [E1 | S1].
    SCase "E1".
      inversion H2 as [E2 | S2].
        inversion E1. inversion E2. subst. trivial.
        inversion E1. subst. apply ex_falso_quodlibet.
          apply H. apply indexes_in_table_correct1 with (v := v2). 
          trivial.
    SCase "S1".
      inversion H2 as [E2 | S2].
      SSCase "E2".
        apply ex_falso_quodlibet. apply H.
          apply indexes_in_table_correct1 with (v := v1).
          inversion E2. subst. trivial.
      SSCase "S2".
        simpl in S1.
        simpl in S2.
          inversion S1 as [E1 | Sk1].
          SSSCase "E1".
            inversion E1. subst. 
            inversion S2 as [E2 | Sk2].
            SSSSCase "E2".
              inversion E2. trivial.
            SSSSCase "Sk2".
              apply W with (n := k).
                simpl. left. trivial.
                simpl. right. trivial.
          SSSCase "Sk1".
            inversion S2 as [E2 | Sk2].
              inversion E2. subst. apply W with (n:= k).
                simpl. right. trivial.
                simpl. left. trivial.
              apply IHts with (n := n)(v := v)(n0 := k).
                intros i a1 a2 I1 I2. apply W with (n := i).
                  simpl. right. trivial.
                  simpl. right. trivial.
                intros C. apply H. simpl. apply set_add_intro. right. trivial.
                simpl. apply set_add_intro. right. trivial.
                simpl. apply set_add_intro. right. trivial.
Qed.


Lemma not_in_wf_table1 : 
  forall (t : table)
         (AH : forall (e f : tentry), {e = f} + {e <> f})
         (n: J) (v: A),
  wf_table t -> 
    ~ (set_In n (indexes_in_table t)) -> 
      wf_table (set_add AH (n,v) t).
Proof.
  unfold wf_table.
  intros t.
  induction t as [ | (m,w) ts]; intros AH n v W H.
  Case "t = nil".
    simpl. intros m w1 w2.
    intros H1 H2.
    inversion H1 as [E1 | F1].
      inversion H2 as [E2 | F2].
        inversion E1. inversion E2. subst. trivial.
        tauto.
      tauto.
  Case "t = (m,w)::ts".
    intros k v1 v2 H1 H2.
    apply set_add_elim in H1.
    apply set_add_elim in H2.
    inversion H1 as [E1 | S1].
    SCase "E1".
      inversion H2 as [E2 | S2].
        inversion E1. inversion E2. subst. trivial.
        inversion E1. subst. apply ex_falso_quodlibet.
          apply H. apply indexes_in_table_correct1 with (v := v2). 
          trivial.
    SCase "S1".
      inversion H2 as [E2 | S2].
      SSCase "E2".
        apply ex_falso_quodlibet. apply H.
          apply indexes_in_table_correct1 with (v := v1).
          inversion E2. subst. trivial.
      SSCase "S2".
        simpl in S1.
        simpl in S2.
          inversion S1 as [E1 | Sk1].
          SSSCase "E1".
            inversion E1. subst. 
            inversion S2 as [E2 | Sk2].
            SSSSCase "E2".
              inversion E2. trivial.
            SSSSCase "Sk2".
              apply W with (n := k).
                simpl. left. trivial.
                simpl. right. trivial.
          SSSCase "Sk1".
            inversion S2 as [E2 | Sk2].
              inversion E2. subst. apply W with (n:= k).
                simpl. right. trivial.
                simpl. left. trivial.
              apply IHts with (AH := AH)(n := n)(v := v)(n0 := k).
                intros i a1 a2 I1 I2. apply W with (n := i).
                  simpl. right. trivial.
                  simpl. right. trivial.
                intros C. apply H. simpl. apply set_add_intro. right. trivial.
                simpl. apply set_add_intro. right. trivial.
                simpl. apply set_add_intro. right. trivial.
Qed.





End Table.

