(**
  SIAAM abstract operational semantics and isolation properties.
  Author: Jean-Bernard Stefani
  Date: June 2013
  (c) INRIA
*)

(**
  This file gathers various lemmas about (finite) sets, represented as lists, as per the 
  ListSet module of the standard library.
*)


Require Import Arith.
Require Import Lists.List.
Require Import Lists.ListSet.
Require Import SfLib.
Require Import Recdef.


(*
  Auxilliary lemmas on lists and sets.
*)



Definition subset {A : Type} (x y : set A): Prop :=
  forall (a: A), set_In a x -> set_In a y.


Definition set_Eq {A : Type} (x y : set A) : Prop := 
  subset x y /\ subset y x.

Fixpoint lasto {A : Type} (l : list A) : option A :=
  match l with 
  | [] => None
  | [a] => Some a
  | a :: ls => lasto ls
  end.

Lemma subset_refl : forall {A : Type} (x : set A),
  subset x x.
Proof.
  intros A x. unfold subset. intros a H. trivial.
Qed.

Lemma subset_trans: forall {A : Type} (x y z : set A),
  subset x y -> subset y z -> subset x z.
Proof.
  unfold subset.
  intros A x y z H1 H2.
  intros a I.
  apply H2. 
  apply H1.
  trivial.
Qed.


Lemma empty_iff_no_elements : forall (A : Type)
                                 (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                                 (x : set A),
  x = [] <-> forall (a : A), not (In a x).
Proof.
  split.
  Case "->".
    intros E. intros a C. rewrite E in C. inversion C.
  Case "<-".
    intros H. destruct x as [ | e xs].
    trivial.
    assert (C : In e (e:: xs)). simpl. left. trivial.
      apply H in C. inversion C.
Qed.

(*
  Lemmas concerning length.
*)
Lemma length_add : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s : set A) (a : A),
  length (set_add Aeq_dec a s) <= S (length s).
Proof.
  intros A Aeq s. induction s as [ | u ss]; intros a.
  Case "s = nil".
    simpl. trivial.
  Case "s = u :: ss".
    remember (Aeq a u) as b. destruct b.
    SCase "b = true".
      simpl. rewrite <- Heqb. simpl. constructor. trivial.
    SCase "b = false".
      simpl. rewrite <- Heqb. simpl. apply le_n_S. apply IHss.
Qed.
  

Lemma add_length : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s : set A) (a : A),
  length s <= length (set_add Aeq_dec a s).
Proof.
  intros A Aeq s. induction s as [ | e ss]; intros a.
  simpl. omega.
  simpl. destruct (Aeq a e).
    simpl. omega.
    simpl. apply le_n_S. apply IHss.
Qed.

Lemma length_diff : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s: set A) (t : set A),
  length (set_diff Aeq_dec s t) <= length s.
Proof.
  intros A Aeq s. induction s as [ | a ss] ; intros t.
  Case "s = nil".
   simpl. trivial.
  Case "s = a :: ss".
   remember (set_mem Aeq a t) as b. destruct b.
   SCase " b = true".
     simpl. rewrite <- Heqb. apply le_trans with (m := length ss). apply IHss.
     constructor. trivial.
   SCase "b = false".
     simpl. rewrite <- Heqb. 
     assert (length (set_add Aeq a (set_diff Aeq ss t)) <=
                 S (length (set_diff Aeq ss t))) as L.
       apply length_add.
     apply le_trans with (m := S (length (set_diff Aeq ss t))).
     apply L. apply le_n_S. apply IHss.
Qed.
 

Lemma length_set_diff_not_empty : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s : set A) (l : set A),
  set_inter Aeq_dec s l <> nil -> 
      length (set_diff Aeq_dec s l) < length s.
Proof.
  intros A Aeq s. induction s as [ | a ss] ; intros l H.
  Case "s = nil".
    simpl in H. apply ex_falso_quodlibet. apply H. trivial.
  Case "s = a :: ss".
    remember (set_mem Aeq a l) as b.
    destruct b.
      SCase "b = true".
        simpl. rewrite <- Heqb.
        assert (length (set_diff Aeq ss l) <= length ss) as L.
          apply length_diff.
        apply le_lt_n_Sm. apply L.
      SCase "b = false".
        simpl. rewrite <- Heqb.
        assert (length (set_add Aeq a (set_diff Aeq ss l)) <=
                  S (length (set_diff Aeq ss l))) as L.
          apply length_add.
        assert (set_inter Aeq (a :: ss) l = set_inter Aeq ss l) as I.
          simpl. rewrite <- Heqb. trivial.
        rewrite I in H. apply IHss in H. apply lt_n_S in H.
        apply le_lt_trans with (m := S (length (set_diff Aeq ss l))).
          trivial. trivial.
Qed.


Lemma length_diff_not_empty_decreases : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s l: set A)  (n : nat),
  length s < S n -> 
    set_inter Aeq_dec s l <> nil ->
      length (set_diff Aeq_dec s l) < n.
Proof.
  intros A Aeq s l n L I. apply lt_S_n. 
  assert (length (set_diff Aeq s l) < length s) as HL.
    apply length_set_diff_not_empty. trivial.
  omega.
Qed.

Lemma length_union1 : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (x y : set A),
   length x <= length (set_union Aeq_dec x y).
Proof.
  intros A Aeq x y. generalize dependent x. induction y as [ | a ys]; intros x.
  simpl. omega.
  simpl. apply le_trans with (m := length (set_union Aeq x ys)).
    apply IHys.
    apply add_length.
Qed.

(*
  Lemmas concerning intersection.
*)


Lemma inter_mem : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s t : set A) (a : A),
  set_mem Aeq_dec a t = false ->
    set_inter Aeq_dec t (a :: s) = set_inter Aeq_dec t s.
Proof.
  intros A Aeq s t. generalize dependent s. 
  induction t as [ | u  ts]; intros s a H.
  Case "t  = nil".
    simpl. trivial.
  Case "t = u :: ts".
    simpl. remember (Aeq u a) as eua. destruct eua.
    SCase "u = a".
      simpl in H. destruct (Aeq a u). inversion H.
      apply ex_falso_quodlibet. apply n. auto.
    SCase "u <> a".
      simpl in H. destruct (Aeq a u). inversion H.
      rewrite IHts. trivial. trivial.
Qed.

Lemma inter_cons_not_empty : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s t : set A) (a : A),
  set_inter Aeq_dec s t <> [] -> set_inter Aeq_dec s (a :: t) <> [].
Proof.
  intros A Aeq s. induction s as [ | u ss].
  Case "s = nil".
    intros t a H. simpl in H. apply ex_falso_quodlibet. apply H. trivial.
  Case "s = u :: ss".
    intros t a H. simpl. remember (Aeq u a) as eua. destruct eua.
      subst. intros I. inversion I.
      remember (set_mem Aeq u t) as mut. destruct mut.
        intros C. inversion C.
        simpl in H. rewrite <- Heqmut in H. apply IHss. trivial.
Qed.


Lemma inter_empty_empty : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s : set A),
  set_inter Aeq_dec s [] = [].
Proof.
  intros A Aeq s. induction s as [ | a ss].
    simpl. trivial.
    simpl. apply IHss.
Qed.


Lemma inter_inter_not_empty : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s t : set A),
  set_inter Aeq_dec t s <> [] ->
    set_inter Aeq_dec (set_inter Aeq_dec t s) s <> [].
Proof.
  intros A Aeq s t. generalize dependent s. induction t as [ | u ts].
  Case "t = nil".
    intros s H. simpl in H. apply ex_falso_quodlibet. apply H. trivial.
  Case "t = u :: ts".
    intros s H. simpl. remember (set_mem Aeq u s) as mus. destruct mus.
    SCase "u in s".
      simpl. rewrite <- Heqmus. intros C. inversion C.
    SCase "u not in s".
      apply IHts. simpl in H. rewrite <- Heqmus in H. trivial.
Qed.


Lemma inter_not_empty_com : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (s t : set A),
  set_inter Aeq_dec s t <> [] <-> set_inter Aeq_dec t s <> [].
Proof.
  intros A Aeq s. induction s as [ | a ss].
  Case "s = nil".
    intros t. split. 
    SCase "->".
      intros H. simpl in H. apply ex_falso_quodlibet. apply H. trivial.
    SCase "<-".
      intros H. simpl. intros C. apply H. apply inter_empty_empty.
  Case "s = a :: ss".
    intros t. split.
    SCase "->".
      intros H. simpl in H. remember (set_mem Aeq a t) as mat. destruct mat.
      SSCase "a in t".
        induction t as [ | u ts].
        simpl. apply ex_falso_quodlibet. simpl in Heqmat. inversion Heqmat.
        simpl. remember (Aeq u a) as eua. destruct eua.
          intros C. inversion C.
          remember (set_mem Aeq u ss) as muss. destruct muss.
            intros C. inversion C.
            apply IHts. simpl in Heqmat. remember (Aeq a u) as eau. destruct eau.
              apply ex_falso_quodlibet. apply n. auto.
              trivial. intros C. inversion C.
      SSCase "a not in t".
       induction t as [ | u ts].
       SSSCase "t = nil".
         apply ex_falso_quodlibet. apply H. apply inter_empty_empty.
       SSSCase "t = u ::ts".
         simpl. remember (Aeq u a) as eua. destruct eua.
           intros C. inversion C.
           remember (set_mem Aeq u ss) as muss. destruct muss.
             intros C. inversion C.
             apply IHts. simpl in Heqmat. remember (Aeq a u) as eau. destruct eau.
               inversion Heqmat.
               trivial.
               rewrite IHss in H. simpl in H. rewrite <- Heqmuss in H. 
                  rewrite IHss. trivial.
    SCase "<-".
      intros H. simpl. remember (set_mem Aeq a t) as mat. destruct mat.
      SSCase "a in t".
        intros C. inversion C.
      SSCase "a not in t".
        induction t as [ | u ts].
        SSSCase "t = nil".
          intros C. apply H. simpl. trivial.
        SSSCase "t = u::ts".
          rewrite IHss. simpl. remember (set_mem Aeq u ss) as muss. destruct muss.
          SSSSCase "u in ss".
            intros C. inversion C.
          SSSSCase "u not in ss".
            rewrite <- IHss. apply IHts. simpl in H.
            remember (Aeq u a) as eua. destruct eua.
              subst. simpl in Heqmat. destruct (Aeq a a).
                inversion Heqmat. apply ex_falso_quodlibet. apply n. trivial.
              rewrite <- Heqmuss in H. trivial.
              simpl in Heqmat. remember (Aeq a u) as eau. destruct eau.
                inversion Heqmat. trivial.
Qed.

Lemma inter_included1 : forall {A : Type}
                         (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                         (x y z : set A),
  subset x z -> subset (set_inter Aeq_dec x y) z.
Proof.
  intros A Aeq x y z. unfold subset. intros Sxz a Ia.
  apply set_inter_elim1 in Ia. apply Sxz. trivial.
Qed.

Lemma inter_included2 : forall {A : Type}
                         (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                         (x y z : set A),
  subset y z -> subset (set_inter Aeq_dec x y) z.
Proof.
  intros A Aeq x y z. unfold subset. intros Syz a Ia.
  apply set_inter_elim2 in Ia. apply Syz. trivial.
Qed.


(* 
  Lemmas concerning union.
*)

Lemma union_included : forall {A : Type} 
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b}) 
                        (x y : set A),
  subset x (set_union Aeq_dec x y).
Proof.
  intros A Aeq_dec x y. generalize dependent x. induction y as [ | a ys]; intros x.
  Case "y = nil".
    simpl. apply subset_refl. 
  Case "y = a :: ys".
    unfold subset. intros e. apply set_union_intro1.
Qed.

Lemma union_included1 : forall {A : Type}
                          (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                          (x y z: set A),
  subset x z -> subset y z -> subset (set_union Aeq_dec x y) z.
Proof.
  intros A Aeq x y z. unfold subset. intros Sxz Syz.
  intros a Ia.
  apply set_union_elim in Ia. inversion Ia as [Iax | Iay].
  apply Sxz. trivial.
  apply Syz. trivial.
Qed.

Lemma union_included2 : forall {A : Type} 
                           (Aeq_dec : forall (a b : A),{a = b} + {a <> b}) 
                           (x y : set A),
  subset x (set_union Aeq_dec y x).
Proof.
  intros A Aeq_dec x. induction x as [ | a xs]; intros y.
  Case "x = nil".
    unfold subset. intros a C. inversion C.
  Case "x = a::xs".
    simpl. unfold subset. intros e S.
    destruct (Aeq_dec e a).
      SCase "e = a".
        apply set_add_intro2. trivial.
      SCase "e <> a".
        assert (He: set_In e xs).
          simpl in S. destruct S as [C | T].
          apply ex_falso_quodlibet. apply n. auto.
          trivial.
        apply set_add_intro. right. apply IHxs. trivial.
Qed.


Lemma union_included_union : forall {A : Type} 
                           (Aeq_dec : forall (a b : A),{a = b} + {a <> b}) 
                           (x y z t: set A),
  subset x z ->
    subset y t ->
      subset (set_union Aeq_dec x y) (set_union Aeq_dec z t).
Proof.
  unfold subset.
  intros A Aeq x y z t.
  intros Sxz Syt.
  intros a Uxy.
  apply set_union_elim in Uxy.
  inversion Uxy.
    apply set_union_intro1. apply Sxz. trivial.
    apply set_union_intro2. apply Syt. trivial.
Qed.


(*
  Lemmas concerning subset.
*)


Lemma diff_included : forall {A: Type}
                         (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                         (x y z : set A),
  subset x z -> subset (set_diff Aeq_dec x y) z.
Proof.
  intros A Aeq x y z. unfold subset. intros Sxz a Ia.
  apply set_diff_elim1 in Ia. apply Sxz. trivial.
Qed.


Lemma cons_included : forall {A : Type}
                        (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                        (x y : set A) (a : A),
  subset (a :: x) y -> set_In a y.
Proof.
  intros A Aeq x. intros y a S.
  unfold subset in S. apply S. simpl. left. trivial.
Qed.
    
Lemma cons_subset : forall {A: Type}
                      (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                      (x y : set A) (a : A),
  subset (a :: x) y -> subset x y.
Proof.
  intros A Aeq x y a H.
  unfold subset.
  intros c Ic.
  unfold subset in H.
  apply H.
  simpl. right. trivial.
Qed.



Lemma included_diff : forall (A : Type)
                             (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                             (x y z : set A),
  subset x z -> 
    set_inter Aeq_dec x y = [] -> 
      subset x (set_diff Aeq_dec z y).
Proof.
  unfold subset. intros A Aeq x y z Sxz E.
  intros a Sax. 
  apply set_diff_intro. apply Sxz. trivial.
  intros C. apply set_inter_intro with (Aeq_dec := Aeq) (x := x) in C.
  rewrite E in C. inversion C. trivial.
Qed.



Lemma subset_empty : forall {A : Type}
                            (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                            (x : set A),
  subset x [] -> x = [].
Proof.
  intros A Aeq x H.
  unfold subset in H.
  apply empty_iff_no_elements.
    trivial.
    intros a C.
    apply H in C.
    inversion C.
Qed.


Lemma included_diff_empty : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y z t: set A),
  subset x (set_diff Aeq_dec t y) ->
     subset t (set_union Aeq_dec z y) -> 
       set_inter Aeq_dec y z = [] ->
         subset x z.
Proof.
  intros A Aeq x y z t.
  unfold subset.
  intros Sxd Stu Iyz.
  intros a Iax.
  assert (Iat: set_In a t). 
    apply Sxd in Iax. apply set_diff_elim1 in Iax. trivial.
  assert (Nay : not (set_In a y)).
    apply Sxd in Iax. apply set_diff_elim2 in Iax. trivial.
  apply Stu in Iat.
  apply set_union_elim in Iat.
  inversion Iat.
    trivial.
    apply ex_falso_quodlibet. apply Nay. trivial.
Qed.


Lemma inter_empty_com : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y : set A),
  set_inter Aeq_dec x y = [] -> set_inter Aeq_dec y x = [].
Proof.
  intros A Aeq x y H.
  apply empty_iff_no_elements. trivial.
  intros a C.
  assert (Ix: In a x). 
    apply set_inter_elim2 with (Aeq_dec := Aeq) (x := y). trivial.
  assert (Iy : In a y).
    apply set_inter_elim1 with (Aeq_dec := Aeq)(y := x). trivial.
  rewrite empty_iff_no_elements in H.
  assert (Ixy : not (In a (set_inter Aeq x y))).
    apply H. 
  apply Ixy. apply set_inter_intro. trivial. trivial.
  trivial.
Qed.


Lemma inter_union_elim : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y : set A) (a : A),
    In a (set_union Aeq_dec x y) -> 
      ~In a y ->
         In a x.
Proof.
  intros A Aeq x y a Ia Na.
  apply set_union_elim in Ia.
  inversion Ia.
    trivial. 
    apply ex_falso_quodlibet. apply Na. trivial.
Qed.

Lemma included_trans_element  : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y : set A) (a : A),
  In a x -> 
    subset x y ->
      In a y.
Proof.
  intros A Aeq x y a I. unfold subset. intros SI.
  apply SI in I. trivial.
Qed.


Lemma included_diff_union : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y z t: set A),
  subset (set_union Aeq_dec x y) t ->
    subset z t ->
      subset (set_union Aeq_dec 
                         (set_union Aeq_dec x z)
                          (set_diff Aeq_dec y z)) t.
Proof.
  unfold subset. intros A Aeq x y z t SU Iz.
  intros a Ia.
  apply set_union_elim in Ia.
  destruct Ia as [Iaxz | Iayz].
  Case "a in union x z".
    apply set_union_elim in Iaxz. destruct Iaxz.
      apply SU. apply set_union_intro1. trivial.
      apply Iz. trivial.
  Case "a in diff y z".
    apply SU. apply set_union_intro2. 
    apply set_diff_elim1 with (Aeq_dec := Aeq)(y := z).
    trivial.
Qed.
  

Lemma union_diff_included : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y z t: set A),
  subset t (set_union Aeq_dec x y) ->
    subset z t ->
      subset t (set_union Aeq_dec 
                         (set_union Aeq_dec x z)
                         (set_diff Aeq_dec y z)).
Proof.
  unfold subset.
  intros A Aeq x y z t.
  intros SU Szt a Ia.
  apply set_union_intro.
  apply SU in Ia.
  apply set_union_elim in Ia. destruct Ia.
  Case "a in x".
    left. apply set_union_intro. left. trivial.
  Case "a in y".
    destruct (set_In_dec Aeq a z).
    SCase "a in z".
      left. apply set_union_intro. right. trivial.
    SCase "a not in z".
      right. apply set_diff_intro. trivial. trivial.
Qed.


Lemma inter_diff_empty : forall (A : Type)
                               (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                               (x y z t: set A),
  set_inter Aeq_dec x y = [] ->
    subset z t ->
      subset t (set_union Aeq_dec x y) ->
        set_inter Aeq_dec (set_union Aeq_dec x z) (set_diff Aeq_dec y z) = [].
Proof.
  unfold subset.
  intros A Aeq x y z t.
  intros Ixy Szt St.
  rewrite empty_iff_no_elements.
  intros a C. rewrite empty_iff_no_elements in Ixy.
  apply set_inter_elim in C. inversion C as [Caxz Cayz].
  apply set_union_elim in Caxz.
  inversion Caxz as [Cax | Caz].
  Case "a in x".
    apply set_diff_elim1 in Cayz.
    assert (Caxy: not (In a (set_inter Aeq x y))).
      apply Ixy. 
    apply Caxy. apply set_inter_intro. trivial. trivial.
  Case "a in z".
    apply set_diff_elim2 in Cayz. apply Cayz. trivial.
  trivial. trivial.
Qed.

Lemma add_included : forall (A : Type)
                            (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                            (e : A)
                            (x y : set A),
  subset x y -> subset (set_add Aeq_dec e x) (set_add Aeq_dec e y).
Proof.
  intros A Aeq e x y H.
  unfold subset.
  intros a I.
  apply set_add_intro.
  apply set_add_elim in I.
  inversion I as [E | S].
    left. trivial.
    right. apply H. trivial.
Qed.


Lemma add_included_sub : forall (A : Type)
                            (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                            (e : A)
                            (x y : set A),
  subset x y -> subset x (set_add Aeq_dec e y).
Proof.
  intros A Aeq e x y H.
  unfold subset.
  intros a I.
  apply set_add_intro. right. apply H. trivial.
Qed.

Lemma add_included_elim : forall (A: Type)
                            (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                            (e : A)
                            (x y : set A),
  set_In e y -> subset x y -> subset (set_add Aeq_dec e x) y.
Proof.
  intros A Aeq e x y He HS.
  unfold subset.
  intros a Ha.
  apply set_add_elim in Ha.
  inversion Ha as [E | I].
    subst. trivial.
    apply HS. trivial.
Qed.

Lemma add_add_com : forall (A : Type)
                      (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                            (a b: A)
                            (x y : set A),
  subset x (set_add Aeq_dec a (set_add Aeq_dec b y)) ->
        subset x (set_add Aeq_dec b (set_add Aeq_dec a y)).
Proof.
  unfold subset. intros A Aeq a b x y H.
  intros e Ie.
  apply set_add_intro.
  apply H in Ie.
  apply set_add_elim in Ie.
  inversion Ie as [Ea | Je].
    subst. right. apply set_add_intro. left. trivial.
    apply set_add_elim in Je. inversion Je as [Eb | Ke].
      left. trivial.
      right. apply set_add_intro. right. trivial.
Qed.

Lemma in_add_in_sub : forall (A : Type)
                      (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                      (a : A)
                      (x y : set A),
  set_In a y -> 
    subset x (set_add Aeq_dec a y) ->
      subset x y.
Proof.
  unfold subset.
  intros A Aeq a x y H S.
  intros e Ie.
  apply S in Ie.
  apply set_add_elim in Ie.
  inversion Ie as [E | I].
    subst. trivial.
    trivial.
Qed.

Lemma union_add_add_union : forall (A : Type)
                      (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                      (a : A)
                      (x y : set A),
  subset (set_union Aeq_dec (set_add Aeq_dec a x) (set_add Aeq_dec a y))
         (set_add Aeq_dec a (set_union Aeq_dec x y)).
Proof.
  intros A Aeq a x y.
  unfold subset.
  intros e H.
  apply set_add_intro.
  apply set_union_elim in H.
  inversion H as [H1 | H2].
    apply set_add_elim in H1. inversion H1 as [E1 | S1].
      tauto.
      right. apply set_union_intro. tauto.
    apply set_add_elim in H2. inversion H2 as [E2 | S2].
      tauto.
      right. apply set_union_intro. tauto.
Qed.
    
Lemma sub_union_com : forall (A : Type)
                      (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
                      (x y: set A),
  subset (set_union Aeq_dec x y) (set_union Aeq_dec y x).
Proof.
  intros A Aeq x y.
  unfold subset.
  intros a H.
  apply set_union_intro.
  apply set_union_elim in H.
  inversion H.
    right. trivial.
    left. trivial.
Qed.

Lemma sub_union_assoc1 : 
  forall (A : Type)
         (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
         (x y z: set A),
  subset (set_union Aeq_dec x (set_union Aeq_dec y z))
         (set_union Aeq_dec (set_union Aeq_dec x y) z).
Proof.
  intros A Aeq x y z.
  unfold subset.
  intros a H.
  apply set_union_intro.
  apply set_union_elim in H.
  inversion H as [X | YZ].
    left. apply set_union_intro. left. trivial.
    apply set_union_elim in YZ. inversion YZ.
      left. apply set_union_intro. right. trivial.
      tauto.
Qed.

Lemma sub_union_assoc2 : 
  forall (A : Type)
         (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
         (x y z: set A),
  subset (set_union Aeq_dec (set_union Aeq_dec x y) z)
         (set_union Aeq_dec x (set_union Aeq_dec y z)).
Proof.
  intros A Aeq x y z.
  unfold subset.
  intros a H.
  apply set_union_intro.
  apply set_union_elim in H.
  inversion H as [XY | Z].
    apply set_union_elim in XY. inversion XY.
      tauto.
      right. apply set_union_intro. tauto.
    right. apply set_union_intro. tauto.
Qed.
  

Lemma union_empty : forall
  (A : Type)
  (Aeq_dec : forall (a b : A),{a = b} + {a <> b})
  (x : set A),
  subset (set_union Aeq_dec [] x ) x.
Proof.
  intros A Aeq x.
  unfold subset.
  intros a H.
  apply set_union_elim in H.
  inversion H as [Se | S].
    inversion Se.
    trivial.
Qed.